/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "posix_simulator.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if defined(HAVE_POSIX_TIMERS)
// clock_gettime
#include <time.h>

#ifdef SIMULATION_STATISTICS
#include <iostream>
using namespace std;
#endif
#endif

#if defined(_WIN32) || defined(WIN32)
#include <windef.h>
// timeGetTime
#include <mmsystem.h>
#else

// gettimeofday
#include <sys/time.h>
#endif

// getenv
#include <stdlib.h>

static double SIMULATOR_MIN_CPU_RELEASE = 1e-3; // 1ms min by default

// global factory object
IMPLEMENT_SIMULATOR_FACTORY(50, POSIX_Simulator);

// timer loops forever and the thread is re-used
void *timer_routine( void * arg ) {
    // see thread creation
    POSIX_Simulator *simu = static_cast<POSIX_Simulator*>(arg);
    // timespec available from pthread.h even on win32 port
    struct timespec ts;
    while (true) {
        pthread_mutex_lock(&simu->timer_mutex);
        // detection for destruction. Use special negative delay as flag
        if (simu->waitDelay==-2) return 0;
        // down to nanosec matches high res timers. Thanks pthreads!
        // Unfortunately, at this point the goal is to release the CPU, not to aggressively
        // get it back => it's much better to have a longer delay and come back to this point
        // less often (avoids pthread locks & waits).
        if (simu->waitDelay < SIMULATOR_MIN_CPU_RELEASE) simu->waitDelay = SIMULATOR_MIN_CPU_RELEASE;
        ts.tv_sec = (long)(simu->waitDelay); // truncated
        ts.tv_nsec = (long)(simu->waitDelay * 1e9);
        if (ts.tv_nsec > 1000000000) ts.tv_nsec %= 1000000000;
        pthread_cond_timedwait(&simu->timer_condition, &simu->timer_mutex, &ts);
        // detection for timer cancellation. Use negative delay as flag
        if (simu->waitDelay>0) simu->waitTimerCallBack();
        // next loop locks and waits for main thread to unlock
    }
};

static pthread_t timer_thread;

POSIX_Simulator::POSIX_Simulator() : Simulator()
{
#if defined(HAVE_POSIX_TIMERS)
#ifdef SIMULATION_STATISTICS
    cout << "Using posix timers." << endl;
#endif
#endif

    char* envMinRelease = getenv("SIMULATOR_MIN_CPU_RELEASE");
    if (envMinRelease) {
        char** endptr = &envMinRelease;
        double d = strtod(envMinRelease, endptr);
        if (endptr != &envMinRelease) SIMULATOR_MIN_CPU_RELEASE = d; // conversion successful
    }

    pthread_mutex_init(&event_mutex, 0);
    pthread_mutex_init(&control_mutex, 0);
    sem_init(&wait_semaphore, 0, 0);
    pthread_mutex_init(&timer_mutex, 0);
    pthread_cond_init(&timer_condition, 0);

    pthread_mutex_init(&waitSemaphoreMutex, 0);
    pthread_cond_init(&waitSemaphoreCondition, 0);
    waitSemaphoreCounter = 0;

        
    // start lack of mutex in timer loop
    pthread_mutex_lock(&timer_mutex);
    
    // create the timer thread in detached state for automatic thread destruction
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    pthread_create(&timer_thread, &attr, timer_routine, this);
    pthread_attr_destroy(&attr);
    // timer thread is waiting on mutex unlock
}


POSIX_Simulator::~POSIX_Simulator()
{
    waitDelay = -2;
    // unlock timer loop and trigger destruction
    pthread_mutex_unlock(&timer_mutex);
    
    pthread_mutex_destroy(&waitSemaphoreMutex);
    pthread_cond_destroy(&waitSemaphoreCondition);
    pthread_mutex_destroy(&event_mutex);
    pthread_mutex_destroy(&control_mutex);
    sem_destroy(&wait_semaphore);
    pthread_cond_destroy(&timer_condition);
    pthread_mutex_destroy(&timer_mutex);
}


void POSIX_Simulator::eventMutexLock( )
{
    pthread_mutex_lock(&event_mutex);
}

void POSIX_Simulator::eventMutexUnlock( )
{
    pthread_mutex_unlock(&event_mutex);
}


bool POSIX_Simulator::stepControlMutexTryLock( )
{
    return pthread_mutex_trylock(&control_mutex) == 0;
}

void POSIX_Simulator::stepControlMutexLock( )
{
    pthread_mutex_lock(&control_mutex);
}

void POSIX_Simulator::stepControlMutexUnlock( )
{
    pthread_mutex_unlock(&control_mutex);
}

void POSIX_Simulator::waitSemaphorePost( )
{
    // Implement our own semaphore because the cost of delete/create in flushing is too much with pthread
    pthread_mutex_lock(&waitSemaphoreMutex);
    ++waitSemaphoreCounter;
    pthread_cond_signal(&waitSemaphoreCondition);
    pthread_mutex_unlock(&waitSemaphoreMutex);

    //sem_post(&wait_semaphore);
}

void POSIX_Simulator::waitSemaphoreFlush( )
{
    // Implement our own semaphore because the cost of delete/create in flushing is too much with pthread
    pthread_mutex_lock(&waitSemaphoreMutex);
    waitSemaphoreCounter = 0;
    pthread_mutex_unlock(&waitSemaphoreMutex);
    
    //sem_destroy(&wait_semaphore);
    //sem_init(&wait_semaphore, 0, 0);
}

void POSIX_Simulator::waitSemaphoreWait( )
{
    // Implement our own semaphore because the cost of delete/create in flushing is too much with pthread
    pthread_mutex_lock(&waitSemaphoreMutex);
    while (waitSemaphoreCounter<=0) {
        pthread_cond_wait(&waitSemaphoreCondition, &waitSemaphoreMutex);
    }
    --waitSemaphoreCounter;
    pthread_mutex_unlock(&waitSemaphoreMutex);

    //sem_wait(&wait_semaphore);
}

void * run_routine( void * arg )
{
    POSIX_Simulator *simu = static_cast<POSIX_Simulator*>(arg);
    // Event loop might wait some time before acquiring the stepControlMutex
    simu->eventLoop();
}

void POSIX_Simulator::start( bool threaded )
{
    if (!threaded) {
        eventLoop();
        return;
    }
    // create the thread in detached state for automatic thread destruction
    // on thread exit
    // This function returns immediately, but the created thread might
    // wait some time before acquiring the stepControlMutex
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    pthread_t run_thread;
    pthread_create(&run_thread, &attr, run_routine, this);
    pthread_attr_destroy(&attr);
}


void POSIX_Simulator::startWaitTimer( double delay )
{
    if (delay<=0) {                        // should not happen!
        waitTimerCallBack();
        return; 
    }
    waitDelay = delay;
    // unlock timer loop and trigger condition wait
    pthread_mutex_unlock(&timer_mutex);
}

void POSIX_Simulator::cancelWaitTimer( )
{
    // avoid dead-lock if no timer running
    if (pthread_mutex_trylock(&timer_mutex)) {
        //cerr << "trylock failed!" << endl;
        return; // don't cancel before the timer is waiting! Might happen
    }
    // flag to indicate the stop
    waitDelay = -1;
    // stop the wait for the condition
    pthread_cond_signal(&timer_condition);
    // allow the timer thread to continue execution
    pthread_mutex_unlock(&timer_mutex);
}


double POSIX_Simulator::getSystemTime( )
{

// POSIX_TIMERS => nanosec precision
// "Old" unix   => microsec precision, but wrap each day: TODO!
// Windows      => millisec precision

// The configure flag is positionned only if the function is available
#if defined(HAVE_POSIX_TIMERS)

    // Hey, high resolution clock is available. Nice!
    struct timespec ts;
    if (clock_gettime(CLOCK_MONOTONIC, &ts)==0) // 0 is success...
        return ts.tv_sec + ts.tv_nsec * 1e-9;
    // error using high-res clock => use standard clock
#endif

#if defined(_WIN32) || defined(WIN32)
    return timeGetTime() * 1e-3;

#else
    // Fallback for HAVE_POSIX_TIMERS. Too bad, let's use another POSIX call
    struct timeval tv;
    gettimeofday(&tv, 0); // don't care about timezone
    return tv.tv_sec + tv.tv_usec * 1e-6;
#endif

}

