/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>

#include <osg/Geode>
#include <osg/ShapeDrawable>

#include "AgentCore.h"
using namespace Environment;
using namespace std;

double AgentCore::maxForce = 1.0;
double AgentCore::maxSpeed = 1.0;
double AgentCore::maxForceSq = 1.0;
double AgentCore::maxSpeedSq = 1.0;
double AgentCore::energyFootprintRatio = 0.0;

double AgentCore::maxEnergy = Infinity; // See environment.h
double AgentCore::initialMass = 1.0;
double AgentCore::selfMaintainEnergyCost = 0;

bool AgentCore::optimizeBounds = true;

AgentCore::AgentCore()
{
    position.set(0,0,0);
    scale.set(1,1,1);
    pivotPoint.set(0,0,0);
    attitude.set(0,0,0,1); // null rotate quaternion
    if (!batch) {
        transform = new NoDirtyBoundTransform;
        transform->setPosition(position);
        transform->setAttitude(attitude);
        transform->setScale(scale);
        transform->setPivotPoint(pivotPoint);
        transformDirtyBound = false;
    } else transform=0;
    mass = initialMass;
    // Use max instead of 0 for scenarii to do something by default
    energy = maxEnergy; // no point in calling a virtual function in a constructor. This is C++, not Java...

    radius = 0.866025403785; // half sqrt(3) = cube radius
}


//#include <iostream>
// Helper code which will be executed at next OSG update
static void nodeRemoverCallback(osg::Node* node) {


// It appears there is probably a memory corruption bug somewhere.
// It causes dummy parents to appear in the vector, then... crash
// So, do this hack in the meanTime

    // process only one parent
    node->getParent(0)->removeChild( node );
    return;

    // Thanks to Robert Osfield in message on osg-user ML Date: Sun Jan 19 2003 - 02:18:09 PST
    osg::ref_ptr<osg::Node> nodeToRemove = node;
//cout << "getNumParents()=" << node->getNumParents() << endl;
    while( nodeToRemove->getNumParents() > 0 ) {
//cout << "n ref=" << node->referenceCount() << endl;
        osg::Group* g = nodeToRemove->getParent(0);
//cout << "g ref=" << g->referenceCount() << endl;
//cout << hex << g << dec << endl;
        g->removeChild( node );
    }
    // nodeToRemove goes out of scope, unref's the node, if its the last
    // node it will be deleted.


}

AgentCore::~AgentCore() {
    if (transform != 0) {
        viewer->post(&nodeRemoverCallback, static_cast<osg::Node*>(transform));
        transform = 0;
    }
}


void AgentCore::setEnergyFootprintRatio( double newRatio )
{
    energyFootprintRatio = newRatio;
}

double AgentCore::getEnergyFootprintRatio( )
{
    return energyFootprintRatio;
}

void AgentCore::setDefaultInitialMass( double mass )
{
    initialMass = mass;
}

double AgentCore::getDefaultInitialMass( )
{
    return initialMass;
}

void AgentCore::setMaxForce( double newMax )
{
    maxForce = newMax;
    maxForceSq = maxForce * maxForce;
}

double AgentCore::getMaxForce( )
{
    return maxForce;
}

double AgentCore::getMaxForceSquared( )
{
    return maxForceSq;
}

void AgentCore::setMaxSpeed( double newMax )
{
    maxSpeed = newMax;
    maxSpeedSq = maxSpeed * maxSpeed;
}

double AgentCore::getMaxSpeed( )
{
    return maxSpeed;
}

double AgentCore::getMaxSpeedSquared( )
{
    return maxSpeedSq;
}


void AgentCore::setDefaultEnergyFootprintRatio( double newRatio )
{
    energyFootprintRatio = newRatio;
}

double AgentCore::getDefaultEnergyFootprintRatio( )
{
    return energyFootprintRatio;
}

void AgentCore::setDefaultMaxForce( double newMax )
{
    maxForce = newMax;
    maxForceSq = maxForce * maxForce;
}

double AgentCore::getDefaultMaxForce( )
{
    return maxForce;
}

double AgentCore::getDefaultMaxForceSquared( )
{
    return maxForceSq;
}

void AgentCore::setDefaultMaxSpeed( double newMax )
{
    maxSpeed = newMax;
    maxSpeedSq = maxSpeed * maxSpeed;
}

double AgentCore::getDefaultMaxSpeed( )
{
    return maxSpeed;
}

double AgentCore::getDefaultMaxSpeedSquared( )
{
    return maxSpeedSq;
}

void AgentCore::setMaxEnergy( double newMax )
{
    maxEnergy = newMax;
}

double AgentCore::getMaxEnergy( )
{
    return maxEnergy;
}

void AgentCore::setDefaultMaxEnergy( double newMax )
{
    maxEnergy = newMax;
}

double AgentCore::getDefaultMaxEnergy( )
{
    return maxEnergy;
}


void AgentCore::setSelfMaintainEnergyCost( double newCost )
{
    selfMaintainEnergyCost = newCost;
}

double AgentCore::getSelfMaintainEnergyCost( )
{
    return selfMaintainEnergyCost;
}

void AgentCore::setDefaultSelfMaintainEnergyCost( double newCost )
{
    selfMaintainEnergyCost = newCost;
}

double AgentCore::getDefaultSelfMaintainEnergyCost( )
{
    return selfMaintainEnergyCost;
}



const osg::Vec3d& AgentCore::getPosition() const { return position; }
const osg::Quat& AgentCore::getAttitude() const { return attitude; }
const osg::Vec3d& AgentCore::getScale() const { return scale; }
const osg::Vec3d& AgentCore::getPivotPoint() const { return pivotPoint; }

void AgentCore::setPosition(const osg::Vec3d& pos) {
    Localized::setPosition(pos);
    if (!batch) transform->setPosition(position);
}

void AgentCore::setPosition(const double& x, const double& y, const double& z) {
    
    Localized::setPosition(x,y,z);
    if (!batch) transform->setPosition(position);
}

void AgentCore::setAttitude(const osg::Quat& quat) {
    
    attitude = quat;
    if (!batch) transform->setAttitude(quat);
}

void AgentCore::setAttitude(const double& x, const double& y, const double& z, const double& w) {
    
    attitude.set(x,y,z,w);
    if (!batch) transform->setAttitude(attitude);
}

void AgentCore::setScale(const osg::Vec3d& sca) {
    
    scale = sca;
    if (!batch) transform->setScale(sca);
}

void AgentCore::setScale(const double& x, const double& y, const double& z) {
    
    scale.set(x,y,z);
    if (!batch) transform->setScale(scale);
}

void AgentCore::setPivotPoint(const osg::Vec3d& pivot) {
    
    pivotPoint = pivot;
    if (!batch) transform->setPivotPoint(pivot);
}

void AgentCore::setPivotPoint(const double& x, const double& y, const double& z) {
    
    pivotPoint.set(x,y,z);
    if (!batch) transform->setPivotPoint(pivotPoint);
}

// modification to the scene graph is done at the right time
static void addAvatarHelper(osg::Group * group, osg::Node* transform) {
    group->addChild(transform);
}

void AgentCore::addAvatarToGroup( osg::Group * group, bool delayToNextFrame )
{
    if (transform) {
        osg::Node *avatar = buildAvatar();
        transform->addChild(avatar);
        if (delayToNextFrame) viewer->post(&addAvatarHelper, group, static_cast<osg::Node*>(transform));
        else addAvatarHelper(group, transform);
    }
}

// Shared geometry between all the agents, thanks to osg ref counting
// private to this compilation unit
static osg::Node *commonAgentCoreAvatar = 0;

// Code shamelessly stolen from OSG tutorial
// Changed to share the node object
osg::Node * AgentCore::buildAvatar( )
{
    // will be ref counted when added to another parent!
    if (commonAgentCoreAvatar!=0) return commonAgentCoreAvatar;

    // create a block shape
    osg::Box *bCube = new osg::Box(osg::Vec3(0,0,0),1,2,0.5);
        
    // create a container that makes the cube drawable
    osg::ShapeDrawable *sdCube = new osg::ShapeDrawable(bCube);
    
    sdCube->setColor(osg::Vec4(1.0f,1.0f,1.0f,1.0f));

    // create a geode object as a container for our drawable cube object
    osg::Geode* geodeCube = new osg::Geode();   
    
    // add our drawable cube to the geode container
    geodeCube->addDrawable(sdCube);

    commonAgentCoreAvatar = geodeCube;
    return commonAgentCoreAvatar;
}

void AgentCore::NoDirtyBoundTransform::update( )
{
    updateMutex.lock();
    // no need to test for dirty to be true: it is, at this point (protected method was posted)
    if (AgentCore::optimizeBounds)
        // don't notify parent, and since we're in callback called at OSG frequency
        // we may safely assume this is best we can do for as less frequently updating
        // bounds as necessary.
        computeBound();
    else
        // otherwise propagate the change the full way and transform will be recomputed later on
        dirtyBound();
    dirty = false;
    updateMutex.unlock();
}

