/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ENVIRONMENT_H
#define ENVIRONMENT_H

#include <osg/Vec3d>
#include <osg/Group>

#include "simulator.h"
#include "EventQueueViewer.h"
#include "DrawableSporeCloud.h"

#include <limits>
#if (defined(_WIN32) || defined(WIN32)) && !defined(__MINGW32__)
#include "float.h"
#endif

/**

Environment contains variables for the physics, and the topological nature
of the world (cyclic in X/Y or not)

@author Nicolas Brodu
*/
namespace Environment {

    /**
        Set this value to the world diameter in X
        Default is 0.0, meaning no cycle.

        Note: This value is overriden by a call to createLocalityDatabase
        with a cyclic condition (the default). Thus, just calling createLocalityDatabase
        would set this value properly. This value is still settable for cases where
        a locality database isn't used.
    */
    extern double cycleX;
    
    /**
        Set this value to the world diameter in Y
        Default is 0.0, meaning no cycle.

        Note: This value is overriden by a call to createLocalityDatabase
        with a cyclic condition (the default). Thus, just calling createLocalityDatabase
        would set this value properly. This value is still settable for cases where
        a locality database isn't used.
    */
    extern double cycleY;

    /**
        Set this value to the offset in world position which represents a boundary limit
        Default is 0.0

        Note: This value is overriden by a call to createLocalityDatabase
        with a cyclic condition (the default). Thus, just calling createLocalityDatabase
        would set this value properly. This value is still settable for cases where
        a locality database isn't used.
    */
    extern double cycleOffsetX;

    /**
        Set this value to the offset in world position which represents a boundary limit
        Default is 0.0
        
        Note: This value is overriden by a call to createLocalityDatabase
        with a cyclic condition (the default). Thus, just calling createLocalityDatabase
        would set this value properly. This value is still settable for cases where
        a locality database isn't used.
    */
    extern double cycleOffsetY;

    /** Drag allows moving objects to stop when no energy remains.
        Moving objects are slowed down by dragCoefficient times their
        squared speed value. Default value is 0.0.
    */
    extern double dragCoefficient;

    /** Gravity takes place after clamping to max force, it's not of
        the agents resort. It adds a -z component to the force.

        Note: gravity is only applied when calling setSteeringForce, which is
        normally called at AI update frequency. In the case of an object with
        no AI, please call setSteeringForce once after setting the gravity, or
        the object won't move at all!
        Note2: There is no gravity by default.
    */
    extern double gravity;

    /// Global simulation manager. Null by default, please allocate this according to your system
    extern Simulator* simu;

    /// Simulation time steps. To be provided by the main application
    extern double agentsAITimeStep;
    extern double agentsIntegrationTimeStep;
    extern double cloudIntegrationTimeStep;

    /// Whether to update the graphical osg parts or not
    extern bool batch;

    
    /// Window parameters. Defaults to 800, 600, 240, 212 (centered on 1280/1024)
    /// TODO: Auto-detect screen dimensions.
    /// Temporary workaround: set these values in your main program before allocating the viewer
    extern int WindowWidth;
    extern int WindowHeight;
    extern int WindowXOffset;
    extern int WindowYOffset;
    
    /// In graphical mode, global reference to the (modified) OSG Viewer.
    /// Null by default, please allocate this in the main program, with the next function if necessary
    extern EventQueueViewer* viewer;

    /// The spore cloud for the agents. Null by default, please set this up in the main program
    /// if you use it
    extern DrawableSporeCloud* cloud;
    extern osg::Group* root;
    
    /// allow to use const temporaries instead of declaring variables
    inline const double square(const double& x) {return x*x;}

    /** Compute the squared distance between 2 points, taking in account 
        the world wrap if this is the case.
     */
    const double wrappedDist2(const osg::Vec3d& p1, const osg::Vec3d& p2);

    /** Return a position inside the world primary cycle/offset from the
        given position. Has no effect when not wrapping
     */
    const osg::Vec3d wrapPosition(const osg::Vec3d& position);

    /** Return the p1 - p2 vector, taking in account the world wrap.
        This is not the same as wrapping the positions and substracting!
     */
    const osg::Vec3d wrappedDiffVector(const osg::Vec3d& p1, const osg::Vec3d& p2);

    /** OS dependent code: do our best */
    const double Infinity = 
    std::numeric_limits<double>::has_infinity ?
    std::numeric_limits<double>::infinity() :
#if (defined(_WIN32) || defined(WIN32)) && !defined(__MINGW32__)
    // for some reason numeric_limits<double>::max(); fails under Visual C++ 6.0
    DBL_MAX;
#else
    std::numeric_limits<double>::max();
#endif
    
}

#endif
