/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef EXPLICITAIAGENT_H
#define EXPLICITAIAGENT_H

#include <vector>

#include "CrogaiAgent.h"
#include "Environment.h"
#include "Logger.h"

/** Agent with hard-coded explicit AI.

    This agent will basically follow this simple algorithm:
        IF not enough energy THEN hunt
        ELSE IF good quiet place THEN spawn
        ELSE seek good quiet place

    The parameters like the amount of energy necessary to stop hunting,
    or the definition of a quiet place, have sensible default values, but
    left open to evolution.
    
    @author Nicolas Brodu
*/
class ExplicitAIAgent : public CrogaiAgent
{
public:
    
    ExplicitAIAgent(int specie, const std::vector<int>& predatorSpecies, const std::vector<int>& preySpecies);
    ~ExplicitAIAgent();

    virtual void updateAI( const double elapsedTime, const double currentTime);


    /// To be set to some relevant value by your application. Default is 120 simulator time
    static double sporeHalfLife;

    /// To be set to some relevant value by your application
    static double maxGrazingSpeed;
    
    /// To be set to some relevant value by your application
    static double sporeEmissionEnergyThreshold;
    
    /// To be set to some relevant value by your application
    static double sporeEmissionEnergyCost;

    /// To be set to some relevant value by your application
    /// Agents metabolize part of the energy they get into mass as growth process (see genome for ratio)
    static double maxMass;
    /// Then they can spend it to create children
    /// see genome for ratio of child mass/parent mass, child can have anything between
    /// min and (max-min) mass as initial value => maxMass should be at least > 2 * minMass
    /// or no child can ever be created (parent mass + child mass = 2*min mass at least)
    static double minMass;

    
    /// Log the previous statistics. These can't be updated in real time, and
    /// generations co-exist at the same time and their statistics may be
    /// updated in parallel (which would require to "go back" in the log file
    /// to update a value). => the logger is used only at the end of the program,
    /// when the last agent is destroyed, but it is also used as a flag indicator
    /// whether to bother computing the statistics at all!
    /// null by default
    static Logger* logGenomes;
    
    /// Creates the logGenomes with the given ID and max number of species
    /// utility routine
    static void createGenomesLogger(const std::string& id, int numSpecies);

    /// Tells the logger to record all its statistics now. It won't be possible to change
    /// the values afterward, so this should normally be called at the end of your program
    static void recordGenomesStatistics();

        
protected:

    /// hunt strategy for the AI
    virtual void hunt( const double elapsedTime, const double currentTime);
    
    /// graze strategy for the AI
    virtual void graze( const double elapsedTime, const double currentTime);

    /// Searching a quiet place for reproduction. Neighbors list was already updated
    virtual void searchQuietPlace( const double elapsedTime, const double currentTime);

    /// Check if here is a quiet place for reproduction. Neighbors list was already updated
    virtual bool checkForQuietPlace( const double elapsedTime, const double currentTime);

    struct Genome : public DrawableSporeCloud::ColoredGenome {
        // agent may have died, or may have multiple species with this AI. keep it here
        int specie;

        // the "genes"
        double hungryRatioThreshold;
        double hungryLastResortRatioThreshold;
        double hungryGrassWeight;
        double hungryPreyWeight;
        double hungryPredatorWeight;
        double hungryWeightedThreshold;
        double grazeRate;
        double grazeNextAmountThreshold;
        double grazeForceCoef;
        double huntMateForwardCoef;
        double huntMateEvasionCoef;
        double huntWanderCoef;
        double huntWanderRate;
        double huntAvoidCoef;
        double huntPreyCoef;
        double huntPredatorCoef;
        double huntMateCoef;
        double seekQuietPreyCutoff;
        double seekQuietOtherEvasionCoef;
        double seekQuietWanderCoef;
        double seekQuietWanderRate;
        double seekQuietAvoidCoef;
        double seekQuietPreyCoef;
        double seekQuietPredatorCoef;
        double seekQuietOtherCoef;
        double quietWeightForObstacle;
        double quietWeightForUnknown;
        double quietWeightForPrey;
        double quietWeightForPredator;
        double quietWeightForMate;
        double quietThreshold;
        double energyForFightRatio;
        double energyConvertionRatio;
        double childEnergyRatio;
        double childMassRatio;

        // provide default values
        Genome(int specie);

        // provide 0-initialized genome, for stat usage
        Genome();
    };


    /// Keep statistics about genomes mean on a per-specie, per generation basic
    /// Map of specie -> generation vector of genomes.
    /// Note: this is just the sum, it will still need to be divided by the total numbers
    ///       in the generationsPopulationNumber map for the final mean values
    static std::map<int, std::vector<Genome*> > genomesMean;
    
    /// Keep statistics about genomes variance on a per-specie, per generation basic
    /// Map of specie -> generation vector of genomes
    /// Note: this is just the sum of squares, it will still need to be substracted the
    ///       squared mean and divided by the total numbers in the
    ///       generationsPopulationNumber map for the final variance values
    static std::map<int, std::vector<Genome*> > genomesVariance;


    /// spawn, using the specified genome as spore
    virtual void spawn(Genome* spore);
    
    /// This agent genome. For internal use during reproduction
    Genome *genome;

    std::vector<int> predatorSpecies, preySpecies;

    /// Overload from parent: get the color from genome
    osg::Node *buildAvatar();

    virtual bool isPrey(CrogaiAgent* agent);
    
    virtual bool isPredator(CrogaiAgent* agent);

    virtual double getEnergyForFight();

    /// emit spores when gaining enough energy!
    virtual void energyGainAction(double gain);

    // cache for critical mass to be able to create a child
    double spawnCriticalMass;
};

#endif
