/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>

#include "KeepUpAttitudeCorrection.h"
#include <iostream>

KeepUpAttitudeCorrection::KeepUpAttitudeCorrection( const osg::Vec3d & targetUpVector, const osg::Vec3d & initUp )
 : targetUp(targetUpVector), oldUp(initUp)
{
}

void KeepUpAttitudeCorrection::set( const osg::Vec3d & targetUpVector, const osg::Vec3d & initUp )
{
    targetUp = targetUpVector;
    oldUp = initUp;
}



KeepUpAttitudeCorrection::~KeepUpAttitudeCorrection()
{
}

void KeepUpAttitudeCorrection::updateAttitude( osg::Quat & newAttitude, const osg::Quat & oldAttitude, const osg::Vec3d & newForward, const osg::Vec3d & oldForward )
{
    // Compute new attitude: find quaternion that convert initial velocity
    // to new velocity vectors, and multiply current attitude by it.
    // No costly acos or sin are used, the quaternion is found only
    // with fast geometrical properties.
    // See for example http://logiciels.cnes.fr/MARMOTTES/marmottes-mathematique.pdf
    // for an explanation of the math used.


    double dotProdPlus1 = 1.0 + oldForward * newForward;
    // Check for degenerate case of full u-turn. Use epsilon for detection
    if (dotProdPlus1 < 1e-7) {
        // Get an orthogonal vector of the given vector
        // in a plane with maximum vector coordinates.
        // Then use it as quaternion axis with pi angle
        if (fabs(oldForward.x()) < 0.6) {
            const double norm = sqrt(1.0 - oldForward.x() * oldForward.x());
            qpassage.set(0.0, oldForward.z()/norm, -oldForward.y()/norm, 0.0);
        } else if (fabs(oldForward.y()) < 0.6) {
            const double norm = sqrt(1.0 - oldForward.y() * oldForward.y());
            qpassage.set(-oldForward.z()/norm, 0.0, oldForward.x()/norm, 0.0);
        } else {
            const double norm = sqrt(1.0 - oldForward.z() * oldForward.z());
            qpassage.set(oldForward.y()/norm, -oldForward.x()/norm, 0.0, 0.0);
        }
    }
    else {
        // Find the shortest angle quaternion that transforms normalized vectors
        // into one other. Then, multiply it to the current quaternion.
        const double s = sqrt(0.5 * dotProdPlus1);
        const double* ptr = (oldForward ^ newForward / (2.0*s)).ptr();
        qpassage.set(ptr[0], ptr[1], ptr[2], s);
    }

    // update new attitude
    newAttitude = oldAttitude * qpassage;
    
    // what has become old up position?
    oldUp = qpassage * oldUp;

    // what we seek: new up in the plane of target up
    newUp = targetUp - newForward * (targetUp * newForward);
    double norm = newUp.normalize();

    // problem when new forward is opposite of target Up!!!
    // in this case, keep old Up to avoid insane rotations around forward axis
    if (norm<1e-7) return;

    // rotate again, around current forward axis this time
    dotProdPlus1 = 1.0 + oldUp * newUp;
    // Check for degenerate case of full u-turn. Use epsilon for detection
    if (dotProdPlus1 < 1e-7) {
        qpassage.set(newForward.x(), newForward.y(), newForward.z(), 0);
    }
    else {
        // Find the shortest angle quaternion that transforms normalized vectors
        // into one other. Then, multiply it to the current quaternion.
        const double s = sqrt(0.5 * dotProdPlus1);
        const double* ptr = (oldUp ^ newUp / (2.0*s)).ptr();
        qpassage.set(ptr[0], ptr[1], ptr[2], s);
    }

    // new forward unchanged by second rotation
    oldUp = newUp;
    newAttitude *= qpassage;

    // Renormalize the quaternion if it is too badly conditionned after many multiplications
    // reuse variable, at the expense of clarity.
    dotProdPlus1 = newAttitude.length2();
    
    // Use less restrictive conditions
    if ((dotProdPlus1<0.9999) || (dotProdPlus1>1.0001)) {

        newAttitude /= sqrt(dotProdPlus1);

        // Sanity check: after that many rotations, ensure forward direction is still OK
        // => local space forward and attitude tranformed world space forward should match
        //if (newForward * (newAttitude * osg::Vec3d(0,1,0)) < 1.0 - 1e-7) {
        if (newForward * osg::Vec3d(-newAttitude.z(),newAttitude.w(),newAttitude.x()) < 1.0 - 1e-7) {
            // avoid infinite recursion? This shouldn't happen with double precision
            if (oldAttitude.w()!=1) {
                oldUp.set(0,0,1);
                updateAttitude(newAttitude, osg::Quat(0,0,0,1), newForward, osg::Vec3d(0,1,0));
            }
        }
    }
    
}

void KeepUpAttitudeCorrection::matchTarget( osg::Quat & newAttitude, osg::Vec3d & newForward )
{

    double dotProdPlus1 = 1.0 + oldUp * targetUp;

    // nothing to do if vectors already match
    if (dotProdPlus1 > 2.0 - 1e-7) return;
    
    // Check for degenerate case of full u-turn. Use epsilon for detection
    if (dotProdPlus1 < 1e-7) {
        // Get an orthogonal vector of the given vector
        // in a plane with maximum vector coordinates.
        // Then use it as quaternion axis with pi angle
        if (fabs(oldUp.x()) < 0.6) {
            const double norm = sqrt(1.0 - oldUp.x() * oldUp.x());
            qpassage.set(0.0, oldUp.z()/norm, -oldUp.y()/norm, 0.0);
        } else if (fabs(oldUp.y()) < 0.6) {
            const double norm = sqrt(1.0 - oldUp.y() * oldUp.y());
            qpassage.set(-oldUp.z()/norm, 0.0, oldUp.x()/norm, 0.0);
        } else {
            const double norm = sqrt(1.0 - oldUp.z() * oldUp.z());
            qpassage.set(oldUp.y()/norm, -oldUp.x()/norm, 0.0, 0.0);
        }
    }
    else {
        // Find the shortest angle quaternion that transforms normalized vectors
        // into one other. Then, multiply it to the current quaternion.
        const double s = sqrt(0.5 * dotProdPlus1);
        const double* ptr = (oldUp ^ targetUp / (2.0*s)).ptr();
        qpassage.set(ptr[0], ptr[1], ptr[2], s);
    }

    // forward is modified by this rotation
    newForward = qpassage * newForward;
    newAttitude *= qpassage;
    oldUp = targetUp;
}


