/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef KEEPUPATTITUDECORRECTION_H
#define KEEPUPATTITUDECORRECTION_H

#include <AttitudeCorrection.h>

/**

    When mapping old direction to the new one, the rotation
    around the new direction is made so the up vector
    lies in the plane of the target up vector.

    Alternatively, a function is provided to force the up
    vector to match perfectly the target vector, which is
    not the same, as it modifies the forward vector.

    No costly acos or sin are used, the quaternion is found only
    with fast geometrical properties.
    See for example http://logiciels.cnes.fr/MARMOTTES/marmottes-mathematique.pdf
    for an explanation of the math used.
    
    Care is taken for reconditionning the quaternion
    after too many rotations, when this becomes necessary.
    
    @author Nicolas Brodu
*/
class KeepUpAttitudeCorrection : public AttitudeCorrection
{
public:
    
    /**
        The target Up vector. See the class description.
        @see matchTarget
    */
    osg::Vec3d targetUp;

    ~KeepUpAttitudeCorrection();

    /**
        In addition to the target up vector, the initial up vector must be provided.
        The matchTarget function also needs a current Up vector.
    */    
    KeepUpAttitudeCorrection(const osg::Vec3d &targetUpVector, const osg::Vec3d &initUp);

    /** Avoid building a new object when we can re-use an existing one.
        Equivalent to operator=
        In addition to the target up vector, the initial up vector must be provided.
        The matchTarget function also needs a current Up vector.
    */    
    virtual void set(const osg::Vec3d &targetUpVector, const osg::Vec3d &initUp);

    /// read-only accessor => use matchTarget or set to change
    inline const osg::Vec3d& getUpVector() const {return oldUp;}

    /// See class descrition
    virtual void updateAttitude(osg::Quat &newAttitude, const osg::Quat &oldAttitude, const osg::Vec3d &newForward, const osg::Vec3d &oldForward);

    /** Force the up vector to match perfectly the target vector,
        and modify the forward direction accordingly.
        The mapping from Up to targetUp is done using
        shortest-angle quaternion rotation.
    */
    virtual void matchTarget(osg::Quat &newAttitude, osg::Vec3d &newForward);

protected:
    osg::Quat qpassage;
    osg::Vec3d newUp, oldUp;

};

#endif
