/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <OpenThreads/Mutex>
using namespace OpenThreads;

#include "Localized.h"
#include "Environment.h"
using namespace Environment;

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// Why couldn't they also make hash_map standard from the STL???
// See configure script. Take advantage of SGI extensions in g++ over C++ standard.
#ifdef HAVE_HASH_CONTAINERS
#include <map> // for other maps
#include HASH_MAP_INCLUDE
namespace HASH_NAMESPACE {
template<> struct hash<Localized*> {
    size_t operator()(Localized* const & x) const {
        return reinterpret_cast<size_t>(x);
    }
};
}
typedef HASH_NAMESPACE::hash_map<Localized*, double> NeighborsMap;
#else
#include <map>
typedef std::map<Localized*, double> NeighborsMap;
#endif
using namespace std;


list<Localized*> Localized::allObjects;
lqDB* Localized::localityDatabase = 0;


Localized::Localized( )
{
    lqProxy = 0;
    position.set(0,0,0);
    allObjects.push_back(this);
    neighbors = new std::multimap<const double, Localized*>;
}

Localized::~ Localized( )
{
    neighbors->clear();
    delete neighbors;

    removeFromLocalityDatabase();

    // first remove us for loop below
    allObjects.remove(this);
    
    // remove us also from other Localized objects neighbor list!
    for (list<Localized*>::iterator it = allObjects.begin(); it != allObjects.end(); ++it) {
        for (multimap<const double,Localized*>::iterator itn = (*it)->neighbors->begin(); itn != (*it)->neighbors->end(); ++itn) {
            if (itn->second == this) {
                (*it)->neighbors->erase(itn);
                break;
            }
        }
    }
}

void Localized::destroy( )
{
    // no delay by default
    delete this;
}


const osg::Vec3d& Localized::getPosition() const { return position; }

void Localized::setPosition(const osg::Vec3d& pos) {
    position = pos;
    // Update locality database
    if (localityDatabase && lqProxy) {
        if ((Environment::cycleX!=0.0) && (Environment::cycleY!=0.0))
            lqXYCyclicUpdateForNewLocation (localityDatabase, lqProxy, pos.x(), pos.y(), pos.z());
        else
            lqUpdateForNewLocation (localityDatabase, lqProxy, pos.x(), pos.y(), pos.z());
    }
}

void Localized::setPosition(const double& x, const double& y, const double& z) {
    position.set(x,y,z);
    // Update locality database
    if (localityDatabase && lqProxy) {
        if ((Environment::cycleX!=0.0) && (Environment::cycleY!=0.0))
            lqXYCyclicUpdateForNewLocation (localityDatabase, lqProxy, x, y, z);
        else
            lqUpdateForNewLocation (localityDatabase, lqProxy, x, y, z);
    }
}

Localized::LocalityMode Localized::localityMode = Localized::Brute;

void Localized::setLocalityMode( LocalityMode mode )
{
    localityMode = mode;
}

Localized::LocalityMode Localized::getLocalityMode( )
{
    return localityMode;
}

void Localized::setDefaultLocalityMode( LocalityMode mode )
{
    localityMode = mode;
}

Localized::LocalityMode Localized::getDefaultLocalityMode( )
{
    return localityMode;
}



void Localized::addToLocalityDatabase( )
{
    if (!localityDatabase) return;
    if (lqProxy) return;
    lqProxy = new lqClientProxy();
    lqInitClientProxy(lqProxy, this);
}

void Localized::removeFromLocalityDatabase( )
{
    if (!localityDatabase) return;
    if (!lqProxy) return;
    lqRemoveFromBin(localityDatabase, lqProxy);
    delete lqProxy;
    lqProxy = 0;
}


void Localized::createLocalityDatabase( double originx, double originy, double originz, double sizex, double sizey, double sizez, int divx, int divy, int divz, bool cyclicXY) {

    if (localityDatabase) return;
    
    localityDatabase = lqCreateDatabase (originx, originy, originz, sizex, sizey, sizez, divx, divy, divz);
    if (cyclicXY) {
        Environment::cycleX = sizex;
        Environment::cycleY = sizey;
        Environment::cycleOffsetX = originx;
        Environment::cycleOffsetY = originy;
    }
    
}

extern "C" {

static void lqDestroyHelper (void* clientObject, double distanceSquared, void* clientQueryState) {
    if (((Localized*)clientObject)->lqProxy != 0) {
        delete ((Localized*)clientObject)->lqProxy;
        ((Localized*)clientObject)->lqProxy = 0;
    }
}

}

void Localized::deleteLocalityDatabase() {

    if (localityDatabase) {
        lqMapOverAllObjects(localityDatabase, &lqDestroyHelper, 0);
        lqDeleteDatabase(localityDatabase);
    }
    localityDatabase = 0;
}



struct ClientQueryStateInfo {
    multimap<const double,Localized*>* map;
    Localized* local;
};


/*
function used to map over client objects
static to this compilation unit
*/
extern "C" {

static void findNeighborsCallBackFunction(void* clientObject, double distanceSquared, void* clientQueryState) {

    // don't insert calling object
    if (((ClientQueryStateInfo*)clientQueryState)->local == clientObject) return;
    
    ((ClientQueryStateInfo*)clientQueryState)->map->insert(pair<const double, Localized*>(distanceSquared, (Localized*)(clientObject)));
    
}

}


void Localized::findNeighbors( double radius )
{

    // Prevent other functions from messing with neighbors.
    // See comment in header file. Basically we have no choice but a global lock,
    // since any object can be impacted through neighborhood chains. Besides
    // which, calls are mostly serial, so it doesn't matter if the lock is global.

    if ((localityMode==Brute) || ((localityMode==Bricks) && (localityDatabase==0))) {
        const double r2 = radius * radius;
        // Discard our previous neighbor list to rebuild it up now. Brute is brute...
        neighbors->clear();
        // Too bad => O(n^2) algo
        for (list<Localized*>::iterator it = allObjects.begin(); it != allObjects.end(); ++it) {
            if ((*it) == this) continue;
            const double dist2 = wrappedDist2((*it)->position, position);
            if (dist2<=r2) {
                neighbors->insert(pair<const double, Localized*>(dist2, *it));
            }
        }
        return;
    }

    multimap< const double, Localized* > results;

    // Now, Mixed or NPropagation, or Bricks with DB.
    // Unless NPropagation explicitely stated, locality is needed
    if ((localityMode!=NPropagation) && localityDatabase) {

        ClientQueryStateInfo info;
        info.local = this;
        
        if (localityMode==Bricks) {
            neighbors->clear();
            info.map = neighbors;
        }
        else info.map = &results;
    
        // ideally one would allow concurrent read-only access
        if ((cycleX!=0.0) && (cycleY!=0.0)) lqXYCyclicMapOverAllObjectsInLocality(localityDatabase,
                    position.x(), position.y(), position.z(),
                    radius,
                    &findNeighborsCallBackFunction,
                    &info);
        else lqMapOverAllObjectsInLocality(localityDatabase,
                    position.x(), position.y(), position.z(),
                    radius,
                    &findNeighborsCallBackFunction,
                    &info);

        // That's all for bricks
        if (localityMode==Bricks) return;
    }

    // uniqueness of object, keep dist too
    NeighborsMap uniqueNeighbors;
    double len;
    
    // look for old neighbors still in viciny and their own neighbors new status
    for (multimap<const double,Localized*>::iterator it = neighbors->begin(); it != neighbors->end(); ++it) {
        if (it->second == this) continue; // better be sure than dead-lock sorry
        len = wrappedDist2(it->second->position, position);

        // old neighbor still in viciny
        if (len <= radius) uniqueNeighbors.insert(pair<Localized*,const double>(it->second,len));

        // it may or may still not be, but check its own neighbors too
        // see comment, currently in global lock section so OK
        for (multimap<const double,Localized*>::iterator itn = it->second->neighbors->begin();
        itn != it->second->neighbors->end();
        ++itn) {
            if (itn->second == this) continue; // don't add ourselves
            len = wrappedDist2(itn->second->position, position);
            // new neighbor comes in viciny
            if (len <= radius) uniqueNeighbors.insert(pair<Localized*,const double>(itn->second,len));
        }
    }

    // If mixed, blend in neighbors found from DB, keep uniqueness of results
    if (localityMode==Mixed) {
        for (multimap<const double,Localized*>::iterator it = results.begin(); it != results.end(); ++it) {
            if (it->second == this) continue;
            
            uniqueNeighbors.insert(pair<Localized*,const double>(it->second,it->first));
            
            // but check its own neighbors too because those might not be in DB
            // see comment, currently in global lock section so OK
            for (multimap<const double,Localized*>::iterator itn = it->second->neighbors->begin();
            itn != it->second->neighbors->end();
            ++itn) {
                if (itn->second == this) continue; // don't add ourselves
                len = wrappedDist2(itn->second->position, position);
                // new neighbor comes in viciny
                if (len <= radius) uniqueNeighbors.insert(pair<Localized*,const double>(itn->second,len));
            }
        }
    }

    // Discard our previous neighbor list
    neighbors->clear();

    // Now, build our new neigbor list
    for (NeighborsMap::iterator it = uniqueNeighbors.begin(); it != uniqueNeighbors.end(); ++it) {
        neighbors->insert(pair<const double, Localized*>(it->second,it->first));
    }
}

