/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>

//#define NEURONAL_AGENT_GENERATE_DATASET 1

#ifdef NEURONAL_AGENT_GENERATE_DATASET
#include <fstream>
#include <sstream>
static int datasetNumber = 0;
#endif

#include "Obstacle.h"
#include "NeuronalAgent.h"
#include "Environment.h"
using namespace std;
using namespace Environment;
using namespace OpenThreads;

const int NeuronalAgent::ninput = 10;
const int NeuronalAgent::noutput = 10;
const double NeuronalAgent::noutputInverse = 0.1;

static int displayErrorCountAlone = 0;
static double displayErrorAverageAlone = 0.0;
static int displayErrorCountNeighbor = 0;
static double displayErrorAverageNeighbor = 0.0;

// See comment in .h file
NeuronalAgent::NeuronalAgent( int nhidden, int specie, AggregateStrategy neighborsStrategy, AggregateStrategy forcesStrategy, bool addToGlobalAgentList ) : CrogaiAgent(specie, addToGlobalAgentList), neighborTLP(ninput/*+nmem*/,nhidden,noutput), aloneTLP(ninput/*+nmem*/,nhidden,noutput) {

    aggregateAgents = neighborsStrategy;
    aggregateForces = forcesStrategy;
    
    input = new double[ninput/*+nmem*/]; // TODO
    output = new double[noutput/*+nmem*/];
    outputForces = new osg::Vec3d[noutput/*+nmem*/];
    aggregateOutputForces = new osg::Vec3d[noutput/*+nmem*/];

    if ((aggregateAgents==WinnerTakeAll) || (aggregateForces==WinnerTakeAll))
        outputForcesLenCache = new double[noutput/*+nmem*/];

    monitorError = false;
    mutationRate = mutationJitter = 0.0;

#ifdef NEURONAL_AGENT_GENERATE_DATASET
//    ostringstream os;
//    os << "dataset" << ++datasetNumber << ".txt";
    datasetFile = new ofstream("dataset0.txt");
#error ARGH!
#endif
}

NeuronalAgent::~NeuronalAgent()
{
    delete[] input;
    delete[] output;
    delete[] outputForces;
    delete[] aggregateOutputForces;
    if ((aggregateAgents==WinnerTakeAll) || (aggregateForces==WinnerTakeAll))
        delete[] outputForcesLenCache;
#ifdef NEURONAL_AGENT_GENERATE_DATASET
    datasetFile->close();
#endif
}

void NeuronalAgent::addTeacher( CrogaiAgent * teacher )
{
    teacher->addStudent(this);
    teachers.push_back(teacher);
}

void NeuronalAgent::removeTeacher( CrogaiAgent * teacher )
{
    teacher->removeStudent(this);
    teachers.remove(teacher);
}


// See comment in .h file
void NeuronalAgent::stopTraining() {
    for (list<CrogaiAgent*>::iterator it = teachers.begin(); it!=teachers.end(); ++it) {
        if (*it!=0) (*it)->removeStudent(this);
    }
}


// See comment in .h file
void NeuronalAgent::learnAI(const double elapsedTime, const double currentTime, CrogaiAgent* teacher) {
    // sanity check because I'm paranoid. Should never be null at this point
    if (!teacher) return;
    
    // Can't do both update parameters while learning and use them. TODO
    learnOrUpdateAI(elapsedTime, currentTime, teacher);

    // TODO: remove from teacher if training off/error condition reached
}

// See comment in .h file
void NeuronalAgent::updateAI(const double elapsedTime, const double currentTime) {

    // Can't do both update parameters while learning and use them. TODO
    CrogaiAgent::updateAI(elapsedTime,currentTime);

    learnOrUpdateAI(elapsedTime, currentTime, this);
}

inline void toMaxForce(osg::Vec3d &vec, CrogaiAgent* valueProvider) {
    vec.normalize();
    vec *= valueProvider->getMaxForce();
}

// See comment in .h file
void NeuronalAgent::learnOrUpdateAI( const double elapsedTime, const double currentTime, CrogaiAgent* valueProvider ) {
    
    // Initialize accumulators
    input[0] = 0.0; // number of obstacles
    input[1] = 0.0; // number of friends
    input[2] = 0.0; // number of ennemies

    // get our new neighbors, or cached teacher neighbors (no computations done in this last case)
    // distance does not matter much in case of teacher (cached results are returned, so it is not used)
    // and in our case, it just sets the number of neighbors to apply the network on
    valueProvider->findNeighbors(valueProvider->searchRadius);

    // not the same as multimap size: we see only what's in front
    int numberOfNeighbors = 0;
    
    // first compute the total numbers of each specie
    for (multimap<const double,Localized*>::iterator it = valueProvider->neighbors->begin(); it != valueProvider->neighbors->end(); ++it) {

        // should not happen
        if (it->second==valueProvider) continue;
        
        // see only what's in front of us
        if (wrappedDiffVector(it->second->getPosition(), valueProvider->getPosition()) * valueProvider->forward < 0) continue;
        
        // not the same as multimap size: we see only what's in front
        ++numberOfNeighbors;

        if (dynamic_cast<Obstacle*>(it->second) !=0) {
            input[0] += 1.0; // accumulate obstacle count
            continue; // proceed next
        }
        NeuronalAgent* ea = dynamic_cast<NeuronalAgent*>(it->second);
        if (ea!=0) {
            if (ea->specie == valueProvider->specie)
                input[1] += 1.0; // accumulate friend count
            else
                input[2] += 1.0; // accumulate ennemy count
            continue; // proceed next
        }
    }
        
    // floor normal is considered one of the steering forces for the outputs. Const ref, no copy
    //const osg::Vec3d& floorNormal = valueProvider->getFloorNormal();
    
    // other common inputs
    // altitude
    input[3] = valueProvider->getAltitude();
    // remaining energy of this agent
    input[4] = valueProvider->energy;
    // mass with energy footprint effect of this agent
    input[5] = valueProvider->mass + valueProvider->energy * valueProvider->getEnergyFootprintRatio();
    

    // Handle special case where there is no visible neighbor (some may be behind, though...)
    if (numberOfNeighbors==0) {

        input[6] = 0.0;  // the squared distance to target
        input[7] = 0.0;  // direction of target relatively to this agent
        input[8] = 0.0;  // nothing forward
        input[9] = -2.0; // flag for no neighbors
        //input[10] = 0.0; // no message from nowhere

#ifdef NEURONAL_AGENT_GENERATE_DATASET
        *datasetFile << "INPUT: ";
        for (int i=0; i<ninput; ++i) {
            if (i>0) *datasetFile << " ";
            *datasetFile << input[i];
        }
        *datasetFile << endl;
#endif

        // compute our outputs for these inputs
        aloneTLP.computeOutput(input, output);

        outputForces[0].set(0,0,0); // seek what?
        outputForces[1].set(0,0,0); // flee what?
        outputForces[2].set(0,0,0); // avoid what?
        outputForces[3].set(0,0,0); // pursue what?
        outputForces[4].set(0,0,0); // evade what?
        outputForces[5].set(0,0,0); // no position diff
        outputForces[6].set(0,0,0); // no forward from nowhere
        // if nothing else, we can always wander around...
        outputForces[7] = valueProvider->steerForWander(0.5);
        toMaxForce(outputForces[7], valueProvider);
        // avoid floor, or take off
        outputForces[8] = valueProvider->getFloorNormal() * valueProvider->getMaxForce();
        // our own previous direction
        outputForces[9] = valueProvider->forward * valueProvider->getMaxForce();

#ifdef NEURONAL_AGENT_GENERATE_DATASET
        *datasetFile << "OUTPUT: ";
        for (int i=0; i<ninput; ++i) {
            if (i>0) *datasetFile << " ";
            *datasetFile << outputForces[i];
        }
        *datasetFile << endl;
#endif

        
        int winner = -1; // may be reused between training and final force
        
        // Compute the resulting force
        switch(aggregateForces) {
            case Average:
                steeringForce.set(0,0,0);
                for (int i=0; i<noutput; ++i) steeringForce += outputForces[i] * output[i];
                //SUM steeringForce *= noutputInverse;
                break;
            
            case WinnerTakeAll: {
                // we must find the winner now
                double maxlen2 = -1.0;
                for (int i=0; i<noutput; ++i) {
                    double len2 = (outputForces[i] * output[i]).length2();
                    // we have a new winner
                    if (len2 > maxlen2) {
                        winner = i;
                        maxlen2 = len2;
                    }
                }
                // one extra computation, strictly speacking
                steeringForce = outputForces[winner] * output[winner];
                break;
            }
        }
        
        if (false) {//valueProvider != this) {
            
            // train on the clamped version
            clampToMaxForce();
            
            double Q = steeringForce.length2();
            double R;
            if (Q>1e-7) {
                Q = 1.0 / Q;
                R = valueProvider->steeringForce.length2();
                if (R>1e-7) R = 1.0 / R;
                else valueProvider = this; // Cancel training in case of singularity
            }
            else valueProvider = this; // Cancel training in case of singularity
            
            if (valueProvider != this) {
    
                const osg::Vec3d g = steeringForce * Q;
                const osg::Vec3d u = valueProvider->steeringForce * sqrt(R*Q);
                const osg::Vec3d Delta = (steeringForce - valueProvider->steeringForce) * R;
                osg::Vec3d h = steeringForce * (u*g) + Delta - u;

// test with no direction part
//h = Delta;
                
                double* gradout = new double[noutput/*+nmem*/];
                            
                // learn from teacher
    
                // compute output gradients. Error function is
                //     TODO: change this E = 1 - resultingForce * targetForce
                // with "*" = scalar product
                // resulting force depends on the aggregation mode, see below
                // targetForce is the teacher steering force set by its own AI
                
                switch(aggregateForces) {
    
                    case Average:
                        for (int i=0; i<noutput; ++i)
                            //SUM gradout[i] = noutputInverse * (outputForces[i] * h);
                            gradout[i] = outputForces[i] * h;
                        break;
                    
                    case WinnerTakeAll: {
                        for (int i=0; i<noutput; ++i) gradout[i] = 0.0;
                        // all others remain 0
                        gradout[winner] = outputForces[winner] * h;
                        break;
                    }
                }
                
                // backpropagation of error gradient
                aloneTLP.backPropagate(input, output, gradout);
    
                // learn
                aloneTLP.learn();
        
                delete[] gradout; // cleanup
                
                // compute error
                if (monitorError) {
                    displayError(1.5 + 0.5 * R / Q - steeringForce * (u + valueProvider->steeringForce * R), true);

// test with no direction part
//displayError(0.5 * R * (steeringForce - valueProvider->steeringForce).length2(), true );
                }
            
            }        
        }
        
        // Done!
        return;
    }


    // Now handle case with more than one neighbor
    double* trainInputs = 0;
    double* trainHiddenValues = 0;
    double* trainOutputs = 0;
    osg::Vec3d *trainOutputForces = 0;
    int* winnerAgent = 0;
    double* gradout = 0;
    
    // when training, it is necessary to keep the mappings
    if (valueProvider != this) {
        trainInputs = new double[numberOfNeighbors*ninput];
        trainHiddenValues = new double[numberOfNeighbors*neighborTLP.getNHidden()];
        trainOutputs = new double[numberOfNeighbors*noutput];
        trainOutputForces = new osg::Vec3d[numberOfNeighbors*noutput];
        if (aggregateAgents==WinnerTakeAll) winnerAgent = new int[noutput];
        gradout = new double[noutput/*+nmem*/];
    }
    
    // initialize agents aggregate
    for (int i=0; i<noutput; ++i) aggregateOutputForces[i].set(0.0, 0.0, 0.0);
    if (aggregateAgents==WinnerTakeAll)
        for (int i=0; i<noutput; ++i) outputForcesLenCache[i] = -1.0; // force first comparison to win

    int neighborNumber = 0;
    
    // now proceed on a per-neighbor basis
    for (multimap<const double,Localized*>::iterator it = valueProvider->neighbors->begin(); it != valueProvider->neighbors->end(); ++it) {
        if (it->second==valueProvider) continue; // should not happen
        
        // See only what's in front of us
        // Store this result as input: indicates the direction of target relatively to this agent
        const osg::Vec3d diffVec = wrappedDiffVector(it->second->getPosition(), valueProvider->getPosition());
        input[7] = diffVec * valueProvider->forward;
        if (input[7] < 0) continue;
        
        // the squared distance to target
        input[6] = it->first;
        
        // cache results of dynamic casting
        Obstacle* obstacle = dynamic_cast<Obstacle*>(it->second);
        Agent* agent = 0;
        
        if (obstacle !=0) {
            input[8] = -1.0; // Always in front of obstacles
            input[9] = 0.0; // flag for obstacles
            //input[10] = 0.0; // no message from obstacles
        } else {
            NeuronalAgent* ea = dynamic_cast<NeuronalAgent*>(it->second);
            if (ea!=0) {
                input[8] = ea->forward * valueProvider->forward; // cos of forward directions, if we are behind or in front of the other
                input[9] = ea->specie;            // friend or foo
                //input[10] = ea->message;          // the other agent message to the world. Let's have communication!
                // note: message is heard by friends and foos. Should we have a by-specie message?
                // note2: message could possibly be an IA mode, like a group of preys
                //        fleeing without waiting to see the predator.
                //        It's up to the net & GA to decide what to do of it, however.
                agent = ea;
            }
            else {
                agent = dynamic_cast<Agent*>(it->second);
                if (agent!=0) input[8] = agent->forward * valueProvider->forward; // this notion still applies for generic agents
                else input[8] = -1.0;                      // as for obstacles, always face danger!
                input[9] = -1.0;                           // flag for unknown beast
                //input[10] = 0.0;                           // no message from generic agents
            }
        }

#ifdef NEURONAL_AGENT_GENERATE_DATASET
        *datasetFile << "INPUT: ";
        for (int i=0; i<ninput; ++i) {
            if (i>0) *datasetFile << " ";
            *datasetFile << input[i];
        }
        *datasetFile << endl;
#endif

        // compute our outputs for these inputs
        neighborTLP.computeOutput(input, output);

        // apparent position of the target, counting world wrap...
        const osg::Vec3d apparentPos = valueProvider->getPosition() + diffVec;
        
        // compute output forces
        outputForces[0] = valueProvider->steerForSeek(apparentPos);
        toMaxForce(outputForces[0], valueProvider);
        outputForces[1] = valueProvider->steerForFlee(apparentPos);
        toMaxForce(outputForces[1], valueProvider);
        if (obstacle != 0) {
            outputForces[2] = valueProvider->steerToAvoid(obstacle->getNode(), valueProvider->searchRadius);
            toMaxForce(outputForces[2], valueProvider);
            // pursuing an unmoving obstacle is considered same as seeking
            outputForces[3] = outputForces[0];
            // evading an unmoving obstacle is considered same as fleeing
            outputForces[4] = outputForces[1];
        
            // target forward: obstacle always head front, see input
            outputForces[6] = -valueProvider->forward * valueProvider->getMaxForce();
        }
        else if (agent!=0) {
            outputForces[2] = valueProvider->steerToAvoid(agent, valueProvider->searchRadius);
            toMaxForce(outputForces[2], valueProvider);
            outputForces[3] = valueProvider->steerForPursuit(agent);
            toMaxForce(outputForces[3], valueProvider);
            outputForces[4] = valueProvider->steerForEvasion(agent);
            toMaxForce(outputForces[4], valueProvider);
            
            // target forward
            outputForces[6] = agent->forward * valueProvider->getMaxForce();
        }
        // unknown beast?
        else {
            // consider avoiding a beast is same as fleeing
            outputForces[2] = outputForces[1];
            // consider pursuing a beast is same as seeking
            outputForces[3] = outputForces[0];
            // consider evading a beast is same as fleeing
            outputForces[4] = outputForces[1];
            
            // target forward: see input for unknown beast
            outputForces[6] = -valueProvider->forward * valueProvider->getMaxForce();
        }

        // other parameters linked to the target
        outputForces[5] = diffVec;
        toMaxForce(outputForces[5], valueProvider);

        // 6 done above
        
        // if nothing else, we can always wander around...
        // Note: this value is cached for each agent during one iteration step, so we get
        // the same value the teacher used, with its own wander parameter
        outputForces[7] = valueProvider->steerForWander(0.5);
        toMaxForce(outputForces[7], valueProvider);

        // avoid floor, or take off
        outputForces[8] = valueProvider->getFloorNormal() * valueProvider->getMaxForce();

        // our own previous direction
        outputForces[9] = valueProvider->forward * valueProvider->getMaxForce();

        // save data for later training
        if (valueProvider != this) {
            int offsetInput = neighborNumber * ninput;
            int offsetOutput = neighborNumber * noutput;
            for (int i=0; i<ninput; ++i) trainInputs[offsetInput+i] = input[i];
            neighborTLP.getHidden(trainHiddenValues + neighborNumber * neighborTLP.getNHidden());
            for (int i=0; i<noutput; ++i) {
                trainOutputs[offsetOutput+i] = output[i];
                trainOutputForces[offsetOutput+i] = outputForces[i];
            }
        }

#ifdef NEURONAL_AGENT_GENERATE_DATASET
        *datasetFile << "OUTPUT: ";
        for (int i=0; i<ninput; ++i) {
            if (i>0) *datasetFile << " ";
            *datasetFile << "(" << outputForces[i] << ")";
        }
        *datasetFile << endl;
#endif

        // Make use of net outputs to weight the output forces AFTER saving
        // the raw vectors for gradient computations, if training
        for (int i=0; i<noutput; ++i) outputForces[i] *= output[i];

        // two strategies:
        // - sum coefs for each agent to get final coef for this steering force
        // - winner take all: get strongest force vector as the only one for this steering force
        switch(aggregateAgents) {

            case Average:
                for (int i=0; i<noutput; ++i)
                    aggregateOutputForces[i] += outputForces[i];
                break;
            
            case WinnerTakeAll:
                for (int i=0; i<noutput; ++i) {
                    double len2 = outputForces[i].length2();
                    // we have a new winner
                    if (len2 > outputForcesLenCache[i]) {
                        aggregateOutputForces[i] = outputForces[i];
                        outputForcesLenCache[i] = len2;
                        winnerAgent[i] = neighborNumber;
                    }
                }
                break;
        }

        
        ++neighborNumber;
    }

    // End of per-neighbor computations

    // We now have either a bunch of behavior forces (seek, flee...) averaged on all agents,
    // OR we have the strongest vector in each of the behaviors
    int winnerForce = 0;

    // Apply second aggregate rule, on the forces this time
    switch(aggregateForces) {
        case Average: {
            steeringForce.set(0,0,0);
            for (int i=0; i<noutput; ++i) steeringForce += aggregateOutputForces[i];
            /*SUM if (aggregateAgents==Average)
                steeringForce *= noutputInverse / (double)numberOfNeighbors;
            else if (aggregateAgents==WinnerTakeAll)
                steeringForce *= noutputInverse;
            */
            break;
        }
        
        case WinnerTakeAll: {
            // if len cache was computed previously, use it, otherwise compute it now
            if (aggregateAgents==Average) {
                for (int i=0; i<noutput; ++i)
                    outputForcesLenCache[i] = aggregateOutputForces[i].length2();
            }
            // now use the len cache for comparisons. We have computed len2 no more than necessary
            // Look for the maximum strength
            double maxlen2 = outputForcesLenCache[0];
            for (int i=1; i<noutput; ++i) {
                if (outputForcesLenCache[i] > maxlen2) {
                    winnerForce = i;
                    maxlen2 = outputForcesLenCache[i];
                }
            }
            steeringForce = aggregateOutputForces[winnerForce];
            break;
        }
    }

    // Training phase
    // Apply batch backpropagation on all neighbor mappings
    if (valueProvider != this) {
        
        // train on the clamped version
        clampToMaxForce();
        
        int offsetInput = 0;
        int offsetHidden = 0;
        int offsetOutput = 0;

        // See report about derivation of this result, I won't put it all here
        // In short, the error function is
        //      E = 1 - normalized(F)*normalized(T) + 0.5*norm(F-T)^2/norm(T)^2
        // With F the final output force from the network and T the training target force.
        // This error is null when T and F are equal, and comprise both a directional and a length component otherwise
        //
        // The goal is now to find the error derivative with respect to each network output, then backpropagate it.
        // The final formula is somewhat convoluted because of the division by norm(F) in the error.
        // See the report for the full derivation, here is the error derivative with respect to each of the outputs :
        //   d(E)/d(o_j_i) = a_j_i * b_j * { u dot [ Sum_d(S_j_i_d*F_d) * g - S_j_i ] + Sum_d(Delta_d*S_j_i_d) }
        //   TODO: Write simplified h formula!
        // o_j_i is each network output, with j the output number and i the agent number.
        // a_j_i depends on the agent aggregate policy: 1/n for Average Mode, or 1 for the winning agent, 0 for others
        // b_j   depends on the force aggregate policy: 1/noutput for Average, or 1 for the winning force, 0 for others
        // S_j_i is the steering force associated with output j,i.
        // u is sqrt(Q*R)*T with T target training vector, Q = 1/squared norm F, R = 1/squared norm T
        // g is Q*F, F network total resulting force
        // Delta is R*(F-T)
        // _d represent dimension component of a vector, that is x, y or z

        // Pre-compute all the factors not depending on i,j only once.
        double Q = steeringForce.length2();
        double R;
        if (Q>1e-7) {
            Q = 1.0 / Q;
            R = valueProvider->steeringForce.length2();
            if (R>1e-7) R = 1.0 / R;
            else valueProvider = this; // Cancel training in case of singularity
        }
        else valueProvider = this; // Cancel training in case of singularity
        
        if (valueProvider != this) {

            const osg::Vec3d g = steeringForce * Q;
            const osg::Vec3d u = valueProvider->steeringForce * sqrt(R*Q);
            const osg::Vec3d Delta = (steeringForce - valueProvider->steeringForce) * R;
            osg::Vec3d h = steeringForce * (u*g) + Delta - u;

// test with no direction part
//h = Delta;
            
            // Also compute a_j_i * b_j for either the winner or for all i,j
            double averageFactor = 1.0;
            switch(aggregateAgents) {
                case Average: switch(aggregateForces) {
                    case Average:
                        //SUM averageFactor = noutputInverse / (double)numberOfNeighbors;
                        break;
                    case WinnerTakeAll:
                        //SUM averageFactor = 1.0 / (double)numberOfNeighbors;
                        break;
                }
                break;
                case WinnerTakeAll: switch(aggregateForces) {
                    case Average:
                        //SUM averageFactor = noutputInverse;
                        break;
                    
                    // no need to do the full loop for just one non-null gradient
                    case WinnerTakeAll:
                        // train only with the winner
                        offsetInput = winnerAgent[winnerForce] * ninput;
                        offsetHidden = winnerAgent[winnerForce] * neighborTLP.getNHidden();
                        offsetOutput = winnerAgent[winnerForce] * noutput;

                        // initialize all gradients to 0
                        for (int j=0; j<noutput; ++j) gradout[j] = 0.0;

                        // const ref to avoid triple indirection
                        //const osg::Vec3d& s = trainOutputForces[offsetOutput+winnerForce];

                        // Winner takes really everything, averagefactor is 1.0
                        gradout[winnerForce] = trainOutputForces[offsetOutput+winnerForce] * h;

                        //-trainOutputForces[offsetOutput+winnerForce] * valueProvider->steeringForce;
                        
                        // first restore hidden values for backprop
                        neighborTLP.setHidden(trainHiddenValues + offsetHidden);
                        neighborTLP.backPropagate(trainInputs+offsetInput, trainOutputs+offsetOutput, gradout);

                        numberOfNeighbors = 0; // useless now, use it as flag
                        break;
                }
                break;
            }
    
            // loop not entered if win&win case
            for (int i=0; i<numberOfNeighbors; ++i) {
    
                // First build up error gradient on output vectors
                switch(aggregateAgents) {
                    case Average: switch(aggregateForces) {
                        case Average:
                            for (int j=0; j<noutput; ++j) {
                                gradout[j] = averageFactor * (trainOutputForces[offsetOutput+j] * h);
                            }
                            break;
                        case WinnerTakeAll:
                            for (int j=0; j<noutput; ++j) gradout[j] = 0.0;
                            gradout[winnerForce] = averageFactor * (trainOutputForces[offsetOutput+winnerForce] * h);
                            break;
                    }
                    break;
                    case WinnerTakeAll:
                        // Necessarily aggregateForces = Average at this point, see above
                        for (int j=0; j<noutput; ++j) {
                            // One agent may win several forces
                            if (winnerAgent[j]==i)
                                gradout[j] = averageFactor * (trainOutputForces[offsetOutput+j] * h);
                            else
                                gradout[j] = 0.0;
                        }
                    break;
                }
                
                // now batch train the network
                // first restore hidden values for backprop
                neighborTLP.setHidden(trainHiddenValues + offsetHidden);
                
                if (i==0) neighborTLP.backPropagate(trainInputs+offsetInput, trainOutputs+offsetOutput, gradout);
                else neighborTLP.batchBackPropagateAccumulate(trainInputs+offsetInput, trainOutputs+offsetOutput, gradout);
            
                offsetInput += ninput;
                offsetHidden += neighborTLP.getNHidden();
                offsetOutput += noutput;
    
            }
    
            // no need to divide by 1
            if (numberOfNeighbors>1) neighborTLP.batchBackPropagateTerminate(numberOfNeighbors);

                
            // learn
            neighborTLP.learn();
            
            // compute error
            if (monitorError) {
//OLD                displayError(1.0 - steeringForce * valueProvider->steeringForce);
                displayError(1.5 + 0.5 * R / Q - steeringForce * (u + valueProvider->steeringForce * R), false);

// test with no direction part
//displayError(0.5 * R * (steeringForce - valueProvider->steeringForce).length2(), false );

                if (displayErrorCountNeighbor%1000 == 0) cout << steeringForce << " target was : " << valueProvider->steeringForce << endl;
            }
        }

        // cleanup    
        delete[] trainInputs;
        delete[] trainHiddenValues;
        delete[] trainOutputs;
        delete[] trainOutputForces;
        if (aggregateAgents==WinnerTakeAll) delete[] winnerAgent;
        delete[] gradout;
        
    } // end of training
    

/*        // Memory: after training, use a few extra neurons (set to 0 during training) and
        // keep them between 2 AI steps
        // Thus, position memory inputs for next step.
        for (int i=0; i<nmem; ++i) {
            input[ninput+i] = output[noutput+i];
        }

        // Message?
*/
}


void NeuronalAgent::displayError( double errorValue , bool alone)
{

    int& displayErrorCount = (alone) ? displayErrorCountAlone : displayErrorCountNeighbor;
    double& displayErrorAverage = (alone) ? displayErrorAverageAlone : displayErrorAverageNeighbor;
    TwoLayerPerceptron& tlp = (alone) ? aloneTLP : neighborTLP;

    ++displayErrorCount;
    displayErrorAverage+=errorValue;
    if ((displayErrorCount%1000) == 0) {

        cout << "Average " << ((alone) ? "alone network" : "neighbor network") << " training error on last " <<
        displayErrorCount << " steps is " <<
        displayErrorAverage/ displayErrorCount<<
        " for agent=" << hex << (void*)this << dec <<
        ", specie=" << specie <<
        ", nhidden=" << tlp.getNHidden() <<
        endl;

        displayErrorCount = 0;
        displayErrorAverage = 0.0;
    }
}

void NeuronalAgent::setMutationParameters( double mutationRate, double mutationJitter )
{
    this->mutationRate = mutationRate;
    this->mutationJitter = mutationJitter;
}


void NeuronalAgent::updateChild( CrogaiAgent * child )
{
    CrogaiAgent::updateChild(child); // Does nothing, just in case it'd change in the future
    
    NeuronalAgent* na = dynamic_cast<NeuronalAgent*>(child);
    if (!na) return;
    
    na->neighborTLP = neighborTLP;
    na->aloneTLP = aloneTLP;
    na->monitorError = monitorError;
    na->mutationRate = mutationRate;
    na->mutationJitter = mutationJitter;
    
    // no mutation => done
    if ((mutationRate==0.0) && (mutationJitter==0.0)) return;

    // now introduce mutations
    na->aloneTLP.mutate(mutationRate,mutationJitter,mutationRate,mutationJitter,mutationRate,mutationJitter,mutationRate,mutationJitter);
    na->neighborTLP.mutate(mutationRate,mutationJitter,mutationRate,mutationJitter,mutationRate,mutationJitter,mutationRate,mutationJitter);
}
