/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef NEURONALAGENT_H
#define NEURONALAGENT_H

#include <OpenThreads/Mutex>

#include "CrogaiAgent.h"
#include "TwoLayerPerceptron.h"

/**
    An agent using a two layer perceptron neural network for its AI decisions.

    First, train it to mimick another agent with an explicit AI. This will give
    a non-random starting point for the later evolution, and at least gives this
    agent a chance to survive.
    Then, use a genetic algorithm to evolve of population of neuronal agents into
    something else.

    If you're really lucky, the offsprings may be more clever than the original AI,
    and escape into the Internet to take over the world.
    Please mail me if this happens :)


    For each neighbor agent, the network is run with the following inputs. Parameters
    marked 'c' are common to all neighbors computations, but provided nonetheless on a
    per-neighbor basis. Parameters marked 'n' are neighbor-dependent, and set to
    respectively 0,0,0,-2 when there is no neighbor. Parameters marked 'e' are only used
    during the evolution phase, and not the training phase (TODO: work in progress). They represent
    the acquired knownledge part of the agent, and are reset when creating a new offspring.
    - c: number of obstacles
    - c: number of friends
    - c: number of ennemies
    - c: altitude
    - c: remaining energy
    - c: mass taking in account energy storage
    - n: squared distance to target
    - n: target direction relatively to forward direction. Is it on the side?
    - n: target forward direction relatively to this agent forwarddirection: are we in front or behind the target?
    - n: specie of the target. -2 is a flag to indicate no neighbor instead of a specie, -1 is reserved for unknown objects, 0 for obstacles.
    // TODO: work in progress
    - en: message of the target to the world.
    - ec: current message of this agent to the world.
    - ec: memory (private message of this agent to itself) from previous AI update, index 0.
    - ec: ...
    - ec: memory from previous AI update, index number of memory - 1

    The network outputs 'n' and 'c' are all coefficients to apply to a vector. See Agent.h for
    steering vector definitions. All neighbor-dependent vector are set to 0 when there is no neighbor.
    The network outputs 'e' are scalar values only used after training (not part of the error function)
    - n: seek target position
    - n: flee target position
    - n: avoid collision with target
    - n: pursue target
    - n: evade target
    - n: position difference vector
    - n: target forward direction
    - c: wander direction (constrained random direction, see Agent.h)
    - c: floor normal at this point
    - c: our own previous forward direction
    // TODO: work in progress
    - ec: new message of this agent to the rest of the world
    - ec: new memory (private message of this agent to itself) for next AI update, index 0
    - ec: ...
    - ec: new memory for next AI update, index number of memory - 1
    
    The neural network is run for each neigbhor, then the computations are aggregated
    according to the following policies. For each result vector:
    - Average: average vectors from all neighbor computation contributions to get final steering vector
    - Winner take all: get largest norm output*force vector as the only one for this steering force

    Once a steering force is computed for each output, the forces are then once again
    aggregated according to one of the following policies for the final steering force:
    - Average: average contributions from all individual steering results, computed as defined above
    - Winner take all: get largest norm individual steering force, computed as defined above, as the only one

    Average should in principle give smoother results and allows training all network weights.
    Winner take all is useful for more natural results (ex: avoid the closest), to decide only
    one clear action in a case where multiple choices could be made, instead of an average and
    not so good result. Unfortunately as gradients are 0 for unused weights in this case, the
    training phase may not go so well.
    Averaging both neighbor computations and result vectors is thus the default, but you're encouraged
    to play with other policies.

    During training phase, the Error function is computed as follow:
        E = 1 - (ResultForce dot TeacherResultForce)
    where ResultForce is the output of all previous computations, and TeacherResultForce is
    what the teacher would have done in the same situation.

    Note: The training phase is stateless, without memory, so the neural network can be registered
          on many teachers at the same time to get data from many different situations
    
@author Nicolas Brodu
*/
class NeuronalAgent : public CrogaiAgent
{
public:
            
    /** Strategies to aggregate the steering force vectors from each neighor computations:
        + For each steering force resulting from a steerForXXX behavior:
          - Average: average vectors from all neighbor computation contributions to get final steering vector
          - Winner take all: get largest norm output*force vector as the only one for this steering force
        + For the final steering force
          - Average: average contributions from all individual steering results, computed as defined above
          - Winner take all: get largest norm individual steering force, computed as defined above, as the only one
    */
    enum AggregateStrategy {Average, WinnerTakeAll, User}; // for subclasses

    /**
        @param nhidden Number of hidden units for the two layer perceptron
        @param specie Specie, or team, this agent belongs to. It's up to you
            to set one specie for all agents with the same number of hidden
            units, or to do other weird mixes.
        @param teacher Give this agent a teacher to show it what to do.
            You can explicitely set the teacher to 0 to de-activate
            the training phase and start with random weights if
            you really want so, but I would advice against it if you want the
            agents to have a reasonable chance to survive. Also the reason why 0 is not
            a default parameter. Use setTeacher(...) later on to deactivate/reactivate
            training;
        @param neighborsStrategy How to aggregate the results of computations done for each neighbor.
            Default is average, but this may not be the most useful for collision avoidance. See below.
        @param forcesStrategy How to aggregate the results of computations done for each steering force.
            Default is average, but this may not be the most useful for collision avoidance. See below.
        @param addToGlobalAgentList Whether or not this agent should be taken into account for playing the
            game. Default is false, contrary to parent default, because the network needs training first.
    */
    NeuronalAgent(int nhidden, int specie, AggregateStrategy neighborsStrategy = Average, AggregateStrategy forcesStrategy = Average, bool addToGlobalAgentList = false);

    virtual void addTeacher(CrogaiAgent* teacher);
    virtual void removeTeacher(CrogaiAgent* teacher);
    
    /** stops training and remove us from all teacher's students lists
        As soon as the training stops, this agent is added to the global agent list and may play the game!
        You can safely position the agent and prepare its avatar before calling this function, it won't
        interfere with the training or other agents so long as this function is not called.
    */
    virtual void stopTraining();

    /** See comment above about policy:
        - aggregateAgents: For getting each individual steering force from neighbor computation contributions
        - aggregateForce: For getting the final steering force from each individual steering force.
    */
    AggregateStrategy aggregateAgents, aggregateForces;//, aggregateMemory;


    /** number of inputs and outputs are common to all agents.
        They are respectively the number of parameters for the AI algorithm and the number
        of steering forces to get a coefficient for.
    */
    static const int ninput;
    static const int noutput;

    /// Our brain. Not very much, but more than most living things in the real world out there!
    TwoLayerPerceptron neighborTLP, aloneTLP;

    /** During training, error can optionally be computed and displayed.
        This class display it on cout, but subclasses may change this behavior.
        See the displayError function.
        monitorError is false by default
    */
    bool monitorError;
        
    /** Specifies the parameters for the spawning "updateChild" function.
        These parameters are inherited by the child, so setting them once for the first generation
        is enough.
    
        @param mutationRate The network weights are added a random number between -mutationRate*currentValue
            and +mutationRate*currentValue. 0 thus makes a perfect copy, which may be useful
            at the beginning to build a large population from only a few teachers. Default is 0.
        @param mutationJitter The network weights are also added a random number between
            -jitter value & +jitter value, then normalized by the network layer dimensions.
            This allows to give a chance to null weights, which would otherwise not be 
            affected by the mutation rate. Default is 0.
    */
    virtual void setMutationParameters(double mutationRate = 0.0, double mutationJitter = 0.0);

protected:
    
    /** Subclasses may call this function when spawning to set base class parameters

        Each agent has an initial genetic payload consisting of neural network weights,
        which are NOT its current weights. Indeed, each agent can use its memory and
        message communication skills during its life, and can also learn and train its
        network during its life.

        A new child inherits only the initial weights as genetic information, and does not
        inherit the parent learnt changes. At the very begining of your program, the first
        parent may use a teacher to get a reasonable AI and have a chance to survive.
        As time passes, the new children can evolve these initial weights using genetic
        algorithm and the hope is that they get better than the initial parent.

        Error monitoring, aggregate policies, and spawn parameters are transfered to
        the child unchanged.
    */
    virtual void updateChild(CrogaiAgent* child);


    /// Dislay the argument as the current training error for this network
    /// Default is to display it on cout, but subclasses may change this behavior.
    virtual void displayError(double errorValue, bool alone);

    // parent API. Implement the neural network AI
    virtual void updateAI(const double elapsedTime, const double currentTime);

    // Learn a teacher AI
    virtual void learnAI(const double elapsedTime, const double currentTime, CrogaiAgent* teacher);

    // Our own function called in both case with different parameters, to avoid code duplication
    virtual void learnOrUpdateAI(const double elapsedTime, const double currentTime, CrogaiAgent* valueProvider);
    
    /// avoid creating input and output parameters each time
    double *input;
    double *output, *gradout;
    osg::Vec3d *outputForces, *aggregateOutputForces;

    // parameters specific to winner take all
    double *outputForcesLenCache;

    // speed up computations, don't divide each time!
    static const double noutputInverse;
    //double nmem;
    
    /// Our teachers. Unused after the training phase
    std::list<CrogaiAgent*> teachers;
    OpenThreads::Mutex teacherAccessMutex;

    /// @see setSpawnParameters
    double mutationRate, mutationJitter;

    virtual ~NeuronalAgent();

#ifdef NEURONAL_AGENT_GENERATE_DATASET
    std::ofstream *datasetFile;
#endif
};

#endif
