/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>

#include "ShortestAngleAttitudeCorrection.h"

ShortestAngleAttitudeCorrection::ShortestAngleAttitudeCorrection()
 : AttitudeCorrection()
{
}


ShortestAngleAttitudeCorrection::~ShortestAngleAttitudeCorrection()
{
}

void ShortestAngleAttitudeCorrection::updateAttitude( osg::Quat & newAttitude, const osg::Quat & oldAttitude, const osg::Vec3d & newForward, const osg::Vec3d & oldForward )
{
    // Compute new attitude: find quaternion that convert initial velocity
    // to new velocity vectors, and multiply current attitude by it.
    // No costly acos or sin are used, the quaternion is found only
    // with fast geometrical properties.
    // See for example http://logiciels.cnes.fr/MARMOTTES/marmottes-mathematique.pdf
    // for an explanation of the math used.

    double dotProdPlus1 = 1.0 + oldForward * newForward;
    // Check for degenerate case of full u-turn. Use epsilon for detection
    if (dotProdPlus1 < 1e-7) {
        // Get an orthogonal vector of the given vector
        // in a plane with maximum vector coordinates.
        // Then use it as quaternion axis with pi angle
        if (fabs(oldForward.x()) < 0.6) {
            const double norm = sqrt(1.0 - oldForward.x() * oldForward.x());
            newAttitude = oldAttitude * osg::Quat(0.0, oldForward.z()/norm, -oldForward.y()/norm, 0.0);
        } else if (fabs(oldForward.y()) < 0.6) {
            const double norm = sqrt(1.0 - oldForward.y() * oldForward.y());
            newAttitude = oldAttitude * osg::Quat(-oldForward.z()/norm, 0.0, oldForward.x()/norm, 0.0);
        } else {
            const double norm = sqrt(1.0 - oldForward.z() * oldForward.z());
            newAttitude = oldAttitude * osg::Quat(oldForward.y()/norm, -oldForward.x()/norm, 0.0, 0.0);
        }
    }
    else {
        // Find the shortest angle quaternion that transforms normalized vectors
        // into one other. Then, multiply it to the current quaternion.
        const double s = sqrt(0.5 * dotProdPlus1);
        const double* ptr = (oldForward ^ newForward / (2.0*s)).ptr();
        newAttitude = oldAttitude * osg::Quat(ptr[0], ptr[1], ptr[2], s);
    }

    // Renormalize the quaternion if it is too badly conditionned after many multiplications
    // reuse variable, at the expense of clarity.
    dotProdPlus1 = newAttitude.length2();

    // Use less restrictive conditions
    if ((dotProdPlus1<0.9999) || (dotProdPlus1>1.0001)) {

        // renormalize
        newAttitude /= sqrt(dotProdPlus1);

        // Sanity check: after that many rotations, ensure forward direction is still OK
        // => local space forward and attitude tranformed world space forward should match
        //if (newForward * (newAttitude * osg::Vec3d(0,1,0)) < 1.0 - 1e-7) {
        if (newForward * osg::Vec3d(-newAttitude.z(),newAttitude.w(),newAttitude.x()) < 1.0 - 1e-7) {
            
            // avoid infinite recursion? This shouldn't happen with double precision
            if (oldAttitude.w()!=1) {
                
                // Problem: we don't have an up vector to restore the current rotation around
                // forward vector (composition of two shortest angle rotations is not the 
                // shortest angle of the result!)
                // keep the current world space up vector as target
                osg::Vec3d oldUp(0,0,1); // in local space
                osg::Vec3d targetUp = newAttitude * oldUp; // world space

                // compute shortest angle rotation from local space to current forward
                // => restore match between local space forward and attitude tranformed world space forward
                //    that was badly conditionned
                updateAttitude(newAttitude, osg::Quat(0,0,0,1), newForward, osg::Vec3d(0,1,0));

                // what we seek: new world space up in the plane of target up
                osg::Vec3d newUp = targetUp - newForward * (targetUp * newForward);
                newUp.normalize();

                // put oldUp in world space with new attitude
                oldUp = newAttitude * oldUp; 
            
                // rotate again, around current forward axis this time
                dotProdPlus1 = 1.0 + oldUp * newUp;
                
                // Check for degenerate case of full u-turn. Use epsilon for detection
                if (dotProdPlus1 < 1e-7) {
                    newAttitude *= osg::Quat(newForward.x(), newForward.y(), newForward.z(), 0);
                }
                else {
                    // Find the shortest angle quaternion that transforms normalized vectors 
                    // into one other. Then, multiply it to the current quaternion.
                    const double s = sqrt(0.5 * dotProdPlus1);
                    const double* ptr = (oldUp ^ newUp / (2.0*s)).ptr();
                    newAttitude *= osg::Quat(ptr[0], ptr[1], ptr[2], s);
                }
            
                // new forward unchanged by second rotation

            }
        }
    }


}


