/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef SPORECLOUD_H
#define SPORECLOUD_H

#include <list>
#include <map>

/** Add a cloud of spores in the environment for the agents genetic algorithm

    We globally have these 2 options for th G.A.:
    - Monoparental reproduction & mutation: like bacteria division
    - Crossover & mutation: but the only way is to make agents meet each other.
    
    The idea is that agents don't have yet a mating AI which alows them to reproduce
    by choosing their partners on self-organized criteria (like for example a pretty
    blue plume on the head).

    Thus we're left with agents meeting each other on an explicit criterion, or chance.
    The first poses the problem of defining this criterion, and the second is not
    possible on collision only : A GA relying on collisions to reproduce would possibly
    end up favoring AI that collide... which we don't want.

    So, a way to maximize agents interactions without relying on collision or other explicit
    rendez-vous points is to do like plants, fish, mushrooms... This leads to the concept
    of a spore cloud (thanks goes to Nawwaf Kharma for pointing this idea out).

    I simulate here the spore cloud by a volumic concentration leading to probabilities of fecundation.
    As for the locality database, the world is divided into elementary volumes. Then when an agent wishes
    to get fecunded it simply gets spores from the environment:
    - No spore in volume => probability 0 of being fecunded
    - Volume saturated => probability 1 of being fecunded
    Actually, this probability is split proportionally to the content of the cloud. Suppose spores from
    individuals 1,2,3 are present in proportions 10%, 30%, 60% respectively, and the volume is 60%
    saturated. Then probabilities of being fecunded by individual 1,2,3 are respectively 0.06, 0.18, 0.36.
    
    The volumic concentrations are updated by diffusion equations, using the simulator once more. Note
    that the integration update frequency of the cloud spore diffusion may be different than the frequency
    of the agents physics.

    I also add a decay time after which the spores are non-functional and thus eliminated. A future extension
    may be to gradually reduce the spores efficiency by playing on probabilities, instead of this clear cut.
    Think for example of radioactive half-life time.
    

    @note As a special case of thread safety, the step and addSpore method should never be called simultaneously
          This is guaranteed if the only access through these methods are done via simulator steps, for example.
    
    @author Nicolas Brodu
    
*/
class SporeCloud {
protected:
    // forward decl
    struct SporeVolume;
    
public:

    /// Default factory creates volume objects for this class.
    /// Subclass may thus choose to create volumes of its own type.
    /// Too bad C++ doesn't allow virtual functions from constructor (I know, big debate...)
    static SporeVolume* sporeVolumeFactory(void* user);

    typedef SporeVolume* (*SporeVolumeFactory)(void*);
    
    // Same API as for the locality DB. Add diffusion coefficients
    SporeCloud(double originx, double originy, double originz,
             double sizex, double sizey, double sizez,
             int divx, int divy, int divz, double baseDiffusionCoef, double lowestConcentration, SporeVolumeFactory factory = &sporeVolumeFactory, void* user = 0);

    
    /// for info purpose
    inline double getDiffusionCoefficient() {return diffusionCoefficient;}

    virtual ~SporeCloud();

    // Interface for spore generating agents. Generic name, use SporeCloud namespace.
    // ref counted, so spore cloud share unique instance & destroyed when last cell releases its ref
    struct Genome {
        int count;
        inline Genome() {count=1;}
        inline Genome* ref() {++count; return this;}
        inline void unref() {if (--count==0) delete this;}
        // don't copy count, just the values.
        inline Genome& operator=(const Genome& g) {decayConstant = g.decayConstant; return *this;}
        
        // too bad, but we need polymorphic type for dynamic_cast
        // => only way for multiple species to find back their compatible genomes
        virtual ~Genome();
        
        /// Spore activity follow exponential decay: x_t = x_0 * exp(decayConstant * t)
        /// This value is <0 for decay
        double decayConstant;

        /// Utility method to set up the decay constant.
        /// Express half-life in terms of simulator time
        inline void setHalfLife(double halfLife) {
            decayConstant = (-0.69314718056) / halfLife; // - ln2 / halfLife
        }
    };

    /// numerical integration step. Warning... condition stability applies...
    /// please use the schedule method and check the return value!
    virtual void step();

    /// Check that the integrationTimeStep respects the stability condition
    /// and in this case, schedule the various parts of the computations
    /// by spreading them over the time step to reduce CPU peaks
    /// return true if the stability condition is OK, false otherwise
    virtual bool schedule(double integrationTimeStep);

    /// Adds an individual agent spore cloud to the global cloud for diffusion
    /// The genome will be ref'd in case the agent dies
    /// An half-life in simulator time units should be provided for the spore vitality!
    virtual void addSpore(double x, double y, double z, Genome* genome, double concentration);

    /// Was initially a map, but inefficient iterators => Try a sorted list
    /// Still inefficient in accessing elements & iterators (thanks oprofile)
    /// => make own structure!
    //typedef std::list<std::pair<Genome*, std::pair<double, double> > > ConcentrationMap;
    //typedef std::list<std::pair<Genome*, std::pair<double, double> > > ConcentrationMap;
    struct ConcentrationList {
        Genome* genome;
        double concentration;
        double nextConcentration;
        ConcentrationList* next;
        ConcentrationList* prev;
    };

    /// Returns a concentration list at the given location
    /// Used by agents during reproduction
    /// @return the concentration list for the given location. May be null.
    virtual ConcentrationList* getConcentrationList(double x, double y, double z);
    
protected:


    
    /// all the lists present in one volume
    struct SporeVolume {
        /// links to neighbors to avoid recomputing them each time
        SporeVolume *xn, *xp, *yn, *yp, *zn, *zp, *next, *nextProcess;
        /// genome concentrations map
        ConcentrationList *concentrations;
        bool processFlag;
        // unref all genomes in map
        ~SporeVolume();
    };

    /// data structure for xyz bricks, actually a big array of spore volume list.
    /// Array is still necessary for direct access, from (x,y,z) => correct volume
    SporeVolume** volume; // buffer
    SporeVolume** volumeEnd; // buffer end
    int vsize; // divx*divy*divz
    /// But for list traversal, it's more efficient not to visit unused cells
    SporeVolume* firstVolume;
    /// And for each step, it's even more efficient to visit only those with non-empty neighbors
    SporeVolume* firstProcess;
    
    /** Re-used from locality DB: the origin is the super-brick corner minimum coordinates */
    double originx, originy, originz;

    /** Re-used from locality DB: length of the edges of the super-brick */
    double sizex, sizey, sizez;

    /** Re-used from locality DB: number of sub-brick divisions in each direction */
    int divx, divy, divz;

    /// Used as dt for integration
    double timeStep;

    /// Min concentration for a volume. Below this, consider there is no cloud
    double lowestConcentration;
    
    /// Diffusion coefficients constants, precomputed with grid size & time step
    /// Plan for anisotropic diffusion if gravity taken into account & maybe global wind?
    double diffuseXN, diffuseXP, diffuseYN, diffuseYP, diffuseZN, diffuseZP;

    /// for info purpose
    double diffusionCoefficient;

    /// consider world is cyclic in XY => no boundary condition!!!
    /// Boundary condition in Z is closed world = altitude limit
    /// Please use pre-computed traversal pointers in volumes instead of this
    /// This is a utility for the constructor because I'm lazy to split all the special boundary cases!
    /// May return null pointer if this cell is unused
    SporeVolume* cyclicVolume(int x, int y, int z) {
        if (x<0) x += divx;
        else if (x>=divx) x -= divx;
        if (y<0) y += divy;
        else if (y>=divy) y -= divy;
        // May return null pointer if this cell is unused
        return volume[z * (divx * divy) + y * divx + x];
    }

//    virtual void lock(SporeVolume* volume) const;
//    virtual void unlock(SporeVolume* volume) const;

    /// computations are distributed in time
    void intermediateStepDecay();
    void intermediateStepXn();
    void intermediateStepXp();
    void intermediateStepYn();
    void intermediateStepYp();
    void intermediateStepZn();
    void intermediateStepZp();
    void intermediateStepPostProcess();

// Utility function defined as inline for the CPP file. Private use
private:
    inline void diffuseVolume(SporeVolume* thisCell, SporeVolume* neighbor, double diffuseCoef);

};

#endif
