/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <vector>

#include <osg/Geometry>

#include "Terrain.h"
#include "Utility.h"

#include "waveNoise.h"

using namespace std;

Terrain::Terrain( double worldSize, double maxHeight, int level, double persistance, double power ) {

        this->worldSize = worldSize;
        halfWorldSize = worldSize * 0.5;
        this->maxHeight = maxHeight;
        this->persistance = persistance;
        this->power = power;
        this->level = level;
    
        dsize = 1 << level;
        dsizeMask = dsize - 1;
        invWorldSize_dsize = dsize / worldSize;
        
        data = new double[dsize*dsize];
        for (int i=0; i<dsize*dsize; ++i) data[i]=0;
        waveletNoise(data, level, persistance, power);

        // Uncomment this to export height maps
        //writePGM(data,dsize,"waveNoise.pgm");
        
        // The forest osg demo example uses quad strips
        // But this is badly non-planar!!!
        // So, this implementation uses triangle strips.
        // It also computes the normals directly so as to take in account edge wrap.
        // Finally, we're far away from the simple osg demo!!!
        // 
        // TODO: do triangle strips in both directions: fill one dir first,
        // then do the other direction to fill the holes

        osg::Geometry* geometry = new osg::Geometry;

        dsizep1 = dsize + 1;
        const int arraySize = dsizep1*dsizep1;
        osg::Vec3Array& v = *(new osg::Vec3Array(arraySize));
        //osg::Vec2Array& t = *(new osg::Vec2Array(arraySize));
        colorArray = new osg::Vec4Array(arraySize);
        osg::Vec4Array& color = *colorArray;

        const double interval = worldSize / (double)dsize;
//        const double texDelta = 1.0 / dsize;

        const double startxy = -worldSize*0.5;
        double px = startxy, py = startxy;
        //double tx = 0, ty = 0;
        
        int vi=0;
        double *dp = data;
        for(int r=0; r<dsize; ++r) {
            px = startxy;
            //tx = 0.0;
            for(int c=0; c<dsize; ++c) {
                *dp *= maxHeight;
                v[vi].set(px, py, *(dp++));
                //t[vi].set(tx, ty);
                px += interval;
                //tx += texDelta;
                ++vi;
            }
            // copy edge location
            v[vi].set(v[vi-dsize]);
            v[vi].x() = halfWorldSize; // avoid roundup error
            ++vi;
            py += interval;
            //ty += texDelta;
        }
        // copy edge locations
        for(int c=0; c<=dsize; ++c) {
            v[vi].set(v[c]);
            v[vi].y() = halfWorldSize; // avoid roundup error
            ++vi;
        }
        
        geometry->setVertexArray(&v);
        //geometry->setTexCoordArray(0,&t);

        int rowoffset = 0;
        
        // allocate and init normals
        normals = new osg::Vec3Array(arraySize);
        for(int r=0; r<dsizep1; ++r) {
            for(int c=0; c<dsizep1; ++c) {
                (*normals)[rowoffset+c].set(0,0,0);
            }
            rowoffset += dsizep1;
        }

        rowoffset = 0;
        int ei = 0, order = 0;
        vector<int> points;
        osg::Vec3d normal;
        for(int r=0; r<dsize; ++r) {
            order = 0; // P1P4=>1, P2P3=>2, init=0
            for(int c=0; c<dsize; ++c) {
                // point heights
                double z1 = wrapDataEdge(c,r);
                double z2 = wrapDataEdge(c+1,r);
                double z3 = wrapDataEdge(c,r+1);
                double z4 = wrapDataEdge(c+1,r+1);
                // indices in the vertex & normal arrays
                int idx1 = rowoffset + c;
                int idx2 = rowoffset + c + 1;
                int idx3 = rowoffset + dsizep1 + c;
                int idx4 = rowoffset + dsizep1 + c + 1;
                // compare mean heights of opposite corners to decide how to split quad
                // P1=height(r,  c)  P2=height(r,  c+1)
                // P3=height(r+1,c)  P4=height(r+1,c+1)
                // if (P1 + P4 > P2 + P3) ...
                if ( z1 + z4 > z2 + z3 ) {

                    // strip is row oriented, can't go very far...
                    // terminate previous strip
                    if (order!=0) {
                        osg::DrawElementsUShort& drawElements = *(new osg::DrawElementsUShort(GL_TRIANGLE_STRIP,points.size()));
                        ei = 0;
                        for (vector<int>::iterator it = points.begin(); it != points.end(); ++it) drawElements[ei++] = *it;
                        geometry->addPrimitiveSet(&drawElements);
                        points.clear();
                    }
                    
                    // also includes init case
                    // start with current column, may be 0th at init
                    points.push_back(idx3);
                    points.push_back(idx4);
                    points.push_back(idx1);
                    points.push_back(idx2);
                    order = 1;
                

                    // triangle normal (left-hand-side!): (P1-P3) ^ (P4-P3)
                    // v1.set(0, -interval, z1-z3);
                    // v2.set(interval, 0, z4-z3);
                    // (v1 ^ v2) / interval = normal next line
                    normal.set(z3 - z4, z1 - z3, interval);
                    normal.normalize();
                    // add the triangle normal to P1, P3, P4 vertices
                    (*normals)[idx1] += normal;
                    (*normals)[idx3] += normal;
                    (*normals)[idx4] += normal;
                   
                    // process second triangle (left-hand-side!): (P4-P2) ^ (P1-P2)
                    // v1.set(0, interval, z4-z2);
                    // v2.set(-interval, 0, z1-z2);
                    // (v1 ^ v2) / interval = normal next line
                    normal.set(z1 - z2, z2 - z4, interval);
                    normal.normalize();
                    (*normals)[idx1] += normal;
                    (*normals)[idx2] += normal;
                    (*normals)[idx4] += normal;

                }
                else {


                    // terminate previous strip
                    if (order==1) {
                        osg::DrawElementsUShort& drawElements = *(new osg::DrawElementsUShort(GL_TRIANGLE_STRIP,points.size()));
                        ei = 0;
                        for (vector<int>::iterator it = points.begin(); it != points.end(); ++it) drawElements[ei++] = *it;
                        geometry->addPrimitiveSet(&drawElements);
                        points.clear();
                    }

                    // also includes init case
                    if (order!=2) {
                        // start with current column, may be 0th at init
                        points.push_back(idx1);
                        points.push_back(idx3);
                        points.push_back(idx2);
                        points.push_back(idx4);
                        order = 2;
                    }
                    else {
                        // continue strip
                        points.push_back(idx2);
                        points.push_back(idx4);
                    }


                                
                    // triangle normal (left-hand-side!): (P2-P1) ^ (P3-P1)
                    // v1.set(interval, 0, z2-z1);
                    // v2.set(0, interval, z3-z1);
                    // (v1 ^ v2) / interval = normal next line
                    normal.set(z1 - z2, z1 - z3, interval);
                    normal.normalize();
                    // add the triangle normal to P1, P3, P4 vertices
                    (*normals)[idx1] += normal;
                    (*normals)[idx3] += normal;
                    (*normals)[idx2] += normal;

                    // process second triangle normal (left-hand-side!): (P3-P4) ^ (P2-P4)
                    // v1.set(-interval, 0, z3-z4);
                    // v2.set(0, -interval, z2-z4);
                    // (v1 ^ v2) / interval = normal next line
                    normal.set(z3 - z4, z2 - z4, interval);
                    normal.normalize();
                    (*normals)[idx3] += normal;
                    (*normals)[idx2] += normal;
                    (*normals)[idx4] += normal;

                }
            }
            rowoffset += dsizep1;
            // terminate this row strip
            osg::DrawElementsUShort& drawElements = *(new osg::DrawElementsUShort(GL_TRIANGLE_STRIP,points.size()));
            ei = 0;
            for (vector<int>::iterator it = points.begin(); it != points.end(); ++it) drawElements[ei++] = *it;
            geometry->addPrimitiveSet(&drawElements);
            points.clear();
        }
        
        // Compute normals averaged over faces
/*      // smooth visitor does a bad job with normal normalization, plus it is not cyclic.
        osgUtil::SmoothingVisitor sv;
        sv.smooth(*geometry);
        normals = geometry->getNormalArray();
*/
        // first complete the cycles by adding both sides
        rowoffset = 0;
        for(int r=0; r<dsizep1; ++r) {
            (*normals)[rowoffset] += (*normals)[rowoffset + dsize];
            (*normals)[rowoffset + dsize] = (*normals)[rowoffset];
            rowoffset += dsizep1;
        }
        rowoffset = dsizep1*dsize;
        for(int c=0; c<dsizep1; ++c) {
            (*normals)[c] += (*normals)[rowoffset + c];
            (*normals)[rowoffset + c] = (*normals)[c];
        }

        // Now normalize the normals, and process colors

        osg::Vec4 colorVertical(192.0 /255.0, 128.0 /255.0, 90.0 /255.0, 1.0);
        osg::Vec4 colorHorizontal(146.0 /255.0, 162.0 /255.0, 100.0 /255.0, 1.0);
        osg::Vec4 topColor(232.0 /255.0, 229.0 /255.0, 238.0 /255.0, 1.0);
        osg::Vec4 diffColor = colorVertical - colorHorizontal;
        rowoffset = 0;
        double invMaxHeight = 1.0 / maxHeight;
        for(int r=0; r<dsizep1; ++r) {
            for(int c=0; c<dsizep1; ++c) {
                // normalize the normal now that all sides were summed
                (*normals)[rowoffset + c].normalize();
                // use cosine with horizontal vector for interpolating
                // when cos is 1.0, normal is horizontal, thus vertical color is used
                color[rowoffset + c] = colorHorizontal + diffColor * fabs((*normals)[rowoffset + c].x());
                color[rowoffset + c] = color[rowoffset + c] + (topColor - color[rowoffset + c]) * invMaxHeight * wrapDataEdge(c,r);
            }
            rowoffset += dsizep1;
        }
        geometry->setColorArray(&color);
        geometry->setColorBinding(osg::Geometry::BIND_PER_VERTEX);
        
        geometry->setNormalArray(normals);
        geometry->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);

        
/*      // Useless, use the w key...     
        osg::PolygonMode* Wireframe = new osg::PolygonMode();
        Wireframe->setMode(osg::PolygonMode::FRONT_AND_BACK,osg::PolygonMode::LINE);
  
        osg::StateSet *stateset = getOrCreateStateSet();
        stateset->setAttributeAndModes(Wireframe, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON);
        setStateSet(stateset);
*/
        
        // This node is ready to draw
        //addDrawable(new osg::ShapeDrawable(grid));
        addDrawable(geometry);

}


Terrain::~Terrain() {

    // normals reference counted , handled by osg
    delete[] data;
}


/**
    Returns the Z coordinate of the floor at the given (x,y) world coordinate.
    If outside, propagate the border value
*/
double Terrain::getFloorZ(const double x, const double y) {
    
    double px = (x + halfWorldSize) * invWorldSize_dsize;
    double py = (y + halfWorldSize) * invWorldSize_dsize;
    
    if (px>dsize) px = dsize;
    if (px<0) px = 0;
    if (py>dsize) py = dsize;
    if (py<0) py = 0;

    const int lx = (int)(px);
    const int hx = (int)(px+1.0);
    const int ly = (int)(py);
    const int hy = (int)(py+1.0);

    //const int offsetl = (ly>=dsize ? 0 : ly) * dsize;
    const int offsetl = (ly & dsizeMask) << level;
    //const int wlx = (lx>=dsize ? 0 : lx);
    const int wlx = lx & dsizeMask;
    const double d1 = data[offsetl + wlx];

    if (ly==py) {
        if (lx==px) return d1;
        return d1 + (data[offsetl + (hx & dsizeMask)] - d1) *(px - lx);
    }

    //const int offseth = (hy>=dsize ? hy-dsize : hy) * dsize;
    const int offseth = (hy & dsizeMask) << level;
    if (lx==px) return d1 + (data[offseth + wlx] - d1) * (py - ly);

    //const int whx = (hx>=dsize ? hx-dsize : hx);
    const int whx = hx & dsizeMask;
    
    // most general case in between both x and y
    const double d2 = data[offsetl + whx];
    const double d3 = data[offseth + wlx];
    const double d4 = data[offseth + whx];

    const double coefx = (px - lx);
    const double coefy = (py - ly);
    
    // Match triangle split
    if (d1+d4 > d2+d3) {
        if (coefx > coefy) {
            return d1 + coefx * (d2 - d1) + coefy * (d4 - d2);
        } else {
            return d1 + coefy * (d3 - d1) + coefx * (d4 - d3);
        }
    } else {
        if (coefx + coefy < 1) {
            return d1 + coefx * (d2 - d1) + coefy * (d3 - d1);
        } else {
            return d3 + coefx * (d4 - d3) + (1.0 - coefy) * (d2 - d4);
        }
    }

}


osg::Vec3d Terrain::getFloorNormal(const double x, const double y) {
    
    double px = (x + halfWorldSize) * invWorldSize_dsize;
    double py = (y + halfWorldSize) * invWorldSize_dsize;
    
    if (px>dsize) px = dsize;
    if (px<0) px = 0;
    if (py>dsize) py = dsize;
    if (py<0) py = 0;

    const int lx = (int)(px);
    const int hx = (int)(px+1.0);
    const int ly = (int)(py);
    const int hy = (int)(py+1.0);

    const int offsetdl = (ly & dsizeMask) << level; // data offset
    const int offsetnl = ly * dsizep1;              // normal offset
    const int wlx = lx & dsizeMask;
    
    const osg::Vec3 n1 = (*normals)[offsetnl + lx];

    if (ly==py) {
        if (lx==px) return n1;
        else {
            osg::Vec3d ret = n1 + ((*normals)[offsetnl + hx] - n1) *(px - lx);
            ret.normalize();
            return ret;
        }
    }

    if (lx==px) {
        osg::Vec3d ret = n1 + ((*normals)[offsetnl + dsizep1 + lx] - n1) * (py - ly);
        ret.normalize();
        return ret;
    }
    
    // most general case in between both x and y
    const int offsetnh = offsetnl + dsizep1;
    const int offsetdh = (hy & dsizeMask) << level;
    const int whx = hx & dsizeMask;
    

    const double d1 = data[offsetdl + wlx];
    const double d2 = data[offsetdl + whx];
    const double d3 = data[offsetdh + wlx];
    const double d4 = data[offsetdh + whx];
    
    const osg::Vec3 n2 = (*normals)[offsetnl + hx];
    const osg::Vec3 n3 = (*normals)[offsetnh + lx];
    const osg::Vec3 n4 = (*normals)[offsetnh + hx];

    const double coefx = (px - lx);
    const double coefy = (py - ly);


    // Match triangle split
    if (d1+d4 > d2+d3) {
        if (coefx > coefy) {
            osg::Vec3d ret = n1 + (n2 - n1) * coefx + (n4 - n2) * coefy;
            ret.normalize();
            return ret;
        } else {
            osg::Vec3d ret = n1 + (n3 - n1) * coefy + (n4 - n3) * coefx;
            ret.normalize();
            return ret;
        }
    } else {
        if (coefx + coefy < 1) {
            osg::Vec3d ret = n1 + (n2 - n1) * coefx + (n3 - n1) * coefy;
            ret.normalize();
            return ret;
        } else {
            osg::Vec3d ret = n3 + (n4 - n3) * coefx + (n2 - n4) * (1.0 - coefy);
            ret.normalize();
            return ret;
        }
    }
    
}



osg::Group* Terrain::drawNormals(osg::Group* normalGroup) {
    int rowoffset = 0;
    const double interval = worldSize / dsize;
    double x,y;
    y = -halfWorldSize;
    if (normalGroup==0) normalGroup = new osg::Group;
    for(int r=0; r<dsizep1; ++r) {
        x = -halfWorldSize;
        for(int c=0; c<dsizep1; ++c) {
            Utility::drawLine(
                osg::Vec3d(x,y,wrapDataEdge(c,r)),
                osg::Vec3d(x,y,wrapDataEdge(c,r)) + (*normals)[rowoffset+c],
                osg::Vec4(0,0,1,1),
                osg::Vec4(0.7,0.7,1,1),
                normalGroup
            );
            x+= interval;
        }
        y+= interval;
        rowoffset += dsizep1;
    }
    return normalGroup;
}
    


