/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef TERRAIN_H
#define TERRAIN_H

#include <osg/Geode>
#include <osg/Group>
#include <osg/Array>

#include "TerrainInterface.h"

/**
    A Terrain that uses wavelet noise for height field generation
    
    @author Nicolas Brodu
*/
class Terrain : public osg::Geode, public TerrainInterface {
public:
    
    Terrain(double worldSize, double maxHeight, int level, double persistance=0.5, double power=1.0);

    ~Terrain();

    /**
       Returns the Z coordinate of the floor at the given (x,y) world coordinate.
       If outside, propagate the border value
    */
    virtual double getFloorZ(const double x, const double y);
    
    /**
       Returns the normal of the floor at the given (x,y) world coordinate.
       If outside, propagate the border value
    */
    virtual osg::Vec3d getFloorNormal(const double x, const double y);

    /** Draw the terrain normals.
        @param normalGroup The osg group that will contain the normals representation as lines. Null by default,
                           in which case a new group is created to contain all the normals.
        @return The osg group that contains the normals representation as lines
    */
    osg::Group* drawNormals(osg::Group* normalGroup = 0);
    
    // read-only public access
    inline double getWorldSize() const { return worldSize; }
    inline double getHalfWorldSize() const { return halfWorldSize; }
    inline double getMaxHeight() const { return maxHeight; }
    
    inline const double* getData() const {return data;}
    inline const osg::Vec3Array* getNormals() const {return normals;}

    inline double getDataArraySize() const { return dsize; }
    inline double getNormalsArraySize() const { return dsizep1; }

        
protected:

    double worldSize, halfWorldSize, invWorldSize_dsize, maxHeight, persistance, power;
    int dsize, dsizep1, level, dsizeMask;
    double *data;
    osg::Vec3Array *normals;
    osg::Vec4Array *colorArray;

    // make data appear as an array of size (dsize+1) * (dsize+1)
    // merge each edge
    inline double wrapDataEdge(int x, int y) {
        if (y==dsize) y = 0;
        if (x==dsize) x = 0;
        return data[y*dsize+x];
    }

};

#endif
