/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>

#include "Utility.h"

// Uncomment this and see the code at the end of this file for debugging
// Trick not to have to modify the header to avoid recompiling dependent objects
/*
#define TWOLAYERPERCEPTRON_H_DEBUG_MODE \
    public: void gradCheck();
*/

#include "TwoLayerPerceptron.h"

using namespace std;

const double TwoLayerPerceptron::defaultLearningRate = 0.1;

TwoLayerPerceptron::TwoLayerPerceptron( int ninput, int nhidden, int noutput )
{
    this->ninput = ninput;
    this->nhidden = nhidden;
    this->noutput = noutput;
    
    nih = ninput * nhidden;
    nho = nhidden * noutput;

    // input to hidden weights: array of size ninput * nhidden
    ihw = new double[nih];
    // input to hidden weight gradients: array of size ninput * nhidden
    ihwg = new double[nih];
    // initialize
    double scale = 1.0 / sqrt(ninput+1.0);
    for (int i=0; i<nih; ++i) {
        ihw[i] = Utility::NRandom() * scale;
        ihwg[i] = 0.0;
    }
    
    // hidden to output weights: array of size nhidden * noutput
    how = new double[nho];
    // hidden to output weight gradients: array of size nhidden * noutput
    howg = new double[nho];
    // initialize
    scale = 1.0 / sqrt(nhidden+1.0);
    for (int i=0; i<nho; ++i) {
        how[i] = Utility::NRandom() * scale;
        howg[i] = 0.0;
    }

    // hidden unit bias: array of size nhidden
    hb = new double[nhidden];
    // hidden unit bias gradients: array of size nhidden
    hbg = new double[nhidden];
    // initialize
    for (int i=0; i<nhidden; ++i) {
        hb[i] = Utility::NRandom() * scale;
        hbg[i] = 0.0;
    }

    // output unit bias gradients: array of size output
    obg = new double[noutput];
    // output unit bias: array of size output
    ob = new double[noutput];
    // initialize
    scale = 1.0 / sqrt(noutput+1.0);
    for (int i=0; i<noutput; ++i) {
        ob[i] = Utility::NRandom() * scale;
        obg[i] = 0.0;
    }

    // hidden values holder.
    hv = new double[nhidden];
    for (int i=0; i<nhidden; ++i) hv[i] = 0;
}

TwoLayerPerceptron::TwoLayerPerceptron( const TwoLayerPerceptron & tlp )
{
    // avoid code dup...
    ihw = how = ihwg = howg = hb = ob = hbg = obg = hv = 0;
    // ...by reusing operator=
    *this = tlp;
}

TwoLayerPerceptron & TwoLayerPerceptron::operator =( const TwoLayerPerceptron & tlp )
{
    ninput = tlp.ninput;
    nhidden = tlp.nhidden;
    noutput = tlp.noutput;
    nih = tlp.nih;
    nho = tlp.nho;
    
    // for copy constructor call detection
    if (ihw!=0) delete [] ihw; if (how!=0) delete [] how; if (ihwg!=0) delete [] ihwg;
    if (howg!=0) delete [] howg; if (hb!=0) delete [] hb; if (ob!=0) delete [] ob;
    if (hbg!=0) delete [] hbg; if (obg!=0) delete [] obg; if (hv!=0) delete [] hv;
    
    ihw = new double[nih]; ihwg = new double[nih];
    for (int i=0; i<nih; ++i) {
        ihw[i] = tlp.ihw[i];
        ihwg[i] = tlp.ihwg[i];
    }

    how = new double[nho]; howg = new double[nho];
    for (int i=0; i<nho; ++i) {
        how[i] = tlp.how[i];
        howg[i] = tlp.howg[i];
    }

    hb = new double[nhidden]; hbg = new double[nhidden]; hv = new double[nhidden];
    for (int i=0; i<nhidden; ++i) {
        hb[i] = tlp.hb[i];
        hbg[i] = tlp.hbg[i];
        hv[i] = tlp.hv[i];
    }

    obg = new double[noutput]; ob = new double[noutput];
    for (int i=0; i<noutput; ++i) {
        ob[i] = tlp.ob[i];
        obg[i] = tlp.obg[i];
    }
}




TwoLayerPerceptron::~TwoLayerPerceptron()
{
    delete [] ihw;
    delete [] how;
    
    delete [] ihwg;
    delete [] howg;

    delete [] hb;
    delete [] ob;
    
    delete [] hbg;
    delete [] obg;
    
    delete [] hv;
}


// compiler use temporary instead of explicit variable
inline double square(double x) {
    return x*x;
}

// f(x) = x / (1 + abs(x))
static double defaultTransfer(double x) {
    return x / (1.0 + fabs(x));
}

// f'(x) = (1 - abs(f)) ^2
static double defaultTransferDerivativeAsF(double f) {
    return square(1.0 - fabs(f));
}


double (*TwoLayerPerceptron::transfer)(double) = &defaultTransfer;
double (*TwoLayerPerceptron::transferDerivativeAsF)(double) = &defaultTransferDerivativeAsF;


/*
// f(x) = tanh
inline double transfer(const double x) {
    return tanh(x);
}

// f'(x) = 1 - f^2
inline double transferDerivativeAsF(const double f) {
    return 1.0 - f*f;
}
*/

/*
// f(x) = 1/(1+exp(-x))
inline double transfer(const double x) {
    return 1.0/(1.0+exp(-x));
}

// f'(x) = f*(1-f)
inline double transferDerivativeAsF(const double f) {
    return f*(1.0-f);
}
*/

/*
// f(x) = x / (1 + abs(x))
inline double transfer(const double x) {
    return 1.5*x / (1.0 + 1.5*fabs(x));
}

// f'(x) = (1 - abs(f)) ^2
inline double transferDerivativeAsF(const double f) {
    return 1.5 * square(1.0 - fabs(f));
}
*/

void TwoLayerPerceptron::computeOutput( const double * input, double * output )
{
    // compute hidden unit values
    double *rp = hv;  // result pointer
    double *wp = ihw; // weights pointer
    const double* dp; // data pointer
    const double* dpend = input + ninput; // end of data pointer
    for (int h=0; h<nhidden; ++h) {
        *rp = 0.0;
        // compute sum of inputs*weights
        for (dp = input; dp<dpend; ++dp) {
            *rp += *(wp++) * (*dp);
        }
        // apply transfer function
        *rp = transfer(*rp + hb[h]);
        ++rp;
    }
    
    // compute output values
    rp = output;  // result pointer
    wp = how;     // weights pointer
    dpend = hv + nhidden; // end of data pointer
    for (int o=0; o<noutput; ++o) {
        *rp = 0.0;
        // compute sum of hidden*weights
        for (dp = hv; dp<dpend; ++dp) {
            *rp += *(wp++) * (*dp);
        }
        // apply transfer function
        *rp = transfer(*rp + ob[o]);
        ++rp;
    }
    
}

void TwoLayerPerceptron::backPropagate( const double * input, const double * output, const double * gradout )
{

    // Error gradient by weight: dE/dw = dE/do * do/dw = gradout * do/dw
    // Error gradient by bias: dE/db = gradout * do/db
    // o = transfer(b + sum(hidden*weights))
    // So: do/db = transferDerivative(b + sum(values*weights))
    //     do/dw = value * transferDerivative(b + sum(values*weights))
    // So: do/db = transferDerivativeAsF(transfer(b + sum(values*weights)))
    //     do/dw = value * transferDerivativeAsF(transfer(b + sum(values*weights)))
    // So: do/db = transferDerivativeAsF(output_value)
    //     do/dw = value * transferDerivativeAsF(output_value)


    // backpropagation in hidden-to-output layer:
    // dE/dw = gradout * do/dw = gradout * hidden_value * transferDerivativeAsF(output_value)
    // dE/db = gradout * do/db = gradout * transferDerivativeAsF(output_value)
    // Thus: howg[h,o] = gradout[o] * hv[h] * transferDerivativeAsF(output[o])
    //       obg[o] = gradout[o] * transferDerivativeAsF(output[o])
    // Hence: obg[o] = gradout[o] * transferDerivativeAsF(output[o])
    //        howg[h,o] = obg[o] * hv[h]
    
    // backpropagation in input-to-hidden layer:
    // First we need to compute equivalent of gradout for hidden layer
    // with h = value of hidden unit, chain rule on all outputs which depend on h (all of them in this case)
    // dE/dh = sum(dE/do * do/dh), with o = transfer(b + sum(hidden*weights))
    // dE/dh = sum(gradout * weight(h,o) * transferDerivative(bo + sum(h*weight(h,o))))
    // dE/dh = sum(gradout(o) * weight(h,o) * transferDerivativeAsF(o))
    // dE/dh = sum(weight(h,o) * obj[o])
    // Let's store temporarily hbg[h] = sum(how[h,o] * obj[o])
    // Then, as above:
    // dE/dw = dE/dh * dh/dw = dE/dh * input_value * transferDerivativeAsF(h)
    // dE/db = dE/dh * dh/db = dE/dh * transferDerivativeAsF(h)
    // Thus, combining dE/dh with the above formula:
    //     hbg[h]_final = hbg[h]_temp * transferDerivativeAsF(hv[h])
    //     ihwg[i,h] = hbg[h]_final * input[i]

    // step one, compute all hidden-to-output related variables: obg[o], howg[h,o], and temporary dE/dh in hbg[h]
    const double* dp; // hidden value data pointer
    const double* dpend = hv + nhidden; // const because it will be reused for input data
    double* bgp = obg;  // bias gradient pointer: dE/db for b output bias
    double* wgp = howg; // weights gradient pointer: dE/dw for w hidden-to-output weight
    double* tgp;        // gradout equivalent for hidden layer: dE/dh for h hidden value
    double* tgpend = hbg + nhidden;
    double* wp = how;   // weights pointer: for dE/dh computation
    for (tgp = hbg; tgp<tgpend; ++tgp) *tgp = 0.0; // initialize sum for dE/dh 
    for (int o=0; o<noutput; ++o) {
        *bgp = gradout[o] * transferDerivativeAsF(output[o]); // see above
        for (dp = hv; dp<dpend; ++dp) {
            *(wgp++) = *bgp * (*dp);       // see above: howg[h,o] = obg[o] * hv[h]
        }
        for (tgp = hbg; tgp<tgpend; ++tgp) {
            *tgp += *(wp++) * (*bgp);      // see above: hbg[h] += how[h,o] * obg[o]
        }
        ++bgp;
    }

    // step two: compute input-to-hidden variables: ihwg[i,h] and final hbg[h]
    dpend = input + ninput;
    bgp = hbg;  // bias gradient pointer: dE/db for b hidden bias
    wgp = ihwg; // weights gradient pointer: dE/dw for w input-to-hidden weight
    for (int h=0; h<nhidden; ++h) {
        *bgp *= transferDerivativeAsF(hv[h]); // hbg[h]_final = hbg[h]_temp * transferDerivativeAsF(hv[h])
        for (dp = input; dp<dpend; ++dp) {
            *(wgp++) = *bgp * (*dp);          // see above: ihwg[i,h] = hbg[h]_final * input[i]
        }
        ++bgp;
    }

    // Done!!!
}


void TwoLayerPerceptron::batchBackPropagateAccumulate( const double * input, const double * output, const double * gradout )
{
    // See comments in the previous function
    // The only difference here is that gradients are accumulated over previous
    // ones, instead of replacing them
    
    const double* dp; // hidden value data pointer
    const double* dpend = hv + nhidden; // const because it will be reused for input data
    double* bgp = obg;    // bias gradient pointer: dE/db for b output bias
    double* wgp = howg;   // weights gradient pointer: dE/dw for w hidden-to-output weight
    double temp[nhidden]; // will be used to store the temporary sum(how[h,o] * obj[o]), we can't override hbg here
    double* tgp;          // gradout equivalent for hidden layer: dE/dh for h hidden value
    double* tgpend = temp + nhidden;
    double* wp = how;     // weights pointer: for dE/dh computation
    double current_bgp;   // another temporary value for a variable we can't erase
    
    for (tgp = temp; tgp<tgpend; ++tgp) *tgp = 0.0; // initialize sum for dE/dh

    for (int o=0; o<noutput; ++o) {
        current_bgp = gradout[o] * transferDerivativeAsF(output[o]); // see above
        for (dp = hv; dp<dpend; ++dp) {
            *(wgp++) += current_bgp * (*dp);    // see above: howg[h,o] = obg[o] * hv[h]. Accumulate result
        }
        for (tgp = temp; tgp<tgpend; ++tgp) {
            *tgp += *(wp++) * current_bgp;      // see above: hbg[h]_temp += how[h,o] * obg[o]
        }
        *(bgp++) += current_bgp;                // Accumulate result
    }

    // step two: compute input-to-hidden variables: ihwg[i,h] and final hbg[h]
    dpend = input + ninput;
    bgp = hbg;  // bias gradient pointer: dE/db for b hidden bias
    wgp = ihwg; // weights gradient pointer: dE/dw for w input-to-hidden weight
    for (int h=0; h<nhidden; ++h) {
        // hbg[h]_final = hbg[h]_temp * transferDerivativeAsF(hv[h]).
        current_bgp = temp[h] * transferDerivativeAsF(hv[h]); 
        for (dp = input; dp<dpend; ++dp) {
            *(wgp++) += current_bgp * (*dp);    // see above: ihwg[i,h] = hbg[h]_final * input[i]. Accumulate result
        }
        *(bgp++) += current_bgp;                // Accumulate result
    }
}

void TwoLayerPerceptron::batchBackPropagateTerminate(int nmappings)
{
    double factor = 1.0 / (double)nmappings;
    for (int i=0; i<nih; ++i) ihwg[i] *= factor;
    for (int i=0; i<nho; ++i) howg[i] *= factor;
    for (int i=0; i<nhidden; ++i) hbg[i] *= factor;
    for (int i=0; i<noutput; ++i) obg[i] *= factor;
}

void TwoLayerPerceptron::learn( double learningRate )
{
    for (int i=0; i<nih; ++i) ihw[i] -= ihwg[i] * learningRate;
    for (int i=0; i<nho; ++i) how[i] -= howg[i] * learningRate;
    for (int i=0; i<nhidden; ++i) hb[i] -= hbg[i] * learningRate;
    for (int i=0; i<noutput; ++i) ob[i] -= obg[i] * learningRate;
}

double TwoLayerPerceptron::train( const double * input, const double * target, double learningRate )
{
    double output[noutput];
    computeOutput(input, output);
    
    // simple error function: half sum of square => gradient is difference
    double gradout[noutput];
    for (int j=0; j<noutput; ++j) gradout[j] = output[j] - target[j];
    
    // backpropagation of error gradient
    backPropagate(input, output, gradout);
    learn(learningRate);
    
    // compute error
    double error = 0.0;
    for (int j=0; j<noutput; ++j) error += square(output[j] - target[j]);
    return error * 0.5; // half sum of square
}

void TwoLayerPerceptron::getHidden( double * hidden )
{
    for (int i=0; i<nhidden; ++i) hidden[i] = hv[i];
}

void TwoLayerPerceptron::setHidden( const double * hidden )
{
    for (int i=0; i<nhidden; ++i) hv[i] = hidden[i];
}

std::ostream& operator<<(std::ostream& os, const TwoLayerPerceptron& tlp){
    
    os << tlp.ninput << " " << tlp.nhidden << " " << tlp.noutput << endl;
    
    for (int i=0; i<tlp.nih; ++i) os << tlp.ihw[i] << " ";
    os << endl;
    
    for (int i=0; i<tlp.nhidden; ++i) os << tlp.hb[i] << " ";
    os << endl;
    
    for (int i=0; i<tlp.nho; ++i) os << tlp.how[i] << " ";
    os << endl;

    for (int i=0; i<tlp.noutput; ++i) os << tlp.ob[i] << " ";
    os << endl;

    return os;
}

std::istream& operator>>(std::istream& is, TwoLayerPerceptron& tlp) {
    
    is >> tlp.ninput;
    is >> tlp.nhidden;
    is >> tlp.noutput;
    tlp.nih = tlp.ninput * tlp.nhidden;
    tlp.nho = tlp.nhidden * tlp.noutput;
    
    delete [] tlp.ihw; delete [] tlp.how; delete [] tlp.ihwg;
    delete [] tlp.howg; delete [] tlp.hb; delete [] tlp.ob;
    delete [] tlp.hbg; delete [] tlp.obg; delete [] tlp.hv;
    
    tlp.ihw = new double[tlp.nih]; tlp.ihwg = new double[tlp.nih];
    for (int i=0; i<tlp.nih; ++i) {
        is >> tlp.ihw[i];
        tlp.ihwg[i] = 0.0;
    }

    tlp.hb = new double[tlp.nhidden]; tlp.hbg = new double[tlp.nhidden]; tlp.hv = new double[tlp.nhidden];
    for (int i=0; i<tlp.nhidden; ++i) {
        is >> tlp.hb[i];
        tlp.hbg[i] = 0.0;
        tlp.hv[i] = 0.0;
    }

    tlp.how = new double[tlp.nho]; tlp.howg = new double[tlp.nho];
    for (int i=0; i<tlp.nho; ++i) {
        is >> tlp.how[i];
        tlp.howg[i] = 0.0;
    }

    tlp.ob = new double[tlp.noutput]; tlp.obg = new double[tlp.noutput];
    for (int i=0; i<tlp.noutput; ++i) {
        is >> tlp.ob[i];
        tlp.obg[i] = 0.0;
    }

    return is;
}

void TwoLayerPerceptron::mutate( double ihwRate, double ihwJitter, double howRate, double howJitter, double hbRate, double hbJitter, double obRate, double obJitter )
{
    double rangeRate = ihwRate * 2.0;
    double offsetRate = 1.0 - ihwRate;
    double offsetJitter = - ihwJitter / sqrt(ninput+1.0);
    double rangeJitter = - 2.0 * offsetJitter;
    for (int i=0; i<nih; ++i) {
        ihw[i] *= offsetRate + rangeRate * Utility::Random();
        ihw[i] += offsetJitter + rangeJitter * Utility::Random();
    }

    rangeRate = howRate * 2.0;
    offsetRate = 1.0 - howRate;
    offsetJitter = - howJitter / sqrt(nhidden+1.0);
    rangeJitter = - 2.0 * offsetJitter;
    for (int i=0; i<nho; ++i) {
        how[i] *= offsetRate + rangeRate * Utility::Random();
        how[i] += offsetJitter + rangeJitter * Utility::Random();
    }
    
    rangeRate = hbRate * 2.0;
    offsetRate = 1.0 - hbRate;
    offsetJitter *= hbJitter / howJitter; //= - hbJitter / sqrt(nhidden+1.0);
    rangeJitter = - 2.0 * offsetJitter;
    for (int i=0; i<nhidden; ++i) {
        hb[i] *= offsetRate + rangeRate * Utility::Random();
        hb[i] += offsetJitter + rangeJitter * Utility::Random();
    }
    
    rangeRate = obRate * 2.0;
    offsetRate = 1.0 - obRate;
    offsetJitter = - obJitter / sqrt(noutput+1.0);
    rangeJitter = - 2.0 * offsetJitter;
    for (int i=0; i<noutput; ++i) {
        ob[i] *= offsetRate + rangeRate * Utility::Random();
        ob[i] += offsetJitter + rangeJitter * Utility::Random();
    }
    
}

//=========================================
// DEBUG CODE BELOW
//=========================================

#ifdef TWOLAYERPERCEPTRON_H_DEBUG_MODE

#include <iostream>
using namespace std;


// Test routines

void TwoLayerPerceptron::gradCheck()
{

    double* input = new double[ninput];
    for (int i=0; i<ninput; ++i) input[i] = Utility::NRandom();
    double* target = new double[noutput];
    for (int i=0; i<noutput; ++i) target[i] = Utility::NRandom();
    double* output = new double[noutput];

    computeOutput(input, output); // random garbage
    
    // simple error function: half sum of square => gradient is difference
    double* gradout = new double[noutput];
    for (int i=0; i<noutput; ++i) gradout[i] = output[i] - target[i];
    
    // test our backpropagation
    backPropagate(input, output, gradout);

    // compute explicitly the gradients
    int nih = ninput * nhidden;
    int nho = nhidden * noutput;
    double epsilon = 1.0e-5;
    double sav;
    double eplus, eminus;
    // ih weights derivative
    for (int i=0; i<nih; ++i) {
        sav = ihw[i];
        ihw[i] += epsilon;
        computeOutput(input, output);
        eplus = 0.0;
        for (int j=0; j<noutput; ++j) eplus += square(output[j] - target[j]);
        eplus *= 0.5; // half sum of square
        ihw[i] = sav - epsilon;
        computeOutput(input, output);
        eminus = 0.0;
        for (int j=0; j<noutput; ++j) eminus += square(output[j] - target[j]);
        eminus *= 0.5; // half sum of square
        ihw[i] = sav;
        // Use central difference formula for approximation of error gradient
        // difference between explicit and backprop values
        cout << "diff for ihwg[" << i << "] = " << 0.5*(eplus - eminus)/epsilon - ihwg[i] << endl;
    }
    for (int i=0; i<nho; ++i) {
        sav = how[i];
        how[i] += epsilon;
        computeOutput(input, output);
        eplus = 0.0;
        for (int j=0; j<noutput; ++j) eplus += square(output[j] - target[j]);
        eplus *= 0.5; // half sum of square
        how[i] = sav - epsilon;
        computeOutput(input, output);
        eminus = 0.0;
        for (int j=0; j<noutput; ++j) eminus += square(output[j] - target[j]);
        eminus *= 0.5; // half sum of square
        how[i] = sav;
        // Use central difference formula for approximation of error gradient
        // difference between explicit and backprop values
        cout << "diff for howg[" << i << "] = " << 0.5*(eplus - eminus)/epsilon - howg[i] << endl;
    }
    for (int i=0; i<nhidden; ++i) {
        sav = hb[i];
        hb[i] += epsilon;
        computeOutput(input, output);
        eplus = 0.0;
        for (int j=0; j<noutput; ++j) eplus += square(output[j] - target[j]);
        eplus *= 0.5; // half sum of square
        hb[i] = sav - epsilon;
        computeOutput(input, output);
        eminus = 0.0;
        for (int j=0; j<noutput; ++j) eminus += square(output[j] - target[j]);
        eminus *= 0.5; // half sum of square
        hb[i] = sav;
        // Use central difference formula for approximation of error gradient
        // difference between explicit and backprop values
        cout << "diff for hbg[" << i << "] = " << 0.5*(eplus - eminus)/epsilon - hbg[i] << endl;
    }
    for (int i=0; i<noutput; ++i) {
        sav = ob[i];
        ob[i] += epsilon;
        computeOutput(input, output);
        eplus = 0.0;
        for (int j=0; j<noutput; ++j) eplus += square(output[j] - target[j]);
        eplus *= 0.5; // half sum of square
        ob[i] = sav - epsilon;
        computeOutput(input, output);
        eminus = 0.0;
        for (int j=0; j<noutput; ++j) eminus += square(output[j] - target[j]);
        eminus *= 0.5; // half sum of square
        ob[i] = sav;
        // Use central difference formula for approximation of error gradient
        // difference between explicit and backprop values
        cout << "diff for obg[" << i << "] = " << 0.5*(eplus - eminus)/epsilon - obg[i] << endl;
    }
    
}

#include <time.h>

int main(int argc, char** argv) {

//    TwoLayerPerceptron tlp(7,9,5);

    double input[] = {1,2,3,4,5,6,7,8,9};
    double target[] = {0.2,0.3,0.5,0.7,0.11,0.13};
    const int noutput = 6;
    const int ninput = 9;
    const int nhidden = 20;
    double output[noutput];
    double gradout[noutput];

    TwoLayerPerceptron tlp(ninput,nhidden,noutput);
//    tlp.gradCheck();

    // training on only one mapping    
    for (int i=0; i<1000; ++i) {
        double error = tlp.train(input, target, 0.1);
        tlp.computeOutput(input, output);
        cout << "Error at step " << i << " is "<< error << ", outputs are ";
        for (int j=0; j<noutput; ++j) cout << output[j] << " ";
        cout << endl;
    }

    double input2[] = {7,5,9,9,6,4,5,23,1};
    double target2[] = {0.21,-.02,-.36,-.68,0.8,0.4};
    double input3[] = {4,8,9,-5,14,2,3,1,4};
    double target3[] = {-0.5,0.8,0.7,0.1,-0.3,0.2};
    
    TwoLayerPerceptron tlp2(ninput,nhidden,noutput);
    
    // batch backprop on several mappings
    for (int i=0; i<10000; ++i) {
        double error = 0.0;
        
        // see comments in train()
        tlp2.computeOutput(input, output);
        for (int j=0; j<noutput; ++j) gradout[j] = output[j] - target[j];
        tlp2.backPropagate(input, output, gradout);
        for (int j=0; j<noutput; ++j) error += square(output[j] - target[j]);
        
        // second mapping
        tlp2.computeOutput(input2, output);
        for (int j=0; j<noutput; ++j) gradout[j] = output[j] - target2[j];
        tlp2.batchBackPropagateAccumulate(input, output, gradout);
        for (int j=0; j<noutput; ++j) error += square(output[j] - target2[j]);

        // third mapping
        tlp2.computeOutput(input3, output);
        for (int j=0; j<noutput; ++j) gradout[j] = output[j] - target3[j];
        tlp2.batchBackPropagateAccumulate(input, output, gradout);
        for (int j=0; j<noutput; ++j) error += square(output[j] - target3[j]);

        // terminate batch backpropagation: 3 mappings
        tlp2.batchBackPropagateTerminate(3);

        // Learn for next step using gradients from batch backpropagation
        tlp2.learn(0.1);
        
        // terminate error computation
        error /= 3*2; // average on 3 mappings, of half sum of square

        cout << "Batch backprop error at step " << i << " is "<< error << endl;
    }
    
    return 0;
}

#endif // TWOLAYERPERCEPTRON_H_DEBUG_MODE




