/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef UTILITY_H
#define UTILITY_H

#include <osg/Group>
#include <osg/Vec3d>
#include <osg/Vec4>
#include <osgText/Text>

#include <iostream>

#include "EventQueueViewer.h"
#include "Logger.h"

/**
    Utility routines for use with OSG

    @author Nicolas Brodu
*/
namespace Utility {

    /** Create a line between the 2 points with the given end colors,
        and add it to the given group.
        Returns the group, a new one is created if the argument is null.
        @param p1 the line starting point
        @param p2 the line ending point
        @param c1 the line color at the starting point, or the whole line color if c2 isn't specified.
        @param c2 the line color at the ending point. Equals c1 by default.
        @param lineGroup an osg group to add this line to. Null by default, meaning a new group is created.
        @return the group in which this line was added
    */
    osg::Group *drawLine(osg::Vec3d p1, osg::Vec3d p2, osg::Vec4 c1, osg::Group *lineGroup = 0);
    osg::Group *drawLine(osg::Vec3d p1, osg::Vec3d p2, osg::Vec4 c1, osg::Vec4 c2, osg::Group *lineGroup = 0);

    /// For use in createTextHUD method
    enum TextHUDPosition {
        TOP_LEFT, TOP_RIGHT, BOTTOM_LEFT, BOTTOM_RIGHT
    };


    /** Set this to true if you also want status messages sent to cout on graphics
        mode. Default is false.
    */
    extern bool statusToConsole;

    /** Set this to true if you also want timestamp each message by prefixing it by "t=XXX: "
        where XXX is the simulator time at which the message was emitted.
        Default is false.
    */
    extern bool statusTimeStamped;

    /** Display some status text.
        Creates a text HUD if necessary in graphics mode, or print it to stdout in batch mode.
    */
    void setStatus(const std::string& status, TextHUDPosition where = BOTTOM_LEFT, bool alsoOut = statusToConsole);

    /** Show or hides the time */
    void toggleSimulationTime();
    
    /** reset position of the viewer to it's initial position */
    void resetView();

    /** Creates a keyboard handler for use in the main program.
        Right now, this is only supported in graphics mode.
        This function has no effect in batch mode, or if the viewer isn't initialized yet.
        
        It adds the following keys to the current viewer:
        - g: Increase the gravity
        - G: Decrease the gravity
        - p: Simulation pause
        - P: Simulation restart
        - n,N: Show/Hide floor normal (only if TerrainAgent terrain is non null)
        - c: cyclic world repeats, up to 3 levels. Available only if tileableNode != 0
        - t,T: start/stop performance timer
        - r,R: reset position of the viewer to it's initial position
        - space: clear status message
    */
    void createKeyboardHandler(osg::Group* tileableNode = 0, osg::Group* tileableParent = 0);

    /// Creates the viewer. Fills the Environment viewer field too
    EventQueueViewer* createViewer(const char* windowTitle = "Crogai application");

    /// main program loop. Handles both simulator and OSG, both batch and graphics modes
    void mainLoop();

    /** Initialize the random number generator with the given seed.
        
        By default, the seed is taken from system time and printed out
        so the experiment is reproducible.

        You can set here a previous seed to reproduce it.
    
        This interface allows independance from the actual RNG used,
        and/or system functions.

        The RNG used is very precisely the Mersenne twister
        implementation by Agner Fog, released under GNU GPL at
        http://www.agner.org/random/randomc.htm.

        See also Utility.cpp and mersenne.cpp for information.
    */
    void RandomInit(long seed);        // re-seed

    long RandomInit();        // re-seed & return seed

    /** Returns a random integer in the range [min, max], bounds
        included: min <= x <= max, from a uniform distribution.

        See comments in RandomInit.
    */
    int IRandom(int min, int max);       // output random integer

    /** Returns a random number between 0 (included) and 1 (excluded)
        0 <= x < 1 drawn from a uniform distribution.

        See comments in RandomInit. For the Mersene twister there is 32
        random bits in the returned number.
    */
    double Random();                     // output random float

    /** Return 32 random bits as a long */
    long BRandom();                    // output random bits

    /** Utility to get a number from a Normal distribution with
        mean 0, variance 1. Added to the other functions for this project
    */
    double NRandom();

    /** Creates a Logger */
    Logger* createLogger(const std::string& id, const std::vector<std::string>& names);
}

#endif

