/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <osgDB/ReadFile>
#include <osg/Material>
#include <osg/Geode>

#include <iostream>
#include <sstream>
#include <fstream>

#include "Environment.h"
#include "BillboardObstacle.h"
#include "GrassyTerrain.h"
#include "ExplicitAIAgent.h"
#include "Utility.h"

using namespace std;
using namespace Environment;
using namespace Utility;

Terrain* theFloor = 0;

osg::Group* metaroot = 0;

double initialTime;
static Logger *transientLogger = 0;
static int extinctionCount = 0;
static int nspecies; // initialized below

void populationUpdateCallback(int delta, int specie) {
    if (delta>0) {
        if (CrogaiAgent::populations[specie]>=2000) {
            ostringstream os;
            os << "Explosion of specie " << specie << ", t=" << simu->getTime();
            setStatus(os.str());
            // record
            (*transientLogger)[nspecies] = simu->getTime();
            // stop simu loop
            simu->stop();
        }
        return;
    }
    // so, delta negative. Check extinctions
    if (CrogaiAgent::populations[specie]!=0) return;
    
    (*transientLogger)[extinctionCount++] = simu->getTime();

    ostringstream os;
    os << "Extinction of specie " << specie << ", t=" << simu->getTime();
    setStatus(os.str());
    // stop simu loop when all extinctions have occured
    if (extinctionCount==nspecies) simu->stop();
}

void outOfTime() {
    ostringstream os;
    os << "Population dynamics reached " << simu->getTime() << " simulated time!";
    setStatus(os.str());
    if (batch) {
        cout << "Stopping batch now, won't process forever!" << endl;
        cout << "Final population counts: ";
        for (map<int, int>::iterator it = CrogaiAgent::populations.begin(); it != CrogaiAgent::populations.end(); ++it) {
            if (it != CrogaiAgent::populations.begin()) os << ", ";
            os << it->first << ":" << it->second;
        }
        (*transientLogger)[nspecies] = simu->getTime();
        simu->stop();
    }
}

int main(int argc, char *argv[])
{
    if (argc>2) batch = true;
    if (argc<2) {
        cerr << "Argument needed: parameter file." << endl;
        cerr << "It should be a plain text file, containing, in order, the values for:" << endl;
        cerr << "Agent maximum mass" << endl;
        cerr << "Agent minimum mass" << endl;
        cerr << "Agent initial mass" << endl;
        cerr << "Agent energy footprint ratio" << endl;
        cerr << "Agent maximum energy" << endl;
        cerr << "Agent initial energy" << endl;
        cerr << "Agent spore emission energy threshold" << endl;
        cerr << "Agent spore emission energy cost" << endl;
        cerr << "Agent self-maintainance energy cost, per simulator time unit" << endl;
        cerr << "Maximum=initial grass density (energy/surface)" << endl;
        cerr << "Grass growth rate (energy/surface) per simulator time unit" << endl;
        return 1;
    }

    ifstream params(argv[1]);
    if (!params.good()) {
        cerr << "Error reading parameter file: " << argv[1] << endl;
        return 2;
    }
    
    cout << "Using parameters from file: " << argv[1] << endl;

    osg::Referenced::setThreadSafeReferenceCounting(true);

    simu = Simulator::newInstance(); //new OpenThreadSimulator;// POSIX_Simulator();
    if (batch) simu->setTimeFactor(100000.0);
    else simu->setTimeFactor(1.0);

    
    // Use explicit seed instead of system time
    // TODO: use command-line argument
    //RandomInit(47);
    
    // debug code
    //statusToConsole = true;
    //simu->postAt(30.0, &exit, 0);

    if (batch) {
        statusTimeStamped = true;
        CrogaiAgent::displayScores = false;
        CrogaiAgent::displayPopulationCounter = false;
        CrogaiAgent::displayPopulationNumbers = false;
    }
    else toggleSimulationTime();
    
/// Set global parameters
    
    params >> ExplicitAIAgent::maxMass;
    cout << "Agent maximum mass: " << ExplicitAIAgent::maxMass << endl;
    params >> ExplicitAIAgent::minMass;
    cout << "Agent minimum mass: " << ExplicitAIAgent::minMass << endl;
    double d;
    params >> d;
    AgentCore::setDefaultInitialMass(d); // for the first generation
    cout << "Agent initial mass: " << d << endl;
    
    params >> d;
    AgentCore::setDefaultEnergyFootprintRatio(d);
    cout << "Agent energy footprint ratio: " << d << endl;
    
    // Energy spent by parent to create a child. Child will have initial energy.
    // this may possibly be more than initial energy. The system is dissipative, but
    // at this point, the child is created with a mass => this is where the extra energy comes from,
    // when it gets eaten. 
    //ExplicitAIAgent::spawningEnergyCost = 580.0;
    params >> d;
    AgentCore::setDefaultMaxEnergy(d);
    cout << "Agent maximum energy: " << d << endl;
    // Don't give max to begin with... let them seek it!
    params >> d;
    double initialEnergy = d;
    cout << "Agent initial energy: " << initialEnergy << endl;
    params >> d;
    ExplicitAIAgent::sporeEmissionEnergyThreshold = d;
    cout << "Agent spore emission energy threshold: " << ExplicitAIAgent::sporeEmissionEnergyThreshold << endl;
    params >> d;
    ExplicitAIAgent::sporeEmissionEnergyCost = d;
    cout << "Agent spore emission energy cost: " << ExplicitAIAgent::sporeEmissionEnergyCost << endl;
    params >> d;
    AgentCore::setDefaultSelfMaintainEnergyCost(d); // in energy per simulator time unit
    cout << "Agent self-maintainance energy cost, per simulator time unit: " << d << endl;
    ExplicitAIAgent::sporeHalfLife = 120.0;

    AgentCore::setDefaultMaxForce(20.0);
    AgentCore::setDefaultMaxSpeed(10.0);
    ExplicitAIAgent::maxGrazingSpeed = 1.0;
    gravity=16.0;

    CrogaiAgent::seeOnlyInFront = true;

    agentsIntegrationTimeStep = 0.04; // 25 fps
    agentsAITimeStep = 0.1;
    cloudIntegrationTimeStep = 0.3;

    const int nobstacles = 100;
    const int nagents = 400;
    nspecies = 5;

    const double diffusionCoef = 40.0;
    const double diffusionCutoff = 0.001;

    GrassyTerrain* grass = new GrassyTerrain(500, 100, 7, 0.35, 3.5);
    TerrainAgent::terrain = theFloor = grass;
    params >> d;
    grass->setMaximumGrassDensity(d);
    grass->setGrassDensity(d); // init density to the max
    cout << "Maximum=initial grass density (energy/surface): " << d << endl;
    params >> d;
    grass->setGrassGrowthRate(d);
    cout << "Grass growth rate (energy/surface) per simulator time unit: " << d << endl;
    grass->scheduleForGrowth(agentsAITimeStep);

    if (!batch) {
        createViewer("Environmental factors experiment");
        root = new osg::Group();
        root->addChild(theFloor);
        metaroot = new osg::Group();
        metaroot->addChild(root);
        osg::Group * wholeScene = new osg::Group();
        wholeScene->addChild(metaroot);
        createKeyboardHandler(root, metaroot);
        
        viewer->setSceneData( wholeScene );
        viewer->realize();
    }

    // create the spore diffusion cloud
    cloud = new DrawableSporeCloud(theFloor, 32, 32, 10, diffusionCoef, diffusionCutoff);
    // stable condition error already written on cerr in this case
    if (!cloud->schedule(cloudIntegrationTimeStep)) exit(1);
    
    if (!batch) {
        cloud->setDefaultAttributes("smoke.rgb", false, false);
        osg::Geode* geode = new osg::Geode();
        geode->addDrawable(cloud);
        root->addChild(geode);
    }
    
    Localized::setDefaultLocalityMode(Localized::Bricks);
    Localized::createLocalityDatabase(
        -theFloor->getHalfWorldSize(), -theFloor->getHalfWorldSize(), 0,
        theFloor->getWorldSize(), theFloor->getWorldSize(), theFloor->getMaxHeight()*2.0,
        50,50,20);

    // Create obstacles
    osg::Texture2D *texture0 = new osg::Texture2D;
    if (!batch) texture0->setImage(osgDB::readImageFile("tree.rgba"));
    osg::Texture2D *texture1 = new osg::Texture2D;
    if (!batch) texture1->setImage(osgDB::readImageFile("buisson.tga"));
    for (int i=0; i<nobstacles; ++i) {
        double x,y,z;
        
        // Position Obstacle at floor surface with bias for low altitude
        do {
            x = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
            y = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
            z = theFloor->getFloorZ(x,y);
        } while (z * Random() > theFloor->getMaxHeight() * 0.25);

        Obstacle *obs;
        switch (IRandom(0,5)) {
            case 0:
            case 1:
            case 2:
                obs = new BillboardObstacle(x, y, z, texture0,
                    2.0 + 2.0*Random(),
                    4.0 + 2.0*Random()
                );
                break;
            case 3:
            case 4:
                obs = new BillboardObstacle(x, y, z, texture1,
                    2.0 + Random(),
                    2.0 + Random()
                );
                break;
            case 5:
                obs = new Obstacle(x, y, z);
                break;
        };
        if (!batch) obs->addAvatarToGroup(root);
    }

    
// Agent code
    vector<int> preySpecies(2), predatorSpecies(2);
    for (int i=2; i<=nagents+1; ++i) { // start at 2 for specie computations, see below
        int specie = (i%nspecies) + 1;
        preySpecies[0] = ((i+1)%nspecies) + 1;
        preySpecies[1] = ((i+2)%nspecies) + 1;
        predatorSpecies[0] = ((i-1)%nspecies) + 1;
        predatorSpecies[1] = ((i-2)%nspecies) + 1;
        ExplicitAIAgent *agent = new ExplicitAIAgent(specie, predatorSpecies, preySpecies);
        double x = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
        double y = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
        // Position Agent at floor surface
        agent->setPosition(x,y, theFloor->getFloorZ(x,y));
        // give it initial energy only instead of max!
        agent->energy = initialEnergy;
        // let it move
        agent->speed = AgentCore::getDefaultMaxSpeed() * Random();
        
        if (!batch) agent->addAvatarToGroup(root);

        agent->aiEvent = simu->postCyclicAt<Agent>(
            // add jitter so all ai computations not done at the same time
            agentsAITimeStep * Random(),
            agentsAITimeStep, agent, &Agent::aiStep);

        agent->integrationEvent = simu->postCyclicAt<Agent>(
            // add jitter so all integration step computations not done at the same time
            agentsIntegrationTimeStep * Random(),
            agentsIntegrationTimeStep, agent, &Agent::integrationStep);
        
        agent->addToLocalityDatabase();
    }

    if (!batch) {
        int max = WindowWidth; if (WindowHeight>max) max = WindowHeight;
        CrogaiAgent::samplePopulationHistory((int)(max * 0.9), nspecies);
    }
    
    CrogaiAgent::createPopulationsLogger("populations",nspecies);
    CrogaiAgent::createScoresLogger("scores",nspecies);
    CrogaiAgent::createStatisticsLogger("statistics",nspecies);
    ExplicitAIAgent::createGenomesLogger("genomes",nspecies);

    CrogaiAgent::addPopulationUpdateCallback(&populationUpdateCallback);
    initialTime = simu->getSystemTime();

    // create transientLogger
    vector<string> names; for (int i=1; i<=nspecies; ++i) {
        ostringstream os;
        os << "extinctionTime" << i;
        names.push_back(os.str());
    }
    names.push_back("explosionTime");
    transientLogger = Utility::createLogger("transientTimes",names);

    // put a time limit on the computations (in batch mode, graphics will just display nice message!)
    simu->postAt(5000, &outOfTime);
    
    mainLoop();

    // record the transients after main loop. Those were also on display status
    transientLogger->record();
    
    // Tell all agents to update their stats now
    for (list<Agent*>::iterator it = Agent::allAgents.begin(); it != Agent::allAgents.end(); ++it) {
        CrogaiAgent* crogai = dynamic_cast<CrogaiAgent*>(*it);
        if (crogai!=0) crogai->updateStatistics();
    }
    CrogaiAgent::recordStatistics();
    ExplicitAIAgent::recordGenomesStatistics();
    //CrogaiAgent::recordPopulationSamples("populationSamples");
    // avoid useless updates on agents final destruction
    CrogaiAgent::logStatistics = 0;
    
    return 0;
}



