/*
// ----------------------------------------------------------------------------
//
//
// OpenSteer -- Steering Behaviors for Autonomous Characters
//
// Copyright (c) 2002-2003, Sony Computer Entertainment America
// Original author: Craig Reynolds <craig_reynolds@playstation.sony.com>
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
// ----------------------------------------------------------------------------
*/
/* ------------------------------------------------------------------ */
/*                                                                    */
/* Locality Query facility                                            */
/*                                                                    */
/* (by Craig Reynolds, see lq.h file for documentation)               */
/*                                                                    */
/*  5-17-99: created                                                  */
/*  5-20-99: found elusive "allocate 0 bins" bug                      */
/*  5-28-99: lqMapOverAllObjectsInLocality: clipped, incremental      */
/*  6- 7-99: clean up, split off annotation stuff into debuglq.c      */
/*  6- 8-99: tried screening by sum of coords ("first mean"?) but     */
/*           it was slightly slower, moved unused code to debuglq     */
/* 10-19-99: Change lqClientObject, lqObject from: "struct x {};" to  */
/*           "typedef struct x {} x;" for EE compiler.                */
/* 12- 2-00: Make lqObject "private" using lqInternalDB.              */
/* 12- 5-00: Rename lqObject to lqDB, lqClientObject to lqClientProxy */
/* 12- 6-00: Change lqCallBackFunction from arglist of (void*) to:    */
/*           (void* clientObject, float distanceSquared, void*        */
/*           clientQueryState).  Add void* clientQueryState arg to    */
/*           lqMapOverAllObjectsInLocality and its helper functions   */
/*           lqMapOverAllObjectsInLocalityClipped and                 */
/*           lqMapOverAllOutsideObjects. Change macro                 */
/*           lqTraverseBinClientObjectList to invoke callback         */
/*           function with three arguments, add "state" to its        */
/*           arglist.  Remove extern lqDistanceSquared.               */
/* 12- 7-00: Rename lqInitClientObject to lqInitClientProxy, make     */
/*           "func" be an argument to lqTraverseBinClientObjectList,  */
/*           add comments.                                            */
/* 12- 8-00: Add lqFindNearestNeighborWithinRadius and related        */
/*           definitions: lqFindNearestHelper lqFindNearestState      */
/*           Add lqMapOverAllObjects and lqRemoveAllObjects (plus:    */
/*           lqMapOverAllObjectsInBin and lqRemoveAllObjectsInBin)    */
/*                                                                    */
/* ------------------------------------------------------------------ */
// 27-01-05: by Nicolas Brodu: converted the whole thing to use double
//           instead of float. Rational is double are better supported
//           on current CPU.


#include <stdlib.h>
// NB27-01-05: float.h also contains definitions for double, ex. DBL_MAX
#include <float.h>
#include <math.h>
#include "lq.h"

/* for debugging and graphical annotation (normally unused) */
#ifdef BOIDS_LQ_DEBUG
#include "debuglq.c"
#endif


/* ------------------------------------------------------------------ */
/* This structure represents the spatial database.  Typically one of
   these would be created, by a call to lqCreateDatabase, for a given
   application.  */


typedef struct lqInternalDB
{

    /* the origin is the super-brick corner minimum coordinates */
    double originx, originy, originz;

    /* length of the edges of the super-brick */
    double sizex, sizey, sizez;
    
    /* number of sub-brick divisions in each direction */
    int divx, divy, divz;

    /* radius of the bounding sphere of a brick. Cached, deduced from size and div x,y,z */
    double brickBoundingRadius;

    /* array of pointers, one for each bin. The idea is to keep the same performance
       as the original implementation and use only one pointer dereference (memory
       access hurts!)
     */
    lqBin** bins;

    /* The real bins */
    lqBin* binArray;

    /* extra bin for "everything else" (points outside super-brick) */
    lqBin* other;

    /* NB050227: Non-empty list allows for quiker traversal for sparse environments, or
                 for very large neighborhood queries.

       Main bottleneck is memory access (by far) according to oprofile
       => the goal is to reduce the number of access to cells which fall too far from
       the locality query. What's done is the previous version is to check only for
       cells in a volume around the query. That's good, but checks for up to
       K = [2L]_x * [2L]_y * [2L]_z cells, where L is the neighborhood query distance
       and [..]_xyz the smaller multiple of div_xyz. Actually, K may be smaller when taking
       into account the "extra" bin.
       Maintaining a non-empty list and its size N allows for checking at run-time
       whether K or N is smaller, and thus minimize the number of memory access to useless
       cells.

       Additionally, I've added in both cases a quick check to remove the cell if the distance
       is too far for all the possible objects in this cell, thus avoiding the need to run
       through all this cell elements.
    */
    lqBin* nonEmptyList;
    int nonEmptyListSize;

} lqInternalDB;


/* ------------------------------------------------------------------ */
/* Allocate and initialize an LQ database, return a pointer to it.
   The application needs to call this before using the LQ facility.
   The nine parameters define the properties of the "super-brick":
      (1) origin: coordinates of one corner of the super-brick, its
          minimum x, y and z extent.
      (2) size: the width, height and depth of the super-brick.
      (3) the number of subdivisions (sub-bricks) along each axis.
   This routine also allocates the bin array, and initialize its
   contents. */


lqInternalDB* lqCreateDatabase (double originx, double originy, double originz,
				double sizex, double sizey, double sizez,
				int divx, int divy, int divz)
{
    lqInternalDB* lq = ((lqInternalDB*) malloc (sizeof (lqInternalDB)));

    lqInitDatabase (lq,
		    originx, originy, originz,
		    sizex, sizey, sizez,
		    divx, divy, divz);
    return lq;
}


/* ------------------------------------------------------------------ */
/* Deallocate the memory used by the LQ database */


void lqDeleteDatabase(lqDB* lq)
{
    free (lq->binArray);
    free (lq->bins);
    free (lq->other);
    free (lq);
}


/* ------------------------------------------------------------------ */
/* Given an LQ database object and the nine basic parameters: fill in
   the object's slots, allocate the bin array, and initialize its
   contents. */


void lqInitDatabase (lqInternalDB* lq,
		     double originx, double originy, double originz,
		     double sizex, double sizey, double sizez,
		     int divx, int divy, int divz)
{
    int bincount = divx * divy * divz;
    int arraysize = (sizeof (lqBin) | 7) * bincount; // align to multiple of 8 boundary
    lq->binArray = (lqBin*) malloc (arraysize);
    lq->bins = (lqBin**) malloc (sizeof (lqBin*) * bincount);
    lq->other = (lqBin*) malloc (sizeof (lqBin));
    lq->other->objects = NULL;
    lq->other->nonEmptyPrev = NULL;
    lq->other->nonEmptyNext = NULL;
    lq->other->x = DBL_MAX;
    lq->other->y = DBL_MAX;
    lq->other->z = DBL_MAX;
    lq->originx = originx;
    lq->originy = originy;
    lq->originz = originz;
    lq->sizex = sizex;
    lq->sizey = sizey;
    lq->sizez = sizez;
    lq->divx = divx;
    lq->divy = divy;
    lq->divz = divz;
    lq->nonEmptyListSize = 0;
    lq->nonEmptyList = NULL;
    {
        lqBin* bin;
        int i,j,k, idx;
        double xpos, ypos, zpos;
        double stepx = sizex / (double)divx;
        double stepy = sizey / (double)divy;
        double stepz = sizez / (double)divz;
        double yposinit = originy + 0.5 * stepy;
        double zposinit = originz + 0.5 * stepz;
        double boundingRadius = 0.5 * sqrt(stepx*stepx + stepy*stepy);
        // compute cached max brick radius now
        lq->brickBoundingRadius = boundingRadius;
        boundingRadius = 0.5 * sqrt(stepx*stepx + stepz*stepz);
        if (lq->brickBoundingRadius < boundingRadius) lq->brickBoundingRadius = boundingRadius;
        boundingRadius = 0.5 * sqrt(stepy*stepy + stepz*stepz);
        if (lq->brickBoundingRadius < boundingRadius) lq->brickBoundingRadius = boundingRadius;
        
        xpos = originx + 0.5 * stepx;
        idx = 0;
        for (i=0; i<divx; ++i) {
            ypos = yposinit;
            for (j=0; j<divy; ++j) {
                zpos = zposinit;
                for (k=0; k<divz; ++k) {
                    lq->bins[idx] = NULL; // empty bin
                    bin = lq->binArray + idx++;
                    bin->nonEmptyPrev = NULL;
                    bin->nonEmptyNext = NULL;
                    bin->objects = NULL;
                    bin->x = xpos;
                    bin->y = ypos;
                    bin->z = zpos;
                    zpos += stepz;
                }
                ypos += stepy;
            }
            xpos += stepx;
        }
    }
}


/* ------------------------------------------------------------------ */
/* Determine index into linear bin array given 3D bin indices */


#define lqBinCoordsToBinIndex(lq, ix, iy, iz) \
    ((ix * (lq)->divy * (lq)->divz) + (iy * (lq)->divz) + iz)


/* ------------------------------------------------------------------ */
/* Find the bin ID for a location in space.  The location is given in
   terms of its XYZ coordinates.  The bin ID is a pointer to the bin
   contents list.  */


lqBin* lqBinForLocation (lqInternalDB* lq,
				  double x, double y, double z)
{
    int i, ix, iy, iz;

    /* if point outside super-brick, return the "other" bin */
    if (x < lq->originx)              return lq->other;
    if (y < lq->originy)              return lq->other;
    if (z < lq->originz)              return lq->other;
    if (x >= lq->originx + lq->sizex) return lq->other;
    if (y >= lq->originy + lq->sizey) return lq->other;
    if (z >= lq->originz + lq->sizez) return lq->other;

    /* if point inside super-brick, compute the bin coordinates */
    ix = (int) (((x - lq->originx) / lq->sizex) * lq->divx);
    iy = (int) (((y - lq->originy) / lq->sizey) * lq->divy);
    iz = (int) (((z - lq->originz) / lq->sizez) * lq->divz);

    /* convert to linear bin number */
    i = lqBinCoordsToBinIndex (lq, ix, iy, iz);

    /* return pointer to that bin */
    return &(lq->binArray[i]);
}


/* ------------------------------------------------------------------ */
/* The application needs to call this once on each lqClientProxy at
   setup time to initialize its list pointers and associate the proxy
   with its client object. */ 


void lqInitClientProxy (lqClientProxy* proxy, void* clientObject)
{
    proxy->prev   = NULL;
    proxy->next   = NULL;
    proxy->bin    = NULL;
    proxy->object = clientObject;
}


/* ------------------------------------------------------------------ */
/* Adds a given client object to a given bin, linking it into the bin
   contents list. */


void lqAddToBin (lqInternalDB* lq, lqClientProxy* object, lqBin* bin)
{
    /* if bin is currently empty */    
    if (bin->objects == NULL)
    {
        object->prev = NULL;
        object->next = NULL;
        bin->objects = object;

        /* If non-empty bin list is currently empty */
        if (lq->nonEmptyList == NULL) {
            bin->nonEmptyPrev = NULL;
            bin->nonEmptyNext = NULL;
            lq->nonEmptyList = bin;
        } else {
            bin->nonEmptyPrev = NULL;
            bin->nonEmptyNext = lq->nonEmptyList;
            lq->nonEmptyList->nonEmptyPrev = bin;
            lq->nonEmptyList = bin;
        }

        /* Increase list size */
        ++lq->nonEmptyListSize;

        /* update non-null array */
        if (bin != lq->other) {
            int idx = bin - lq->binArray;
            lq->bins[idx] = bin;
        }
    }
    else
    {
        /* Non-empty bin => already in bin list */
        object->prev = NULL;
        object->next = bin->objects;
        bin->objects->prev = object;
        bin->objects = object;
    }

    /* record bin ID in proxy object */
    object->bin = bin;
}


/* ------------------------------------------------------------------ */
/* Removes a given client object from its current bin, unlinking it
   from the bin contents list. */


void lqRemoveFromBin (lqInternalDB* lq, lqClientProxy* object)
{
    /* adjust pointers if object is currently in a bin */
    if (object->bin != NULL)
    {
        /* If there is a prev object, link its "next" pointer to the
           object after this one. */
        if (object->prev != NULL) object->prev->next = object->next;
        
        /* If there is a next object, link its "prev" pointer to the
           object before this one. */
        if (object->next != NULL) object->next->prev = object->prev;
        
        /* If this object is at the head of the list, move the bin
           pointer to the next item in the list (might be NULL). */
        if (object->bin->objects == object) {
            lqBin* bin = object->bin;
            bin->objects = object->next;

            /* If the bin is now empty, remove it from the non-empty
               bin list*/
            if (bin->objects == NULL) {
        
                /* same steps, on a different list... */
                if (bin->nonEmptyPrev != NULL)
                    bin->nonEmptyPrev->nonEmptyNext = bin->nonEmptyNext;

                if (bin->nonEmptyNext != NULL)
                    bin->nonEmptyNext->nonEmptyPrev = bin->nonEmptyPrev;

                if (lq->nonEmptyList == bin)
                    lq->nonEmptyList = bin->nonEmptyNext;

                /* Decrease list size */
                --lq->nonEmptyListSize;

                bin->nonEmptyPrev = NULL;
                bin->nonEmptyNext = NULL;
            
                /* update non-null array */
                if (bin != lq->other) {
                    int idx = bin - lq->binArray;
                    lq->bins[idx] = 0;
                }

            }
        }

    }

    /* Null out prev, next and bin pointers of this object. */
    object->prev = NULL;
    object->next = NULL;
    object->bin = NULL;
}


/* ------------------------------------------------------------------ */
/* Call for each client object every time its location changes.  For
   example, in an animation application, this would be called each
   frame for every moving object.  */


void lqUpdateForNewLocation  (lqInternalDB* lq, 
			      lqClientProxy* object, 
			      double x, double y, double z)
{
    /* find bin for new location */
    lqBin* newBin = lqBinForLocation (lq, x, y, z);

    /* store location in client object, for future reference */
    object->x = x;
    object->y = y;
    object->z = z;

    /* has object moved into a new bin? */
    if (newBin != object->bin)
    {
        lqRemoveFromBin (lq, object);
        lqAddToBin (lq, object, newBin);
    }
}


/* ------------------------------------------------------------------ */
/* Given a bin's list of client proxies, traverse the list and invoke
   the given lqCallBackFunction on each object that falls within the
   search radius.  */


#define lqTraverseBinClientObjectList(co, radiusSquared, func, state) \
    while (co != NULL)                                                \
    {                                                                 \
	/* compute distance (squared) from this client   */           \
	/* object to given locality sphere's centerpoint */           \
	dx = x - co->x;                                         \
	dy = y - co->y;                                         \
	dz = z - co->z;                                         \
	distanceSquared = (dx * dx) + (dy * dy) + (dz * dz);    \
                                                                      \
	/* apply function if client object within sphere */           \
	if (distanceSquared < radiusSquared)                          \
	    (*func) (co->object, distanceSquared, state);             \
                                                                      \
	/* consider next client object in bin list */                 \
	co = co->next;                                                \
    }


/* ------------------------------------------------------------------ */
/* This subroutine of lqMapOverAllObjectsInLocality efficiently
   traverses of subset of bins specified by max and min bin
   coordinates. */


void lqMapOverAllObjectsInLocalityClipped (lqInternalDB* lq, 
					   double x, double y, double z,
					   double radius,
					   lqCallBackFunction func,
					   void* clientQueryState,
					   int minBinX,
					   int minBinY, 
					   int minBinZ,
					   int maxBinX,
					   int maxBinY,
					   int maxBinZ)
{
    int i, j, k;
    int iindex, jindex, kindex;
    int slab = lq->divy * lq->divz;
    int row = lq->divz;
    int istart = minBinX * slab;
    int jstart = minBinY * row;
    int kstart = minBinZ;
    lqClientProxy* co;
    lqBin* bin;
    double dx;
    double dy;
    double dz;
    double distanceSquared;
    double radiusSquared = radius * radius;
    /* Sphere test: consider the bounding sphere of each brick. It has radius
       precomputed at database creation time. Consider Dc the distance from the
       locality query point to the center of the sphere.
       Then if Dc - brickBoundingRadius > neighborhood query radius, the whole
       brick is outside the neighborhood query sphere.
       Thus, test for Dc^2 > sphereTest with sphereTest a constant defined by
       sphereTest = (neighborhood query radius + brickBoundingRadius)^2
    */
    double sphereTest = radius + lq->brickBoundingRadius;
    sphereTest *= sphereTest;

#ifdef BOIDS_LQ_DEBUG
    if (lqAnnoteEnable) drawBallGL (x, y, z, radius);
#endif

    /* loop for x bins across diameter of cube */
    iindex = istart;
    for (i = minBinX; i <= maxBinX; i++)
    {
        /* loop for y bins across diameter of cube */
        jindex = jstart;
        for (j = minBinY; j <= maxBinY; j++)
        {
            /* loop for z bins across diameter of cube */
            kindex = kstart;
            for (k = minBinZ; k <= maxBinZ; k++)
            {
                bin = lq->bins[iindex + jindex + kindex];
                
                if (bin==0) {
                    kindex += 1;
                    continue;
                }

                /* Eliminate bricks clearly outside sphere */
                dx = bin->x - x;
                dy = bin->y - y;
                dz = bin->z - z;
                distanceSquared = (dx * dx) + (dy * dy) + (dz * dz);
                
                /* If the whole brick is outside sphere, its elements can't be in! */
                if (distanceSquared > sphereTest) {
                    kindex += 1;
                    continue;
                }

                /* get current bin's client object list */
                co = bin->objects;
                
#ifdef BOIDS_LQ_DEBUG
                if (lqAnnoteEnable) drawBin (lq, bin);
#endif
                /* traverse current bin's client object list */
                lqTraverseBinClientObjectList (co,
					   radiusSquared,
					   func,
					   clientQueryState);
                kindex += 1;
            }
            jindex += row;
        }
        iindex += slab;
    }
}


/* ------------------------------------------------------------------ */
/* If the query region (sphere) extends outside of the "super-brick"
   we need to check for objects in the catch-all "other" bin which
   holds any object which are not inside the regular sub-bricks  */


void lqMapOverAllOutsideObjects (lqInternalDB* lq, 
				 double x, double y, double z,
				 double radius,
				 lqCallBackFunction func,
				 void* clientQueryState)
{
    double dx;
    double dy;
    double dz;
    double distanceSquared;
    lqClientProxy* co = lq->other->objects;
    double radiusSquared = radius * radius;

    /* traverse the "other" bin's client object list */
    lqTraverseBinClientObjectList (co,
				   radiusSquared,
				   func,
				   clientQueryState);
}



/* ------------------------------------------------------------------ */
/* If there is a gain by mapping over the non-empty list, in terms of
   number of cells, then use it.
  */

   
void lqMapOverNonEmptyList(lqInternalDB* lq, 
                 double x, double y, double z,
                 double radius,
                 lqCallBackFunction func,
                 void* clientQueryState)
{
    double dx;
    double dy;
    double dz;
    double distanceSquared;
    lqClientProxy* co;
    lqBin* bin;
    double radiusSquared = radius * radius;
    /* Sphere test: consider the bounding sphere of each brick. It has radius
       precomputed at database creation time. Consider Dc the distance from the
       locality query point to the center of the sphere.
       Then if Dc - brickBoundingRadius > neighborhood query radius, the whole
       brick is outside the neighborhood query sphere.
       Thus, test for Dc^2 > sphereTest with sphereTest a constant defined by
       sphereTest = (neighborhood query radius + brickBoundingRadius)^2
    */
    double sphereTest = radius + lq->brickBoundingRadius;
    sphereTest *= sphereTest;


    for (bin = lq->nonEmptyList; bin!=NULL; bin = bin->nonEmptyNext) {
    
        /* Eliminate bricks clearly outside sphere
           The idea is to avoid de-referencing co = bin->objects
        */
        dx = bin->x - x;
        dy = bin->y - y;
        dz = bin->z - z;
        distanceSquared = (dx * dx) + (dy * dy) + (dz * dz);
        
        /* If the whole brick is outside sphere, its elements can't be in! */
        if (distanceSquared > sphereTest) continue;

        /* get current bin's client object list */
        co = bin->objects;
        
        /* traverse current bin's client object list */
        lqTraverseBinClientObjectList (co,
                radiusSquared,
                func,
                clientQueryState);

    }

}



/* ------------------------------------------------------------------ */
/* Apply an application-specific function to all objects in a certain
   locality.  The locality is specified as a sphere with a given
   center and radius.  All objects whose location (key-point) is
   within this sphere are identified and the function is applied to
   them.  The application-supplied function takes three arguments:

     (1) a void* pointer to an lqClientProxy's "object".
     (2) the square of the distance from the center of the search
         locality sphere (x,y,z) to object's key-point.
     (3) a void* pointer to the caller-supplied "client query state"
         object -- typically NULL, but can be used to store state
         between calls to the lqCallBackFunction.

   This routine uses the LQ database to quickly reject any objects in
   bins which do not overlap with the sphere of interest.  Incremental
   calculation of index values is used to efficiently traverse the
   bins of interest. */


void lqMapOverAllObjectsInLocality (lqInternalDB* lq, 
				    double x, double y, double z,
				    double radius,
				    lqCallBackFunction func,
				    void* clientQueryState)
{
    int partlyOut = 0;
    int completelyOutside = 
	(((x + radius) < lq->originx) ||
	 ((y + radius) < lq->originy) ||
	 ((z + radius) < lq->originz) ||
	 ((x - radius) >= lq->originx + lq->sizex) ||
	 ((y - radius) >= lq->originy + lq->sizey) ||
	 ((z - radius) >= lq->originz + lq->sizez));
    int minBinX, minBinY, minBinZ, maxBinX, maxBinY, maxBinZ;

    /* is the sphere completely outside the "super brick"? */
    if (completelyOutside)
    {
	lqMapOverAllOutsideObjects (lq, x, y, z, radius, func,
				    clientQueryState);
	return;
    }

    /* compute min and max bin coordinates for each dimension */
    minBinX = (int) ((((x - radius) - lq->originx) / lq->sizex) * lq->divx);
    minBinY = (int) ((((y - radius) - lq->originy) / lq->sizey) * lq->divy);
    minBinZ = (int) ((((z - radius) - lq->originz) / lq->sizez) * lq->divz);
    maxBinX = (int) ((((x + radius) - lq->originx) / lq->sizex) * lq->divx);
    maxBinY = (int) ((((y + radius) - lq->originy) / lq->sizey) * lq->divy);
    maxBinZ = (int) ((((z + radius) - lq->originz) / lq->sizez) * lq->divz);

    /* clip bin coordinates */
    if (minBinX < 0)         {partlyOut = 1; minBinX = 0;}
    if (minBinY < 0)         {partlyOut = 1; minBinY = 0;}
    if (minBinZ < 0)         {partlyOut = 1; minBinZ = 0;}
    if (maxBinX >= lq->divx) {partlyOut = 1; maxBinX = lq->divx - 1;}
    if (maxBinY >= lq->divy) {partlyOut = 1; maxBinY = lq->divy - 1;}
    if (maxBinZ >= lq->divz) {partlyOut = 1; maxBinZ = lq->divz - 1;}


    /* Apply non-empty list traversal if there is a gain in number of cells */
    if (lq->nonEmptyListSize < (maxBinX - minBinX + 1) * (maxBinY - minBinY + 1) * (maxBinZ - minBinZ + 1)) {
        lqMapOverNonEmptyList(lq, x, y, z, radius, func, clientQueryState);
        return;
    }

    
    /* map function over outside objects if necessary (if clipped) */
    if (partlyOut) 
	lqMapOverAllOutsideObjects (lq, x, y, z, radius, func,
				    clientQueryState);
    
    /* map function over objects in bins */
    lqMapOverAllObjectsInLocalityClipped (lq,
					  x, y, z,
					  radius,
					  func,
					  clientQueryState,
					  minBinX, minBinY, minBinZ,
					  maxBinX, maxBinY, maxBinZ);
}


/* ------------------------------------------------------------------ */
/* internal helper function */


typedef struct lqFindNearestState
{
    void* ignoreObject;
    void* nearestObject;
    double minDistanceSquared;

} lqFindNearestState;


void lqFindNearestHelper (void* clientObject,
			  double distanceSquared,
			  void* clientQueryState)
{
    lqFindNearestState* fns = (lqFindNearestState*) clientQueryState;

    /* do nothing if this is the "ignoreObject" */
    if (fns->ignoreObject != clientObject)
    {
	/* record this object if it is the nearest one so far */
	if (fns->minDistanceSquared > distanceSquared)
	{
	    fns->nearestObject = clientObject;
	    fns->minDistanceSquared = distanceSquared;
	}
    }
}


/* ------------------------------------------------------------------ */
/* Search the database to find the object whose key-point is nearest
   to a given location yet within a given radius.  That is, it finds
   the object (if any) within a given search sphere which is nearest
   to the sphere's center.  The ignoreObject argument can be used to
   exclude an object from consideration (or it can be NULL).  This is
   useful when looking for the nearest neighbor of an object in the
   database, since otherwise it would be its own nearest neighbor.
   The function returns a void* pointer to the nearest object, or
   NULL if none is found.  */


void* lqFindNearestNeighborWithinRadius (lqInternalDB* lq, 
					 double x, double y, double z,
					 double radius,
					 void* ignoreObject)
{
    /* initialize search state */
    lqFindNearestState lqFNS;
    lqFNS.nearestObject = NULL;
    lqFNS.ignoreObject = ignoreObject;
    lqFNS.minDistanceSquared = DBL_MAX;

    /* map search helper function over all objects within radius */
    lqMapOverAllObjectsInLocality (lq, 
				   x, y, z,
				   radius,
				   lqFindNearestHelper,
				   &lqFNS);

    /* return nearest object found, if any */
    return lqFNS.nearestObject;
}


/* ------------------------------------------------------------------ */
/* internal helper function */


void lqMapOverAllObjectsInBin (lqClientProxy* binProxyList, 
			       lqCallBackFunction func,
			       void* clientQueryState)
{
    /* walk down proxy list, applying call-back function to each one */
    while (binProxyList != NULL)
    {
	(*func) (binProxyList->object, 0, clientQueryState);
	binProxyList = binProxyList->next;
    }
}


/* ------------------------------------------------------------------ */
/* Apply a user-supplied function to all objects in the database,
   regardless of locality (cf lqMapOverAllObjectsInLocality) */

void lqMapOverAllObjects (lqInternalDB* lq, 
			  lqCallBackFunction func,
			  void* clientQueryState)
{
    int i;
    int bincount = lq->divx * lq->divy * lq->divz;
    for (i=0; i<bincount; i++)
    {
	lqMapOverAllObjectsInBin (lq->binArray[i].objects, func, clientQueryState);
    }
    lqMapOverAllObjectsInBin (lq->other->objects, func, clientQueryState);
}


/* ------------------------------------------------------------------ */
/* internal helper function */


#define lqRemoveAllObjectsInBin(lq, bin) \
    while ((bin)->objects != NULL) lqRemoveFromBin (lq, (bin)->objects);


/* ------------------------------------------------------------------ */
/* Removes (all proxies for) all objects from all bins */


void lqRemoveAllObjects (lqInternalDB* lq)
{
    int i;
    int bincount = lq->divx * lq->divy * lq->divz;
    for (i=0; i<bincount; i++)
    {
        lqRemoveAllObjectsInBin (lq, &lq->binArray[i]);
        lq->bins[i] = NULL;
    }
    lqRemoveAllObjectsInBin (lq, lq->other);
}


/* ------------------------------------------------------------------ */
