/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Test case with a swarm of pseudo-boids

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <osg/ShapeDrawable>
#include <osgProducer/Viewer>
#include <osgGA/TrackballManipulator>

#include <iostream>

#include <time.h>
#include <unistd.h>
#include <getopt.h>

#include "Agent.h"
#include "Environment.h"
#include "Utility.h"
using namespace Environment;
using namespace std;
using namespace Utility;

struct SwarmAgent : public Agent {

    SwarmAgent() : color(1.0f,1.0f,1.0f,1.0f) {}
    
    osg::Vec4 color;

    static osg::Shape *commonShape;
    static double centerAttraction;
    static bool collisionDetection;
    
    // Code based on OSG tutorial, plus shape sharing
    osg::Node * buildAvatar()
    {
        if (!commonShape) commonShape = new osg::Box(osg::Vec3(0,0,0),1,2,0.5);
        
        // create a container that makes the cube drawable
        osg::ShapeDrawable *sdCube = new osg::ShapeDrawable(commonShape);
        
        sdCube->setColor(color);
    
        // create a geode object as a container for our drawable cube object
        osg::Geode* geodeCube = new osg::Geode();
        
        // add our drawable cube to the geode container
        geodeCube->addDrawable(sdCube);

        return geodeCube;
    }
    
    void updateAI(double elapsedTime, double currentTime) {
        steeringForce.set(0,0,0);
        osg::Vec3d localCenter(0,0,0);
        osg::Vec3d forwardForce(0,0,0);
        
        int n = 0; // not the same as multimap size, since objects behind are not seen
        findNeighbors(30.0);
        for (multimap<const double,Localized*>::iterator it = neighbors->begin(); it != neighbors->end(); ++it) {

                Agent* neighbor = dynamic_cast<Agent*>(it->second);
                if (!neighbor) continue; // not an agent. Should never happen in this simple example

                if ((neighbor->getPosition() - getPosition()) * forward < 0)
                    continue; // see only what's in front of us
                
                steeringForce += steerToAvoid(neighbor);
                forwardForce += neighbor->forward;
                localCenter += neighbor->getPosition();
                ++n;
        }

        // don't move to the origin when no other swagent present
        // In this case, move at random, toward 0 preferentially
        if (n==0) {
            steeringForce += steerForWander(0.4);
            steeringForce += steerForSeek(localCenter) * centerAttraction;
        }

        // Otherwise, swagent behavior:
        // - steer to avoid neighbors
        // - steer toward the local center
        // - and also to keep the local group direction
        else {
            steeringForce /= (double)n; // average of avoidance vectors
            steeringForce += steerForSeek(localCenter/(double)n) * 2.0; // local center has more weight
            steeringForce += forwardForce /(double)n  -  forward; // local group direction
        }

    }

    // Handle collision detection
    void updateIntegration(const double elapsedTime, const double currentTime)
    {
        // May be de-activated
        if (!collisionDetection) return;
        
        // Ad-hoc hack to make collisions effective & more efficient
        const double radius22 = 2.0 * radius * radius;

        // Collision detection. Use current neighbor list
        for (multimap<const double,Localized*>::iterator itn = neighbors->begin(); itn != neighbors->end(); ++itn) {
            if (itn->second == this) continue; // should not happen, but...
            Agent* agent = dynamic_cast<Agent*>(itn->second);
            if (agent==0) continue; // should not happen either
            
            // Collision case
            if (wrappedDist2(itn->second->getPosition(), newPosition) <= radius22) {

                // Bounce back!
                osg::Vec3d collisionForward = wrappedDiffVector(newPosition, agent->getPosition());

                // Add previous directions for consistency, but not so much
                // as to completely prevent the agent from separating from each other
                collisionForward += (forward + agent->forward) * 0.25;

                // Change direction
                if (collisionForward.normalize()==0)
                    // Bounce back if exactly at the same place of colliding target
                    collisionForward = -newForward; 
        
                // Attitude quaternion must be updated since we change direction
                attitudeCorrection->updateAttitude(newAttitude, newAttitude, collisionForward, newForward);
        
                // Now update our field
                newForward = collisionForward;
                
                // One collision is enough to handle
                return;
            }
        }
    
    }
    
};


osg::Shape *SwarmAgent::commonShape = 0;

static struct option long_options[] = {
    {"number", 1, 0, 'n'},
    {"istep", 1, 0, 'i'},
    {"aistep", 1, 0, 'a'},
    {"time-ratio", 1, 0, 't'},
    {"max-force", 1, 0, 'f'},
    {"max-speed", 1, 0, 'd'},
    {"center-force", 1, 0, 'c'},
    {"random-seed", 1, 0, 'r'},
    {"locality", 1, 0, 'l'},
    {"brick-proba", 1, 0, 'p'},
    {"nbricks", 1, 0, 'b'},
    {"world-size", 1, 0, 's'},
    {"collision", 1, 0, 'o'},
    {"simulator", 1, 0, 'u'},
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
};

const int defaultNumber = 1000;
const double defaultIStep = 0.04; // 25 FPS
const double defaultAIStep = 0.15;
const double defaultTimeRatio = 1.0;
const double defaultMaxForce = 50;
const double defaultMaxSpeed = 30;
const double defaultBrickProba = 0.25;
const int defaultBricksNumber = 100;
const double defaultWorldSize = 400.0;
const Localized::LocalityMode defaultLocality = Localized::Mixed;
const double defaultCenterAttraction = 0.005;
const bool defaultCollisionDetection = true;

double SwarmAgent::centerAttraction = defaultCenterAttraction;
bool SwarmAgent::collisionDetection = defaultCollisionDetection;

void help(int argc, char *argv[]) {
cout <<
"Usage: " << argv[0] << " [options]" << endl << endl << 

"Simulate a swarm of boids-like agents." << endl << endl <<

"Options:" << endl << 
"  -n, --number=INTEGER       Set the number of agents. Default is " << defaultNumber << endl <<
"  -i, --istep=FLOAT          Set the integration step cycle, in seconds. Default is " << defaultIStep << endl <<
"  -a, --aistep=FLOAT         Set the AI step cycle, in seconds. Default is " << defaultAIStep << endl <<
"  -t, --time-ratio=FLOAT     Set the simulation time / real time ratio. Default is " << defaultTimeRatio << endl <<
"  -f, --max-force=FLOAT      Set the maximum force the swarm agents can use to change direction. Default is " << defaultMaxForce << endl <<
"  -d, --max-speed=FLOAT      Set the maximum speed the swarm agents can reach. Default is " << defaultMaxSpeed << endl <<
"  -c, --center-force=FLOAT   Set the force toward center to apply when an agent becomes isolated. Default is " << defaultCenterAttraction << endl <<
"  -r, --random-seed=INTEGER  Use this random number generator seed instead of the current system time." << endl <<
"  -l, --locality=STRING      Set the neighbor finding algorithm to one of the following:" << endl <<
"                             . 'brute': brute-checking over all localized objects. O(n^2)" << endl <<
"                             . 'bricks': split world in bricks to quickly reject far bricks: ~O(n)" << endl <<
"                             . 'propagate': maintain and propagate a list of neighbors. Faster but less accurate than 'bricks': isolated individuals are lost, two groups may cross without seeing each other." << endl <<
"                             . 'mixed': put some objects in bricks to complement the list of neighbors. This is the default, and corrects the 'propagate' mode problems after a few steps. Consistency depends on the proportion of objects in bricks, see the 'bricks-proba' option." << endl <<
"  -p, --brick-proba=FLOAT    In mixed mode, each object has this probability to be inserted in a brick. Default is " << defaultBrickProba << endl <<
"  -b, --nbricks=INTEGER      Set the number of bricks along each dimension. Default is " << defaultBricksNumber << endl <<
"  -s, --world-size=FLOAT     Set the world size to split in bricks. Default is " << defaultWorldSize << endl <<
"  -o, --collision=BOOL       Whether collision detection should be computed. Default is " <<
(defaultCollisionDetection? "true" : "false") << endl <<
"  -u, --simulator=STRING     Use a specific simulator implementation ID. Default is autodetection." << endl <<
"  -h, --help                 Display this help message and exit." << endl <<

endl<<
"Here are some standard Open Scene Graph keys to use during the simulation. More keys may exist depending on your current OSG version:" << endl <<
"  Escape: Exit program." << endl <<
"  'f': Fullscreen ON/OFF." << endl <<
"  'w': Wireframe mode." << endl <<
"  Space: Position the view out of the swarm thanks to trackball mode." << endl <<
        

endl << "Author: Nicolas Brodu <nicolas.brodu@free.fr>. The locality bricks database code was re-used from Craig Reynolds OpenSteer library and extended. The graphics framework is provided by the Open Scene Graph project. The Mersenne Twister random number generator was created by M. Matsumoto & T. Nishimura, and implemented by A. Fog. Please report bugs, remarks, suggestions..." << endl;
}

int main(int argc, char *argv[])
{
    
    int number = defaultNumber;
    double istep = defaultIStep;
    double aistep = defaultAIStep;
    double timeRatio = defaultTimeRatio;
    double maxForce = defaultMaxForce;
    double maxSpeed = defaultMaxSpeed;
    double brickProba = defaultBrickProba;
    int bricksNumber = defaultBricksNumber;
    double worldSize = defaultWorldSize;
    Localized::LocalityMode locality = defaultLocality;
    SwarmAgent::centerAttraction = defaultCenterAttraction;
    SwarmAgent::collisionDetection = defaultCollisionDetection;

    unsigned int seed = time(0);
    const char* simulatorImplementationID = 0;
    
    int c;
    while (1) {
        c = getopt_long (argc, argv, "n:i:a:t:f:d:c:r:l:p:b:s:o:u:h", long_options, 0);
        
        if (c==-1) break;

        switch(c) {
            case 'n':
                number = atoi(optarg);
                if (number<=0) number = defaultNumber;
                break;
            case 'i':
                istep = atof(optarg);
                if (istep<=0.0) istep = defaultIStep;
                break;
            case 'a':
                aistep = atof(optarg);
                if (aistep<=0.0) aistep = defaultIStep;
                break;
            case 't':
                timeRatio = atof(optarg);
                if (timeRatio<=0.0) timeRatio = defaultIStep;
                break;
            case 'f':
                maxForce = atof(optarg);
                if (maxForce<=0.0) maxForce = defaultMaxForce;
                break;
            case 'd':
                maxSpeed = atof(optarg);
                if (maxSpeed<=0.0) maxSpeed = defaultMaxSpeed;
                break;
            case 'c':
                SwarmAgent::centerAttraction = atof(optarg);
                if (SwarmAgent::centerAttraction<=0.0) SwarmAgent::centerAttraction = defaultCenterAttraction;
                break;
            case 'r':
                seed = atoi(optarg);
                if (seed==0) seed = time(0);
                break;
            case 'l':
                if (!strcasecmp(optarg,"brute"))
                    locality = Localized::Brute;
                else if (!strcasecmp(optarg,"bricks"))
                    locality = Localized::Bricks;
                else if (!strcasecmp(optarg,"propagate"))
                    locality = Localized::NPropagation;
                else if (!strcasecmp(optarg,"mixed"))
                    locality = Localized::Mixed;
                else locality = defaultLocality;
                break;
            case 'p':
                brickProba = atof(optarg);
                if (brickProba<0.0) brickProba = defaultBrickProba;
                break;
            case 'b':
                bricksNumber = atoi(optarg);
                if (bricksNumber<=0) bricksNumber = defaultBricksNumber;
                break;
            case 's':
                worldSize = atof(optarg);
                if (worldSize<0.0) worldSize = defaultWorldSize;
                break;
            case 'o':
                if (!strcasecmp(optarg,"true"))
                    SwarmAgent::collisionDetection = true;
                else
                    SwarmAgent::collisionDetection = false;
                break;
            case 'u':
                simulatorImplementationID = optarg;
                break;
            case 'h':
            case '?':
                help(argc, argv);
                return 0;
        }
    }

    cout << "Creating a swarm with the following parameters:" << endl <<
    "number of agents = "<< number << endl <<
    "integration step = " << istep << endl <<
    "AI step = " << aistep << endl <<
    "time ratio = " << timeRatio << endl <<
    "max force = " << maxForce << endl <<
    "max speed = " << maxSpeed << endl <<
    "center attraction = " << SwarmAgent::centerAttraction << endl <<
    "random seed = " << seed << endl <<
    "collision detection = " << (SwarmAgent::collisionDetection? "true" : "false") << endl <<
    "locality = ";
    switch (locality) {
        case Localized::Brute:
            cout << "brute" << endl;
            cout << "world size = " << worldSize << endl;
            break;
        case Localized::Bricks:
            cout << "bricks" << endl;
            cout << "number of bricks = " << bricksNumber << endl;
            cout << "world size = " << worldSize << endl;
            break;
        case Localized::NPropagation:
            cout << "propagate" << endl;
            break;
        case Localized::Mixed:
            cout << "mixed" << endl;
            cout << "brick probability = " << brickProba << endl;
            cout << "number of bricks = " << bricksNumber << endl;
            cout << "world size = " << worldSize << endl;
            break;
    };
    cout << endl << "See the --help option for parameters signification." << endl << endl <<
"Here are some standard Open Scene Graph keys to use during the simulation. More keys may exist depending on your current OSG version:" << endl <<
". Escape: Exit program." << endl <<
". 'f': Fullscreen ON/OFF." << endl <<
". 'w': Wireframe mode." << endl <<
". Space: Position the view out of the swarm thanks to trackball mode." << endl;


    osg::Referenced::setThreadSafeReferenceCounting(true);

    RandomInit(seed);

    simu = Simulator::newInstance(simulatorImplementationID); //new OpenThreadSimulator;// POSIX_Simulator();
    simu->setTimeFactor(timeRatio);

    
    osg::Group* root = new osg::Group();

    createViewer("Swarm");
    viewer->setSceneData( root );
    viewer->realize();
    createKeyboardHandler();

    // Create a non-cyclic brick database of the required size.
    // Make it symmetric in all dimensions, this example is just a big swarm around the origin
    if ((locality==Localized::Bricks) || (locality==Localized::Mixed))
        Localized::createLocalityDatabase(
            -worldSize*0.5,-worldSize*0.5,-worldSize*0.5, // origin X,Y,Z coordinates
            worldSize,worldSize,worldSize,                // size along X,Y,Z
            bricksNumber,bricksNumber,bricksNumber,       // number of divisions in X,Y,Z
            false                                         // world is non cyclic
        );
    
    Localized::setDefaultLocalityMode(locality);
    
    AgentCore::setDefaultMaxForce(maxForce);
    AgentCore::setDefaultMaxSpeed(maxSpeed);

    // Create all agents at a random initial position    
    for (int i=0; i<number; ++i) {
    
        SwarmAgent *swagent = new SwarmAgent();

        // Position the new agent randomly in the cube
        swagent->setPosition(osg::Vec3d(
            worldSize*Random() - worldSize*0.5,
            worldSize*Random() - worldSize*0.5,
            worldSize*Random() - worldSize*0.5)
        );

        // Give them a random color, but not too dark
        swagent->color.set(
            0.7 * Random() + 0.3,
            0.7 * Random() + 0.3,
            0.7 * Random() + 0.3,
            1
        );
        
        // Add the agent graphical representation to the scene graph root
        swagent->addAvatarToGroup(root);

        // Post the AI routine of the agent        
        // Add jitter to the cycle start so all AI computations are not done at the same time,
        // then nothing for a whole cycle, which would be a waste!
        simu->postCyclicAt<Agent>(
            aistep * Random(),
            aistep, swagent, &Agent::aiStep);
     
        simu->postCyclicAt<Agent>(
            // add jitter so all integration step computations not done at the same time
            istep * Random(),
            istep, swagent, &Agent::integrationStep);

        // In mixed mode, add the swagents to the locality database
        // with the requested probability.
        if (locality==Localized::Mixed) {
            if (Random() < brickProba) swagent->addToLocalityDatabase();
        }
        // Of course, all agents are unconditionally in the database in bricks mode
        else if (locality==Localized::Bricks) {
            swagent->addToLocalityDatabase();
        }
        
    }
    
    // Post the integration step routine of all the agents
    // Unlike the AI routine, this one is supposed to be fast enough and come back often enough
    // so that posting only one event and call all agent steps in turn is actually faster than posting
    // all agents individually. This may not be true of course, especially if the agent was
    // providing its own integration step routine, but this is not the case in this example.
    //simu->postCyclic(istep, &Agent::globalIntegrationStep);
    
    // Provide an initial list of neighbors to all swagents in corresponding locality modes
    if ((locality==Localized::Mixed) || (locality==Localized::NPropagation)) {
        for (list<Localized*>::iterator it = Localized::allObjects.begin(); it != Localized::allObjects.end(); ++it) {
            (*it)->findNeighbors(30.0);
        }
        // Second order search
        for (list<Localized*>::iterator it = Localized::allObjects.begin(); it != Localized::allObjects.end(); ++it) {
            (*it)->findNeighbors(30.0);
        }
    }

    // Force OSG to initialize so the trackball can compute a view to encompass the swarm
    viewer->sync(); viewer->update();
    resetView();

    mainLoop();

    return 0;
}



