/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <fstream>
#include <iostream>

#include <math.h>
#include <string.h>

#include "waveNoise.h"

#include "Utility.h"

using namespace std;

// Daubechies W6 coefficients for reconstruction
static double lcoef[6] = {
    0.332670552950,
    0.806891509311,
    0.459877502118,
    -0.135011020010,
    -0.085441273882,
    0.035226291882
};
static double hcoef[6] = {
    0.035226291882,
    0.085441273882,
    -0.135011020010,
    -0.459877502118,
    0.806891509311,
    -0.332670552950
};
static const int ncoef = 6;

static const double sqrt2 = sqrt(2.0);
static const double sqrt2inv = sqrt2 * 0.5;

// transform the data
// optimizing this function somewhat wouldn't hurt
static double* waveletTransform(double *data, double *result, const unsigned int n, const int  dinc = 1)
{
    unsigned int nmax = n+ncoef;
    double workVector[nmax];
    // Copy data
    if (dinc==1) memcpy(workVector,data,n*sizeof(double));
    else {
        const double* end = data + n*dinc;
        double *wp = workVector;
        for (double* d = data; d<end; d+=dinc) *(wp++) = *d;
    }
    // Consider it periodic.
    // Note: Use workVector for the case ncoef>n
    // and do not use memcopy in case it overlaps
    double *w=workVector+n, *d=workVector;
    for (int i=n; i<nmax; i++) *(w++)=*(d++);
    nmax=n>>1;
    double *resultL=result, *resultH=result+nmax*dinc;
    for (unsigned int i=0; i<nmax; i++) {
        double l=0, h=0;
        double *index=workVector+(i<<1);
        double *lcp=lcoef, *hcp=hcoef;
        for (int j=0; j<ncoef; j++) {
            double value=*(index++);
            l+=value*(*(lcp++));
            h+=value*(*(hcp++));
        }
        *resultL = l * sqrt2inv;
        *resultH = h;
        resultL += dinc; resultH += dinc;
    }
    return result;
}

static double* waveletReconstruct(double *data, double *result, const unsigned int n, const int  dinc = 1)
{
    unsigned int nmax = n+ncoef;
    double workVector[nmax];
    // Initialize workVector
    memset(workVector,0,nmax*sizeof(double));
    // Now go through the matrix column by column and add the products
    // to each workVector cell so that by the end the matrix product is OK.
    // Also do sum low and high coefs at the same time to reconstruct the data.
    unsigned int nmid=n/2;
    double *dlp=data, *dhp=data+nmid*dinc;
    double *w=workVector;
    for (int i=0; i<nmid; i++) {
        double l=(*dlp) * sqrt2, h=*dhp;
        dlp += dinc; dhp+=dinc;
        double *ww=w, *lcp=lcoef, *hcp=hcoef;
        for (int j=0; j<ncoef; j++) *(ww++)+=(*(lcp++))*l+(*(hcp++))*h;
        w+=2;
    }
    w=workVector;
    double *w2=workVector+n;
    // wrap-around prevents boundary effect by considering the data periodic
    for (int i=n; i<nmax; i++) *(w++)+=*(w2++);
    // copy to result
    if (dinc==1) memcpy(result,workVector,n*sizeof(double));
    else {
        double *end = result + n*dinc;
        double *wp = workVector;
        for (double *r = result; r<end; r+=dinc) *r = *(wp++);
    }
    return result;
}


inline int coordWrapup(int x, int y, int n) {
    if (x<0) x+=n;
    x %= n;
    if (y<0) y+=n;
    y %= n;
    return y*n + x;
}

/**
    takes an array of size 2^n * 2^n to fill with noise.
    Ex: double[64][64] for n=6
*/
void waveletNoise(double* data, const int n, const double* frequencyFactors, const double power) {

    const int size = 1 << n;
    double *dp = data;
    data[0] = 0.0; // data will be rescaled, so this base value is irrelevant

    for (int level=0; level<n; ++level) {
        // Add random values to low-frequency coefficients
        // Renormalization is done in the reconstruction function
        int boundary = 1 << level;
        double f = frequencyFactors[level];

        // Add smoothed noise function to avoid square blocks
        double *noise = new double[boundary*boundary];
        double *np = noise;
        dp = data;
        for (int i=0; i<boundary; ++i) {
            for (int j=0; j<boundary; ++j) {
                *(np++) = *(dp+j) + f * Utility::Random();
            }
            dp+=size;
        }
        dp = data;
        for (int i=0; i<boundary; ++i) {
            for (int j=0; j<boundary; ++j) {
                double corners = (
                      noise[coordWrapup(j+1, i-1, boundary)]
                    + noise[coordWrapup(j-1, i-1, boundary)]
                    + noise[coordWrapup(j-1, i+1, boundary)]
                    + noise[coordWrapup(j+1, i+1, boundary)]
                    ) * 0.0625; // 1/16
                double sides = (
                      noise[coordWrapup(j-1, i, boundary)]
                    + noise[coordWrapup(j+1, i, boundary)]
                    + noise[coordWrapup(j, i-1, boundary)]
                    + noise[coordWrapup(j, i+1, boundary)]
                    ) * 0.125; // 1/8
                *(dp+j) =
                      noise[coordWrapup(j, i, boundary)] * 0.25
                    + corners
                    + sides;
            }
            dp+=size;
        }

        delete[] noise;

        boundary <<= 1;
        // Expand these random values to next level
        dp = data;
        for (int i=0; i<boundary; ++i) {
            waveletReconstruct(dp, dp, boundary, size);
            dp++;
        }
        dp = data;
        for (int i=0; i<boundary; ++i) {
            waveletReconstruct(dp, dp, boundary);
            dp += size;
        }
    }
        
    const int datasize = size*size;

    double *dpend = data + datasize;
    
    // Rescale to 0..1
    double minWN = 1e80;
    double maxWN = -1e80;
    
    for (dp = data; dp<dpend; ++dp) {
        if (*dp < minWN) minWN = *dp;
        if (*dp > maxWN) maxWN = *dp;
    }
    double renorm = maxWN - minWN;
    if (renorm!=0) renorm = 1.0 / renorm;

    // Renormalize to 0..1
    for (dp = data; dp<dpend; ++dp) {
        *dp = (*dp - minWN) * renorm;
    }

    // enhance data differences
    if (power!=1.0) {
        for (dp = data; dp<dpend; ++dp) *dp = pow(*dp, power);
    }


}


void waveletNoise(double* data, const int n, const double persistence, const double power) {
    double frequencyFactors[n];
    double fpower = 1.0;
    for (int i=0; i<n; ++i) {
        frequencyFactors[i] = fpower;
        fpower *= persistence;
    }
    waveletNoise(data, n, frequencyFactors, power);
}




void writePGM(double *data, const unsigned int size, const char* name) {
    
    ofstream f(name);
    f << "P2" << endl;
    f << size << " " << size << endl;
    f << 65536 << endl;
    
    const int datasize = size*size;
    double *dpend = data + datasize;
    
    for (double *dp = data; dp<dpend; ++dp)
        f << (int)((*dp)*65536+0.5) << endl;

    f.close();
}

/*
// Test routine
int main(int argc, char** argv) {
    const int dlen = 8*8;
    double data[dlen];
    for (int i=0; i<dlen; ++i) data[i] = i;
    for (int i=0; i<8; ++i) waveletTransform(data+i*8, data+i*8, 8);
    for (int i=0; i<8; ++i) waveletTransform(data+i, data+i, 8, 8);
    for (int i=0; i<dlen; ++i) cout << data[i] << " ";
    cout << endl;
    for (int i=0; i<8; ++i) waveletReconstruct(data+i, data+i, 8, 8);
    for (int i=0; i<8; ++i) waveletReconstruct(data+i*8, data+i*8, 8);
    for (int i=0; i<dlen; ++i) cout << data[i] << " ";
    cout << endl;
    return 0;
}
*/
