/***************************************************************************
 *                                                                         *
 *   These routines work like Perlin Noise, but a Daubechies W6 wavelet    *
 *   reconstruction is used as the interpolating function.                 *
 *   Well, not quite, since a little blur is applied at each               *
 *   reconstruction step to smooth out the blocky wavelet artifacts.       *
 *                                                                         *
 *   The idea is just to do as in Perlin noise, but interpolating with     *
 *   wavelets for fun :)                                                   *
 *                                                                         *
 *   The data is also considered cyclic for the wavelet and blur filter.   *
 *   This means the generated terrain is tileable :)                       *
 *                                                                         *
 *   Disclamer: The wavelet transform routines there are certainly not the *
 *   best that can be found on the net, but were created from scratch just *
 *   for this purpose. They were designed for the W6 nice features: same   *
 *   low & high filters lenghts, same coefficients for reconstruction.     *
 *                                                                         *
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef CROGAI_WAVELET_NOISE
#define CROGAI_WAVELET_NOISE

/**
    Fill the data array of size 2^n * 2^n with tileable noise.
    @param data The data array to fill with noise. The returned data is rescaled to
                be in the [0..1] range
    @param n The 2-exponent for the size of the array. Ex: double[64*64] for n=6
    @param persistence A measure of smoothness. 0.0 is flat, 1.0 is jagged. The default of
                       0.3 produces nice hills.
    @param power Raise the produced data by this power to enhance the low and high values.
                 1.0 causes no computation at all.
                 The default is 3.0, allowing for plains without making the hills to sharp.

    Note: Utility::Random function is used so you may control the noise seed
*/
void waveletNoise(double* data, const int n, const double persistence=0.3, const double power=3.0);

/**
    Fill the data array of size 2^n * 2^n with tileable noise
    @param data The data array to fill with noise. The returned data is rescaled to
                be in the [0..1] range
    @param n The 2-exponent for the size of the array. Ex: double[64*64] for n=6
    @param frequencyFactors The spacial frequency distribution. This is an array of size
           n, with f[0] being the lowest frequency. The noise produced at each level is
           scaled by the corresponding factor. Default is persistence^level.
    @param power Raise the produced data by this power to enhance the low and high values.
                 1.0 causes no computation at all.
                 The default is 3.0, allowing for plains without making the hills to sharp.
    
    Note: Utility::Random function is used so you may control the noise seed
*/
void waveletNoise(double* data, const int n, const double* frequencyFactors, const double power=3.0);

/** Write the data array of the given size*size in a file
    @param data the data array to write
    @param size the size of the array. data is a double[size*size] array.
    @param name The file name
*/
void writePGM(double *data, const unsigned int size, const char* name);

#endif
