#!/usr/bin/perl

sub help() {
    print "Missing or bad arguments: argument: genomes_XXX.txt [genomes_YYY.txt ...]\n";
    print "\n";
    print "This script should be run on the genomes log files. It uses the corresponding populations file (same random seed) to check on which species went extinct, and monitor the amount of specialization toward grazing/hunting before extinction.\n";
    exit(1);
};

if ($#ARGV < 0) {
    help();
}

$meanIndicator = 0;
$varIndicator = 0;

$totalExtinctionNumber = 0;

# loop over file names
foreach $fileName (@ARGV) {

    if (!($fileName =~ /.*genomes.*/)) {
        print "Incorrect file name: $fileName\n";
        help();
    }
    
    open(FILE,"<$fileName") or die "Cannot open $fileName for reading.\n";
    
    $popFile = $fileName;
    $popFile =~ s/genomes/populations/g;
    open(POPFILE,"<$popFile") or die "Cannot open $popFile for reading.\n";

# need only get the final values
@popEntry = ();
POPLOOP: while (<POPFILE>) {
    chomp;
    @popEntry = split;
}
close(POPFILE);

#discard time => check only on extinctions.
shift @popEntry;

# force allocation of 0 for all species
@specializations = ();
for ($i=0; $i<=$#popEntry; ++$i) {$specializations[$i] = 0;}

@initialSpecializations = ();
for ($i=0; $i<=$#popEntry; ++$i) {$initialSpecializations[$i] = 0;}

$first = 0;
LOOP: while (<FILE>) {

    chomp;

    /^#.*/ && next LOOP;
    /^$/ && next LOOP;

    # init: skip names
    if ($first==0) {
        $first = 1;
        next LOOP;
    }

    @entry = split;
    
    $gen = shift @entry;

    # Compute specialiazation as follow:
    # 
    # hungryLastResortRatioThreshold -> the greater, the more chance to graze
    # hungryGrassWeight              -> the lesser (greater negative), the more chance to graze
    # hungryPreyWeight               -> the lesser (positive), the more chance to graze
    # hungryPredatorWeight           -> has influence toward hunting, but not easily measureable => drop
    # hungryWeightedThreshold        -> the lesser, the more chance to graze
    #
    # But we'e not so much interested in whether the agents graze or hunt.
    # Rather, the amount of specialization is reflected by a smaller variance in each of these factors.
    # wheareas a broad variance indicates a wider range of behaviors.
    # => just add all standard deviations for these ratios, & the lesser the result, the more specialization.

    for ($i=0; $i<=$#specializations; ++$i) {
        $j = $i * 71;
        $numGen = $entry[$j + 70];
        # Get the last generation for which stats have a minimum of signification
        if ($numGen > 30) {
            $specializations[$i] = $entry[$j + 3]; # hungryLastResortRatioThreshold std dev
                                 + $entry[$j + 5]; # hungryGrassWeight std dev
                                 + $entry[$j + 7]; # hungryPreyWeight std dev
                                 + $entry[$j + 11]; # hungryWeightedThreshold std dev
            # Apply a factor for number clarity
            $specializations[$i] *= 100;
        }
    }
    
    if ($gen==0) {
        for ($i=0; $i<=$#specializations; ++$i) {$initialSpecializations[$i] = $specializations[$i];}
    }
}
close(FILE);

print "# Specialization indicator (e=extinct, a=alive):";
for ($i=0; $i<=$#specializations; ++$i) {
    # only monitor extinct species, final explosion for surviving specie has too much influence
    # let's compare what's comparable!
    if ($popEntry[$i]==0) {
        print " $specializations[$i](e)";
        $meanIndicator += $specializations[$i];
        $varIndicator += $specializations[$i] * $specializations[$i];
        $totalExtinctionNumber ++;
    } else {
        print " $specializations[$i](a)";
    }
    
    $meanInitialIndicator += $initialSpecializations[$i];
    $varInitialIndicator += $initialSpecializations[$i] * $initialSpecializations[$i];
}
print "\n";


} # end of file loop

# finish computations
$meanIndicator /= $totalExtinctionNumber;
$varIndicator = sqrt( ($varIndicator / $totalExtinctionNumber) - $meanIndicator * $meanIndicator );

$initialCount = (1 + $#ARGV) * ($#specializations + 1);
$meanInitialIndicator /= $initialCount;
$varInitialIndicator = sqrt( ($varInitialIndicator / $initialCount) - $meanInitialIndicator * $meanInitialIndicator );

print "# Mean specialization indicator before extinction: $meanIndicator, std. dev: $varIndicator\n";
print "# The smaller, the more species are specialized toward grazing or hunting. At the beginning of the runs, the indicator was on average $meanInitialIndicator, with std. dev $varInitialIndicator\n";
print "Specialization ratio (global) (mean end/beginning): ".($meanIndicator/$meanInitialIndicator)."\n";
print "Specialization ratio (inter-specie) (std.dev end/beginning): ".($varIndicator/$varInitialIndicator)."\n";

